@@ -6,6 +6,9 @@
 #if HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#if HAVE_NETDB_H
+#include <netdb.h>
+#endif
 #if HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
@@ -36,6 +39,7 @@
 #if HAVE_NETINET_IN_H
 #include <netinet/in.h>
 #endif
+#include <getopt.h>
 
 #include "asn1.h"
 #include "snmp_api.h"
@@ -45,6 +49,23 @@
 #include "system.h"
 #include "read_config.h"
 #include "snmp_debug.h"
+#include "snmp_parse_args.h"
+#include "agent_trap.h"
+#include "callback.h"
+#include "agent_callbacks.h"
+#include "tools.h"
+#include "snmp_logging.h"
+
+#include "mib_module_config.h"
+
+#ifdef USING_NOTIFICATION_SNMPNOTIFYTABLE_MODULE
+#include "snmp_vars.h"
+#include "snmp-tc.h"
+#include "target/snmpTargetAddrEntry.h"
+#include "target/snmpTargetParamsEntry.h"
+#include "notification/snmpNotifyTable.h"
+#endif
+
 
 struct trap_sink {
     struct snmp_session	*sesp;
@@ -103,10 +124,114 @@ static void send_v2_trap (struct snmp_session *, int, int, int);
 	 * Trap session handling
 	 *
 	 *******************/
-int add_trap_session( struct snmp_session *ss, int pdutype, int version )
+
+static void free_trap_session (struct trap_sink *sp)
 {
-    struct trap_sink *new_sink =
-      (struct trap_sink *) malloc (sizeof (*new_sink));
+    snmp_close(sp->sesp);
+    free (sp);
+}
+
+int add_trap_session( struct snmp_session *ss, int pdutype, int confirm, int version )
+{
+#ifdef USING_NOTIFICATION_SNMPNOTIFYTABLE_MODULE
+#define MAX_ENTRIES 1024
+    struct targetAddrTable_struct *ptr;
+    struct targetParamTable_struct *pptr;
+    struct snmpNotifyTable_data *nptr;
+    int i;
+    char buf[SNMP_MAXBUF_SMALL];
+    oid udpdomain[] = { 1,3,6,1,6,1,1 };
+    int udpdomainlen = sizeof(udpdomain)/sizeof(oid);
+#ifdef HAVE_GETHOSTBYNAME
+    struct hostent *hp;
+#endif
+
+    if (!ss)
+        return (0);
+
+        for(i=0; i < MAX_ENTRIES; i++) {
+            sprintf(buf, "internal%d", i);
+            if (get_addrForName(buf) == NULL && get_paramEntry(buf) == NULL)
+                break;
+        }
+        if (i == MAX_ENTRIES) {
+            snmp_sess_close(ss);
+            return(0);
+        }
+
+        /* address */
+        ptr = snmpTargetAddrTable_create();
+        ptr->name = strdup(buf);
+        memcpy(ptr->tDomain, udpdomain, udpdomainlen*sizeof(oid));
+        ptr->tDomainLen = udpdomainlen;
+
+#ifdef HAVE_GETHOSTBYNAME
+        hp = gethostbyname(ss->peername);
+        if (hp != NULL){
+            /* XXX: fix for other domain types */
+            ptr->tAddressLen = hp->h_length + 2;
+            ptr->tAddress = malloc(ptr->tAddressLen);
+            memmove(ptr->tAddress, hp->h_addr, hp->h_length);
+            ptr->tAddress[hp->h_length] = (ss->remote_port & 0xff00) >> 8;
+            ptr->tAddress[hp->h_length+1] = (ss->remote_port & 0xff);
+        } else {
+#endif /* HAVE_GETHOSTBYNAME */
+            ptr->tAddressLen = 6;
+            ptr->tAddress = calloc(1, ptr->tAddressLen);
+#ifdef HAVE_GETHOSTBYNAME
+        }
+#endif /* HAVE_GETHOSTBYNAME */
+        ptr->timeout = ss->timeout/1000;
+        ptr->retryCount = ss->retries;
+        ptr->tagList = strdup(ptr->name);
+        ptr->params = strdup(ptr->name);
+        ptr->storageType = ST_READONLY;
+        ptr->rowStatus = RS_ACTIVE;
+        ptr->sess = ss;
+        DEBUGMSGTL(("trapsess", "adding to trap table\n"));
+        snmpTargetAddrTable_add(ptr);
+
+        /* param */
+        pptr = snmpTargetParamTable_create();
+        pptr->paramName = strdup(buf);
+        pptr->mpModel = ss->version;
+        if (ss->version == SNMP_VERSION_3) {
+            pptr->secModel = ss->securityModel;
+            pptr->secLevel = ss->securityLevel;
+            memdup((void *) &pptr->secName, (void *) ss->securityName,
+                   ss->securityNameLen+1);
+	    pptr->secName[ss->securityNameLen] = 0;
+        } else {
+            pptr->secModel = ss->version;
+            pptr->secLevel = SNMP_SEC_LEVEL_NOAUTH;
+            pptr->secName = NULL;
+            if (ss->community && (ss->community_len > 0)) {
+                memdup((void *) &pptr->secName, (void *) ss->community,
+                   ss->community_len+1);
+		pptr->secName[ss->community_len] = 0;
+	    }
+        }
+        pptr->storageType = ST_READONLY;
+        pptr->rowStatus = RS_ACTIVE;
+        snmpTargetParamTable_add(pptr);
+            
+        /* notify table */
+        nptr = SNMP_MALLOC_STRUCT(snmpNotifyTable_data);
+        nptr->snmpNotifyName = strdup(buf);
+        nptr->snmpNotifyNameLen = strlen(buf);
+        nptr->snmpNotifyTag = strdup(buf);
+        nptr->snmpNotifyTagLen = strlen(buf);
+        nptr->snmpNotifyCommand = pdutype;
+        nptr->snmpNotifyType = confirm ?
+            SNMPNOTIFYTYPE_TRAP : SNMPNOTIFYTYPE_INFORM;
+        nptr->snmpNotifyStorageType = ST_READONLY;
+        nptr->snmpNotifyRowStatus = RS_ACTIVE;
+
+        snmpNotifyTable_add(nptr);
+#else /* ! USING_NOTIFICATION_SNMPNOTIFYTABLE_MODULE */
+    struct trap_sink *new_sink;
+    
+    new_sink = (struct trap_sink *) malloc (sizeof (*new_sink));
     if ( new_sink == NULL )
 	return 0;
 
@@ -115,9 +240,28 @@ int add_trap_session( struct snmp_session *ss, int pdutype, int version )
     new_sink->version = version;
     new_sink->next    = sinks;
     sinks = new_sink;
+#endif
     return 1;
 }
 
+int remove_trap_session( struct snmp_session *ss ) {
+    struct trap_sink *sp = sinks, *prev = 0;
+    while (sp) {
+        if (sp->sesp == ss) {
+            if (prev) {
+                prev->next = sp->next;
+            } else {
+                sinks = sp->next;
+            }
+            free_trap_session(sp);
+            return 1;
+        }
+        prev = sp;
+	sp = sp->next;
+    }
+    return 0;
+}
+
 int create_trap_session (char *sink, u_short sinkport,
 				char *com,
 				int version, int pdutype)
@@ -135,7 +279,7 @@ int create_trap_session (char *sink, u_short sinkport,
     sesp = snmp_open (&session);
 
     if (sesp) {
-	return( add_trap_session( sesp, pdutype, version ));
+	return( add_trap_session( sesp, pdutype, (pdutype==SNMP_MSG_INFORM), version ));
     }
 
     /* diagnose snmp_open errors with the input struct snmp_session pointer */
@@ -165,13 +309,6 @@ static int create_v2_inform_session (char *sink,  u_short sinkport,
 }
 
 
-static void free_trap_session (struct trap_sink *sp)
-{
-    snmp_close(sp->sesp);
-    free (sp);
-}
-
-
 void snmpd_free_trapsinks (void)
 {
     struct trap_sink *sp = sinks;
@@ -188,6 +325,48 @@ void snmpd_free_trapsinks (void)
 	 *
 	 *******************/
 
+void convert_v2_to_v1( struct variable_list *vars, struct snmp_pdu *template )
+{
+    struct variable_list *v, *trap_v=NULL, *ent_v=NULL;
+    oid  trap_prefix[] = { SNMPV2_TRAPS_PREFIX };
+    int len;
+
+    for ( v = vars ; v ; v = v->next_variable ) {
+	if ( snmp_oid_compare( v->name, v->name_length, 
+    			       snmptrap_oid, OID_LENGTH(snmptrap_oid)) == 0 )
+		trap_v = v;
+	if ( snmp_oid_compare( v->name, v->name_length, 
+    			       snmptrapenterprise_oid,
+			       OID_LENGTH(snmptrapenterprise_oid)) == 0 )
+		ent_v = v;
+    }
+
+    if ( !trap_v )
+	return;		/* Can't find v2 snmpTrapOID varbind */
+
+		/*
+		 * Is this a 'standard' trap?
+		 *  Or at least, does it have the correct prefix?
+		 */
+    if ( snmp_oid_compare( trap_v->val.objid, OID_LENGTH(trap_prefix), 
+    			   trap_prefix,       OID_LENGTH(trap_prefix)) == 0 ) {
+	template->trap_type     = trap_v->val.objid[OID_LENGTH(trap_prefix)]-1;
+	template->specific_type = 0;
+    }
+    else {
+	len = trap_v->val_len /sizeof( oid );
+	template->trap_type     = 6;	/* enterprise specific */
+	template->specific_type = trap_v->val.objid[len-1];
+    }
+
+		/*
+		 *  TODO:
+		 *    Extract the appropriate enterprise value from 'ent_v'
+		 *    Remove uptime/trapOID varbinds from 'vars' list
+		 */
+    
+}
+
 void send_enterprise_trap_vars (int trap, 
 		     int specific,
 		     oid *enterprise, int enterprise_length,
@@ -195,7 +374,7 @@ void send_enterprise_trap_vars (int trap,
 {
     struct variable_list uptime_var, snmptrap_var, enterprise_var;
     struct variable_list *v2_vars, *last_var=NULL;
-    struct snmp_pdu	*template_pdu, *pdu;
+    struct snmp_pdu	*template_pdu;
     struct timeval	 now;
     long uptime;
     struct sockaddr_in *pduIp;
@@ -280,6 +459,7 @@ void send_enterprise_trap_vars (int trap,
 			snmptrap_var.next_variable = vars;
 		}
 		last_var = NULL;	/* Don't need enterprise info */
+		convert_v2_to_v1( vars, template_pdu );
 		break;
 
 			/* "Standard" SNMPv1 traps */
@@ -305,8 +485,10 @@ void send_enterprise_trap_vars (int trap,
 				    sizeof(link_up_oid));
 		break;
 	case SNMP_TRAP_AUTHFAIL:
-		if (snmp_enableauthentraps == SNMP_AUTHENTICATED_TRAPS_DISABLED)
+		if (snmp_enableauthentraps == SNMP_AUTHENTICATED_TRAPS_DISABLED) {
+		    snmp_free_pdu(template_pdu);
 		    return;
+		}
 		snmp_set_var_value( &snmptrap_var,
 				    (u_char *)auth_fail_oid,
 				    sizeof(auth_fail_oid));
@@ -318,13 +500,13 @@ void send_enterprise_trap_vars (int trap,
 		break;
 
 	case SNMP_TRAP_ENTERPRISESPECIFIC:
-		memcpy( &temp_oid,
+		memcpy( temp_oid,
 				    (char *)enterprise,
 				    (enterprise_length)*sizeof(oid));
 		temp_oid[ enterprise_length   ] = 0;
 		temp_oid[ enterprise_length+1 ] = specific;
 		snmp_set_var_value( &snmptrap_var,
-				    (u_char *)&temp_oid,
+				    (u_char *)temp_oid,
 				    (enterprise_length+2)*sizeof(oid));
 		snmptrap_var.next_variable  = vars;
 		last_var = NULL;	/* Don't need version info */
@@ -339,8 +521,8 @@ void send_enterprise_trap_vars (int trap,
     for ( sink = sinks ; sink ; sink=sink->next ) {
 	if ( sink->version == SNMP_VERSION_1 && trap == -1 )
 		continue;	/* Skip v1 sinks for v2 only traps */
-	template_pdu->version = sink->version;
 	template_pdu->command = sink->pdutype;
+
 	if ( sink->version != SNMP_VERSION_1 ) {
 	    template_pdu->variables = v2_vars;
 	    if ( last_var )
@@ -349,21 +531,31 @@ void send_enterprise_trap_vars (int trap,
 	else
 	    template_pdu->variables = vars;
 
-	pdu = snmp_clone_pdu( template_pdu );
-	pdu->sessid = sink->sesp->sessid;	/* AgentX only ? */
-	if ( snmp_send( sink->sesp, pdu) == 0 ) {
-            snmp_sess_perror ("snmpd: send_trap", sink->sesp);
-	    snmp_free_pdu( pdu );
-	}
-	else {
-	    snmp_increment_statistic(STAT_SNMPOUTTRAPS);
-	    snmp_increment_statistic(STAT_SNMPOUTPKTS);
-	}
-		
+        send_trap_to_sess(sink->sesp, template_pdu);
+
 	if ( sink->version != SNMP_VERSION_1 && last_var )
 	    last_var->next_variable = NULL;
     }
 
+    /* send stuff to registered callbacks */
+    /* v2 traps/informs */
+    template_pdu->variables = v2_vars;
+    if ( last_var )
+        last_var->next_variable = &enterprise_var;
+
+    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, SNMPD_CALLBACK_SEND_TRAP2,
+                        template_pdu);
+    
+    if ( last_var )
+        last_var->next_variable = NULL;
+
+    /* v1 traps */
+    template_pdu->command = SNMP_MSG_TRAP;
+    template_pdu->variables = vars;
+    
+    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, SNMPD_CALLBACK_SEND_TRAP1,
+                        template_pdu);
+
 	/* Ensure we don't free anything we shouldn't */
     if ( last_var )
 	last_var->next_variable = NULL;
@@ -371,6 +563,35 @@ void send_enterprise_trap_vars (int trap,
     snmp_free_pdu( template_pdu );
 }
 
+/* send_trap_to_sess: sends a trap to a session but assumes that the
+   pdu is constructed correctly for the session type. */
+void send_trap_to_sess(struct snmp_session *sess,
+                       struct snmp_pdu *template_pdu) {
+    struct snmp_pdu *pdu;
+
+    if (!sess || !template_pdu)
+        return;
+
+    DEBUGMSGTL(("send_trap_to_sess","sending trap type=%d, version=%d\n",
+                template_pdu->command, sess->version));
+    
+    if (sess->version == SNMP_VERSION_1 &&
+        (template_pdu->command == SNMP_MSG_TRAP2 ||
+         template_pdu->command == SNMP_MSG_INFORM))
+        return;	/* Skip v1 sinks for v2 only traps */
+    template_pdu->version = sess->version;
+    pdu = snmp_clone_pdu( template_pdu );
+    pdu->sessid = sess->sessid;	/* AgentX only ? */
+    if ( snmp_send( sess, pdu) == 0 ) {
+        snmp_sess_perror ("snmpd: send_trap", sess);
+        snmp_free_pdu( pdu );
+    }
+    else {
+        snmp_increment_statistic(STAT_SNMPOUTTRAPS);
+        snmp_increment_statistic(STAT_SNMPOUTPKTS);
+    }
+}
+
 void send_trap_vars (int trap, 
 		     int specific,
 		     struct variable_list *vars)
@@ -508,6 +729,62 @@ snmpd_parse_config_informsink(const char *word, char *cptr)
     }
 }
 
+/* this must be standardized somewhere, right? */
+#define MAX_ARGS 128
+
+static int traptype;
+
+static void trapOptProc(int argc, char *const *argv, int opt)
+{
+    switch (opt) {
+        case 'C':
+            while (*optarg) {
+                switch (*optarg++) {
+                    case 'i':
+                        traptype = SNMP_MSG_INFORM;
+                        break;
+                    default:
+                        config_perror("unknown argument passed to -C");
+                        break;
+                }
+            }
+            break;
+    }
+}
+
+
+void
+snmpd_parse_config_trapsess(const char *word, char *cptr) {
+    char args[MAX_ARGS][SPRINT_MAX_LEN], *argv[MAX_ARGS];
+    int argn, arg;
+    struct snmp_session session, *ss;
+
+    /* inform or trap?  default to trap */
+    traptype = SNMP_MSG_TRAP2;
+
+    /* create the argv[] like array */
+    strcpy(argv[0] = args[0], "snmpd-trapsess"); /* bogus entry for getopt() */
+    for(argn = 1; cptr && argn < MAX_ARGS;
+        cptr = copy_word(cptr, argv[argn] = args[argn++])) {
+    }
+
+    arg = snmp_parse_args(argn, argv, &session, "C:", trapOptProc);
+
+    ss = snmp_open (&session);
+
+    if (!ss) {
+        config_perror("snmpd: failed to parse this line");
+        snmp_sess_perror("snmpd: snmpd_parse_config_trapsess()", &session);
+        return;
+    }
+    
+    if (ss->version == SNMP_VERSION_1) {
+	add_trap_session(ss, SNMP_MSG_TRAP, 0, SNMP_VERSION_1);
+    } else {
+	add_trap_session( ss, traptype, (traptype==SNMP_MSG_INFORM), ss->version );
+    }
+}
+
 void
 snmpd_parse_config_trapcommunity(const char *word, char *cptr)
 {
