@@ -46,31 +46,57 @@
 #include "var_struct.h"
 #include "mibII/sysORTable.h"
 
+extern int close_agentx_session(struct snmp_session *session, int sessid);
+
 #define VARLIST_ITERATION	10
 
 struct ax_variable_list {
     struct agent_snmp_session	*asp;
     int				num_vars;
     int				max_vars;
-#ifdef ONE_IDEA
-    {
-	struct variable_list    *entry;
-	int		        index;
-    }				variables[MAX_VARS];
-#else
 		/* Placeholder for dynamically-resized list of variables */
-    struct variable_list*        variables[1];
-#endif
+    struct variable_list*       variables[1];
 };
 
+void
+free_agentx_request(struct request_list *req, int free_cback)
+{
+    if ( !req )
+	return;
+
+    /*
+     *	Only free this call-back data when request processing has
+     *	completed.  Multi-pass requests (i.e. SET and GETBULK handling)
+     *	use this to store information that needs to be persistent
+     *	from one pass to the next.
+     */
+
+    if ( req->cb_data && free_cback )
+	free ( req->cb_data );
+
+     free ( req );
+}
+
+void
+free_agentx_varlist(struct ax_variable_list *vlist)
+{
+    if ( !vlist )
+	return;
+    if ( vlist->variables )
+	free ( vlist->variables );
+     free ( vlist );
+}
+
 	/*
 	 * Remove the specified request from
 	 *  the list of outstanding requests
+	 *  and return it, for later disposal
 	 */
-int
+struct request_list *
 remove_outstanding_request( struct agent_snmp_session *asp, int reqid )
 {
     struct request_list     *req, *prev;
+
     for (req = asp->outstanding_requests, prev=NULL ;
 			req != NULL ;
 			prev = req , req = req->next_request ) {
@@ -79,11 +105,11 @@ remove_outstanding_request( struct agent_snmp_session *asp, int reqid )
 		prev->next_request = req->next_request;
 	    else
 		asp->outstanding_requests = req->next_request;
-	    free( req );
-	    return 0;
+
+	    return req;
 	}
     }
-    return 1;
+    return NULL;
 }
 
 	/*
@@ -100,122 +126,203 @@ handle_agentx_response( int operation,
     struct ax_variable_list *ax_vlist = (struct ax_variable_list *)magic;
     struct variable_list *vbp, *next;
     struct agent_snmp_session *asp  =  ax_vlist->asp;
-    int i;
-    struct ax_variable_list *retry_vlist;
-    struct variable_list    *vb_retry, *vbp2;
+    struct snmp_session *ax_session;
+    struct request_list *req, *oldreq;
+    int  free_cback = 1;
+    int i, type, index;
+    int oldstatus;
     struct subtree          *retry_sub;
-    int j;
     char buf[SPRINT_MAX_LEN];
 
-    remove_outstanding_request( asp, pdu->reqid );
+    oldreq = remove_outstanding_request( asp, pdu->reqid );
 
-    retry_vlist = NULL;
-    vb_retry    = NULL;
-    vbp2        = NULL;
+    switch(operation) {
+	case TIMED_OUT:
+			/*
+			 * Multiple timed out requests probably
+			 *  indicate that the subagent has died.
+			 */
+		if ( SET_SNMP_STRIKE_FLAGS( session->flags )) {
+		    ax_session = session->subsession;
+				/*
+				 * XXX - Need to send a 'close' message
+				 *       to the subagent (even though
+				 *	 we don't think it's listening)
+				 */
+		    (void) close_agentx_session(ax_session, session->sessid);
+		    if ( ax_session->subsession == NULL ) {
+			snmp_close( ax_session );
+		    }
+		}
+		pdu->errstat  = SNMP_ERR_GENERR;
+		pdu->errindex = 0;
+		if (asp->pdu->command != SNMP_MSG_SET)
+		    asp->mode = RESERVE2;
+		break;
 
-    DEBUGMSGTL(("agentx/master","handle_agentx_response() beginning...\n"));
-    for ( i = 0, vbp = pdu->variables ;
-          vbp && i < ax_vlist->num_vars ;
-          i++, vbp = vbp->next_variable ) {
+	case SEND_FAILED:
+		if ( SET_SNMP_STRIKE_FLAGS( session->flags )) {
+		    (void) close_agentx_session(session, -1 );
+		}
+		pdu->errstat  = SNMP_ERR_GENERR;
+		pdu->errindex = 0;
+		if (asp->pdu->command != SNMP_MSG_SET)
+		    asp->mode = RESERVE2;
+		break;
 
-      if (vbp) {
-        DEBUGMSGTL(("agentx/master","  handle_agentx_response: processing: "));
-        DEBUGMSGOID(("agentx/master",vbp->name, vbp->name_length));
-        DEBUGMSG(("agentx/master","\n"));
-      }
-	if ( ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_VERBOSE) ) {
-	    sprint_variable (buf, vbp->name, vbp->name_length, vbp);
-	    DEBUGMSGTL(("snmp_agent", "    >> %s\n", buf));
-	}
 
-	if ( !asp->exact && (vbp->type == SNMP_ENDOFMIBVIEW ||
+	case RECEIVED_MESSAGE:
+		/* This session is alive */
+		CLEAR_SNMP_STRIKE_FLAGS( session->flags );
+		break;
+	default:
+		return 0;
+    }
+
+
+    oldstatus = asp->status;
+    asp->status = pdu->errstat;
+    if ( pdu->errstat != AGENTX_ERR_NOERROR ) {
+		/*
+		 *  If the request failed, locate the
+		 *    original index of the variable resonsible
+		 */
+	if ( pdu->errindex != 0 && pdu->errindex < ax_vlist->num_vars )
+	    asp->index = ax_vlist->variables[pdu->errindex-1]->index;
+        else
+	    asp->index = 0;
+
+    }
+    else {
+		/*
+		 * Otherwise, process successful requests
+		 */
+	DEBUGMSGTL(("agentx/master","handle_agentx_response() beginning...\n"));
+	for ( i = 0, vbp = pdu->variables ;
+              vbp && i < ax_vlist->num_vars ;
+              i++, vbp = vbp->next_variable ) {
+
+            if (vbp) {
+		DEBUGMSGTL(("agentx/master","  handle_agentx_response: processing: "));
+		DEBUGMSGOID(("agentx/master",vbp->name, vbp->name_length));
+		DEBUGMSG(("agentx/master","\n"));
+		if ( ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_VERBOSE) ) {
+		    sprint_variable (buf, vbp->name, vbp->name_length, vbp);
+		    DEBUGMSGTL(("snmp_agent", "    >> %s\n", buf));
+		}
+	    }
+
+	    if ( !asp->exact && (vbp->type == SNMP_ENDOFMIBVIEW ||
 			     in_a_view( vbp->name, &vbp->name_length,
                                         asp->pdu,  vbp->type ))) {
-	    /*
-	     *   Add unfulfilled requests to "retry" list
-	     */
-	    retry_sub = find_subtree_next( vbp->name, vbp->name_length, NULL );
-	    if ( retry_sub == NULL ) {
-		next = ax_vlist->variables[i]->next_variable;
-	        snmp_clone_var( vbp, ax_vlist->variables[i] );
-		ax_vlist->variables[i]->next_variable = next;
-			/* XXX - handle SNMPv1 */
-		ax_vlist->variables[i]->type = SNMP_ENDOFMIBVIEW;
-		continue;
-	    }
-	    if (retry_vlist == NULL ) {
-		retry_vlist = (struct ax_variable_list *)
-				calloc( 1, sizeof( struct ax_variable_list));
-		if (retry_vlist == NULL)
-		    break;
-
-		vb_retry = (struct variable_list *)
-				calloc( 1, sizeof( struct variable_list));
-		if (vb_retry == NULL) {
-		    free(retry_vlist);
-		    break;
-		}
-
-		vbp2 = vb_retry;
+	        /*
+	         *   Retry unfulfilled requests
+	         */
+	        retry_sub = find_subtree_next( vbp->name, vbp->name_length, NULL );
+	        if ( retry_sub ) {
+		    (void)snmp_set_var_objid(ax_vlist->variables[i], retry_sub->name, retry_sub->namelen);
+		    asp->index = ax_vlist->variables[i]->index;
+		    asp->status = handle_one_var(asp, ax_vlist->variables[i]);
+	        }
+	        else
+		    ax_vlist->variables[i]->type = SNMP_ENDOFMIBVIEW;
 	    }
 	    else {
-		vbp2->next_variable = (struct variable_list *)
-				calloc( 1, sizeof( struct variable_list));
-		if (vbp2->next_variable == NULL)
-		    break;
-		vbp2 = vbp2->next_variable;
+		next  = ax_vlist->variables[i]->next_variable;
+		index = ax_vlist->variables[i]->index;
+        	snmp_clone_var( vbp, ax_vlist->variables[i] );
+		ax_vlist->variables[i]->next_variable = next;
+		ax_vlist->variables[i]->index         = index;
+	    }
+
+	    type = ax_vlist->variables[i]->type;
+	    if ((asp->pdu->version == SNMP_VERSION_1) &&
+	        ((type == SNMP_ENDOFMIBVIEW) || (type == SNMP_NOSUCHOBJECT) || (type == SNMP_NOSUCHINSTANCE))) {
+		    asp->index = ax_vlist->variables[i]->index;
+		    asp->status = SNMP_ERR_NOSUCHNAME;
+		    goto finish;
 	    }
-	    j = retry_vlist->num_vars;
-	    retry_vlist->variables[ j ] = ax_vlist->variables[i];
-	    retry_vlist->num_vars++;
-	    memcpy(vbp->name, retry_sub->name, retry_sub->namelen * sizeof(oid));
-	    vbp->name_length = retry_sub->namelen;
-	    next = vbp2->next_variable;
-	    snmp_clone_var( vbp, vbp2 );
-	    vbp2->next_variable = next;
 	}
-	else
-		/* Got an answer, so use it */
-	if ( pdu->errstat != AGENTX_ERR_NOERROR ) {
-	    asp->status = pdu->errstat;
+    }
+
+	/*
+	 *  If we're in the middle of a SET,
+	 *	the state machine will have been updated on the
+	 *	(mistaken) assumption that the request succeeded.
+	 *  In fact the lack of an error from the 'agentx_var()'
+	 *	routine  simply meant that the delegation succeeded. 
+	 *
+	 *  If the subagent reports that the request failed,
+	 *	we need to tweak the mode to reflect this.
+	 */
+    if (asp->pdu->command == SNMP_MSG_SET) {
+	switch( asp->mode ) {
+
+	    case ACTION:	/* I.e. a 'successful' RESERVE2 pass */
+		if ( asp->status != SNMP_ERR_NOERROR )
+			asp->mode = FREE;
+		break;
+
+	    case COMMIT:	/* I.e. a 'successful' ACTION pass */
+		if ( asp->status != SNMP_ERR_NOERROR )
+			asp->mode = UNDO;
+		break;
+
+	    case FREE:
+		asp->mode   = FINISHED_FAILURE;
+		asp->status = oldstatus;
+		break;
+
+	    case FINISHED_SUCCESS:	/* I.e. a 'successful' COMMIT pass */
+		if ( asp->status != SNMP_ERR_NOERROR ) {
+			asp->mode   = FINISHED_FAILURE;
+			asp->status = SNMP_ERR_COMMITFAILED;
+		}
+		break;
+
+	    case FINISHED_FAILURE:	/* I.e. a 'successful' UNDO pass */
+		if ( asp->status != SNMP_ERR_NOERROR ) {
+			asp->mode   = FINISHED_FAILURE;
+			asp->status = SNMP_ERR_UNDOFAILED;
+		}
+		else {
+			asp->status = oldstatus;
+		}
+		break;
 	}
-	else {
-	    next = ax_vlist->variables[i]->next_variable;
-	    snmp_clone_var( vbp, ax_vlist->variables[i] );
-	    ax_vlist->variables[i]->next_variable = next;
-	}	
     }
 
 
-    if ( retry_vlist ) {
+    if ( asp->outstanding_requests ) {
 		/*
-		 * unfulfilled GETNEXT requests
-		 *
-		 * Need a new 'handle_getnext_request' routine
-		 *   (parameters to be determined)
+		 * Send out any newly delegated requests
+		 * 	See 'handle_one_var' above
 		 */
-	asp->start = vb_retry;
-	asp->end   = vbp2;
-	j = handle_var_list( asp );
-	if ( j != SNMP_ERR_NOERROR )
-	    asp->status = j;
-	else for ( i = 0, vbp2 = vb_retry ;
-	           i < retry_vlist->num_vars ;
-	           i++ , vbp2=vbp2->next_variable ) {
-			/*
-			 * XXX - assumes success second time around
-			 *	Needs to handle deferrals / failures
-			 */
-	    		next = retry_vlist->variables[i]->next_variable;
-	 		snmp_clone_var( vbp2, retry_vlist->variables[i] );
-	    		retry_vlist->variables[i]->next_variable = next;
+	for ( req=asp->outstanding_requests ; req ; req=req->next_request ) {
+	    if ( req->pdu ) {
+		snmp_async_send(req->session, req->pdu,
+				req->callback, req->cb_data);
+		req->pdu = NULL;
+	    }
 	}
-	return j;
     }
-    else
-      DEBUGMSGTL(("agentx/master","handle_agentx_response() finishing...\n"));
-	return handle_snmp_packet(operation, session, reqid,
-					asp->pdu, (void*)asp);
+
+finish:
+			 /*
+			  * Free the old request, but if we
+			  * haven't completed a multi-pass request,
+			  * then don't free the callback data
+			  */
+    if ( oldreq ) {
+	if ( oldreq->pdu && oldreq->pdu->command == SNMP_MSG_SET &&
+		(( asp->mode == ACTION ) ||
+		 ( asp->mode == COMMIT )))
+	    free_cback = 0;
+	free_agentx_request( oldreq, free_cback );
+    }
+
+    DEBUGMSGTL(("agentx/master","handle_agentx_response() finishing...\n"));
+    return handle_snmp_packet(operation, session, reqid, asp->pdu, (void*)asp);
 }
 
 
@@ -249,7 +356,7 @@ get_agentx_request(struct agent_snmp_session *asp,
 		    (vlist->max_vars + VARLIST_ITERATION) * sizeof(struct variable_list);
 		vlist = (struct ax_variable_list *)realloc(vlist, new_size);
 		if ( !vlist )
-		    break;
+		    break; 
 		vlist->max_vars += VARLIST_ITERATION;
 		req->cb_data = (void *)vlist;
 	    }
@@ -266,9 +373,9 @@ get_agentx_request(struct agent_snmp_session *asp,
     vlist = (struct ax_variable_list *)calloc( 1, new_size);
     pdu   = snmp_pdu_create( 0 );
     if ( req == NULL || pdu == NULL || vlist == NULL ) {
-	free( pdu );
-	free( vlist );
-	free( req );
+	free_agentx_request( req, 1 );
+	snmp_free_pdu(       pdu );
+	free_agentx_varlist( vlist );
 	return NULL;
     }
 
@@ -317,9 +424,9 @@ get_agentx_request(struct agent_snmp_session *asp,
 		break;
 	default:
                 DEBUGMSGTL(("agentx/master","-> unknown\n"));
-		free( req );
-		free( pdu );
-		free( vlist );
+		free_agentx_request( req, 1 );
+		snmp_free_pdu(       pdu );
+		free_agentx_varlist( vlist );
 		return NULL;
     }
    
@@ -364,27 +471,35 @@ agentx_add_request( struct agent_snmp_session *asp,
 	return AGENTX_ERR_NOERROR;
 
     ax_session = get_session_for_oid( vbp->name, vbp->name_length );
+    if ( !ax_session )
+	return SNMP_ERR_GENERR;
     sessid = ax_session->sessid;
-		/* WHat if this returns NULL ? */
     if ( ax_session->flags & SNMP_FLAGS_SUBSESSION )
 	ax_session = ax_session->subsession;
     request    = get_agentx_request( asp, ax_session, pdu->transid );
+    if ( !request )
+	return SNMP_ERR_GENERR;
     request->pdu->sessid = sessid;	/* Use the registered (sub)session's ID,
 					   not the main listening session ID */
     ax_vlist   = (struct ax_variable_list *)request->cb_data;
 
     ax_vlist->variables[ ax_vlist->num_vars ] = vbp;
+    vbp->index = asp->index;	/* Remember the variable index */
     ax_vlist->num_vars++;
     
+    sub = find_subtree_previous( vbp->name, vbp->name_length, NULL );
     if ( asp->exact )
         snmp_pdu_add_variable( request->pdu,
 			   vbp->name, vbp->name_length, vbp->type,
 			   (u_char*)(vbp->val.string), vbp->val_len);
     else {
-	sub = find_subtree_previous( vbp->name, vbp->name_length, NULL );
         snmp_pdu_add_variable( request->pdu,
 			   vbp->name, vbp->name_length, ASN_PRIV_EXCL_RANGE,
-			   (u_char*)sub->end, sub->end_len);
+			   (u_char*)sub->end, sub->end_len*sizeof(oid));
+    }
+    if ( sub->timeout > request->pdu->time ) {
+	request->pdu->time = sub->timeout;
+	request->pdu->flags |= UCD_MSG_FLAG_PDU_TIMEOUT;
     }
 
     return AGENTX_ERR_NOERROR;
