@@ -62,19 +62,18 @@ struct agent_set_info {
     struct agent_set_info *next;
 };
 
-static struct agent_set_info *Sets;
+static struct agent_set_info *Sets = NULL;
 
-void
+struct agent_set_info *
 save_set_vars( struct snmp_session *ss, struct snmp_pdu *pdu )
 {
     struct agent_set_info *ptr;
     struct timeval now;
-    struct snmp_pdu *pdu2;
     extern struct timeval starttime;
 
     ptr = (struct agent_set_info *)malloc(sizeof(struct agent_set_info));
     if (ptr == NULL )
-	return;
+	return NULL;
 
 	/*
 	 * Save the important information
@@ -85,58 +84,47 @@ save_set_vars( struct snmp_session *ss, struct snmp_pdu *pdu )
     gettimeofday(&now, NULL);
     ptr->uptime = calculate_time_diff(&now, &starttime);
 
-		/* Lazy way to copy the variables! */
-    pdu2 = snmp_clone_pdu( pdu );
-    if ( pdu2 == NULL ) {
+    ptr->var_list = snmp_clone_varbind(pdu->variables);
+    if ( ptr->var_list == NULL ) {
 	free( ptr );
-	return;
+	return NULL;
     }
-    ptr->var_list = pdu2->variables;	
-    pdu2->variables = NULL;
-    snmp_free_pdu( pdu2 );
 
     ptr->next = Sets;
     Sets = ptr;
+
+    return ptr;
 }
 
-void
+struct agent_set_info *
 restore_set_vars( struct agent_snmp_session *asp )
 {
     struct agent_set_info *ptr;
 
     for ( ptr=Sets ; ptr != NULL ; ptr=ptr->next )
-	if ( ptr->sess == asp->session && ptr->transID == asp->pdu->transid ) {
-	    if ( ptr->var_list == NULL )
-		return;
-	    asp->rw      	= WRITE;
-	    asp->pdu->variables = ptr->var_list;
-	    asp->start		= ptr->var_list;
-	    asp->end		= ptr->var_list;
-	    while ( asp->end->next_variable )
-		asp->end = asp->end->next_variable;
-	    ptr->mode		= asp->mode;
-	    return;
-	}
-    return;
+	if ( ptr->sess == asp->session && ptr->transID == asp->pdu->transid )
+	    break;
+
+    if ( ptr == NULL || ptr->var_list == NULL )
+	return NULL;
+
+    asp->rw      	= WRITE;
+    asp->pdu->variables = ptr->var_list;
+    asp->start		= ptr->var_list;
+    asp->end		= ptr->var_list;
+    while ( asp->end->next_variable )
+	asp->end = asp->end->next_variable;
+    asp->mode		= ptr->mode;
+    return ptr;
 }
 
-int
-set_mode( struct snmp_session *ss, struct snmp_pdu *pdu )
-{
-    struct agent_set_info *ptr;
-
-    for ( ptr=Sets ; ptr != NULL ; ptr=ptr->next )
-	if ( ptr->sess == ss && ptr->transID == pdu->transid )
-	    return ptr->mode;
-    return -1;
-}
 
 void
 free_set_vars( struct snmp_session *ss, struct snmp_pdu *pdu )
 {
     struct agent_set_info *ptr, *prev=NULL;
 
-    for ( ptr=Sets ; ptr != NULL ; ptr=ptr->next )
+    for ( ptr=Sets ; ptr != NULL ; ptr=ptr->next ) {
 	if ( ptr->sess == ss && ptr->transID == pdu->transid ) {
 	    if ( prev )
 		prev->next = ptr->next;
@@ -146,6 +134,8 @@ free_set_vars( struct snmp_session *ss, struct snmp_pdu *pdu )
 	    free(ptr);
 	    return;
 	}
+	prev = ptr;
+    }
 }
 
 
@@ -154,10 +144,11 @@ handle_agentx_packet(int operation, struct snmp_session *session, int reqid,
                    struct snmp_pdu *pdu, void *magic)
 {
     struct agent_snmp_session  *asp;
+    struct agent_set_info      *asi;
     int status, allDone, i;
     struct variable_list *var_ptr, *var_ptr2;
 
-    asp = init_agent_snmp_session( session, snmp_clone_pdu(pdu) );
+    asp = init_agent_snmp_session( session, pdu );
 
     DEBUGMSGTL(("agentx/subagent","handling agentx request....\n"));
     switch (pdu->command) {
@@ -241,26 +232,24 @@ handle_agentx_packet(int operation, struct snmp_session *session, int reqid,
 	     * These correspond to the first AgentX pass - TESTSET
 	     */
 	asp->rw      = WRITE;
-
         asp->mode = RESERVE1;
-	save_set_vars( session, pdu );
-	status = handle_next_pass( asp );
+	asi = save_set_vars( session, pdu );
+	if ( asi ) {
+	    status = handle_next_pass( asp );
+	}
+	else
+	    status = AGENTX_ERR_PROCESSING_ERROR;
 
-	if ( status != SNMP_ERR_NOERROR ) {
-	    asp->mode = FREE;
-	    (void) handle_next_pass( asp );
-	    free_set_vars( session, pdu );
-	    break;
+	if ( status == SNMP_ERR_NOERROR ) {
+	    asp->mode = RESERVE2;
+	    status = handle_next_pass( asp );
 	}
 
-        asp->mode = RESERVE2;
-	status = handle_next_pass( asp );
-
-	if ( status != SNMP_ERR_NOERROR ) {
-	    asp->mode = FREE;
-	    (void) handle_next_pass( asp );
-	    free_set_vars( session, pdu );
-	}
+	if ( status == SNMP_ERR_NOERROR )
+	    asi->mode = ACTION;
+	else
+	    asi->mode = FREE;
+	
 	break;
 
 
@@ -272,39 +261,31 @@ handle_agentx_packet(int operation, struct snmp_session *session, int reqid,
     case AGENTX_MSG_COMMITSET:
         DEBUGMSGTL(("agentx/subagent","  -> commitset\n"));
         asp->mode = ACTION;
-        restore_set_vars( asp );
-
-        if ( asp->pdu->variables == NULL )
-            status = AGENTX_ERR_PROCESSING_ERROR;
-        else
+        asi = restore_set_vars( asp );
+	if ( asi ) {
 	    status = handle_next_pass( asp );
-
-	if ( status != SNMP_ERR_NOERROR ) {
-	    asp->mode = UNDO;
-	    (void) handle_next_pass( asp );
-	    free_set_vars( session, pdu );
 	}
+	else
+	    status = AGENTX_ERR_PROCESSING_ERROR;
+
+	if ( status == SNMP_ERR_NOERROR )
+	    asi->mode = COMMIT;
+	else
+	    asi->mode = UNDO;
+	
         asp->pdu->variables = NULL;
 	break;
 
     case AGENTX_MSG_CLEANUPSET:
         DEBUGMSGTL(("agentx/subagent","  -> cleanupset\n"));
-        switch ( set_mode( session, pdu )) {
-            case RESERVE1: 
-            case RESERVE2: 
-            case UNDO: 
-                asp->mode = FREE;
-                break;
-            case ACTION:
-                asp->mode = COMMIT;
-                break;
-        }
-        restore_set_vars( asp );
-
-        if ( asp->pdu->variables == NULL )
-            status = AGENTX_ERR_PROCESSING_ERROR;
-        else
+        asi = restore_set_vars( asp );
+	if ( asi ) {
+	    asp->mode = asi->mode;
 	    status = handle_next_pass( asp );
+	}
+	else
+	    status = AGENTX_ERR_PROCESSING_ERROR;
+
 	free_set_vars( session, pdu );
         asp->pdu->variables = NULL;
 	break;
@@ -312,12 +293,13 @@ handle_agentx_packet(int operation, struct snmp_session *session, int reqid,
     case AGENTX_MSG_UNDOSET:
         DEBUGMSGTL(("agentx/subagent","  -> undoset\n"));
         asp->mode = UNDO;
-        restore_set_vars( asp );
-
-        if ( asp->pdu->variables == NULL )
-            status = AGENTX_ERR_PROCESSING_ERROR;
-        else
+        asi = restore_set_vars( asp );
+	if ( asi ) {
 	    status = handle_next_pass( asp );
+	}
+	else
+	    status = AGENTX_ERR_PROCESSING_ERROR;
+
 	free_set_vars( session, pdu );
         asp->pdu->variables = NULL;
 	break;
@@ -325,7 +307,7 @@ handle_agentx_packet(int operation, struct snmp_session *session, int reqid,
     case AGENTX_MSG_RESPONSE:
         DEBUGMSGTL(("agentx/subagent","  -> response\n"));
 	free( asp );
-	return 0;
+	return 1;
 
     default:
         DEBUGMSGTL(("agentx/subagent","  -> unknown (%d)\n", pdu->command ));
@@ -336,10 +318,18 @@ handle_agentx_packet(int operation, struct snmp_session *session, int reqid,
     
 	
     if ( asp->outstanding_requests == NULL ) {
+	if ( status != SNMP_ERR_NOERROR ) {
+	    snmp_free_pdu( asp->pdu );
+	    asp->pdu = asp->orig_pdu;
+	    asp->orig_pdu = NULL;
+	}
 	asp->pdu->command = AGENTX_MSG_RESPONSE;
-	asp->pdu->errstat = status;
-	snmp_send( asp->session, asp->pdu );
-	free( asp );
+	asp->pdu->errstat  = status;
+	asp->pdu->errindex = asp->index;
+	if (! snmp_send( asp->session, asp->pdu ))
+	    snmp_free_pdu(asp->pdu);
+	asp->pdu = NULL;
+	free_agent_snmp_session( asp );
     }
     DEBUGMSGTL(("agentx/subagent","  FINISHED\n\n"));
 
@@ -361,7 +351,8 @@ agentx_registration_callback(int majorID, int minorID, void *serverarg,
     return agentx_register(agentx_ss,
 		    reg_parms->name, reg_parms->namelen,
 		    reg_parms->priority,
-		    reg_parms->range_subid, reg_parms->range_ubound);
+		    reg_parms->range_subid, reg_parms->range_ubound,
+		    reg_parms->timeout);
   else
     return agentx_unregister(agentx_ss,
 		    reg_parms->name, reg_parms->namelen,
@@ -412,7 +403,7 @@ subagent_register_for_traps(int majorID, int minorID, void *serverarg, void *cli
     DEBUGMSGTL(("agentx/subagent","No session to register\n"));
     return 0;
   }
-  if (add_trap_session( thesession, AGENTX_MSG_NOTIFY, AGENTX_VERSION_1) == 0){
+  if (add_trap_session( thesession, AGENTX_MSG_NOTIFY, 1, AGENTX_VERSION_1) == 0){
     DEBUGMSGTL(("agentx/subagent","Trap session registration failed\n"));
     return 0;
   }
@@ -420,21 +411,22 @@ subagent_register_for_traps(int majorID, int minorID, void *serverarg, void *cli
   return 1;
 }
 
-void
+/* returns non-zero on error */
+int
 subagent_pre_init( void )
 {
     struct snmp_session sess;
 
     DEBUGMSGTL(("agentx/subagent","initializing....\n"));
     if ( ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_ROLE) != SUB_AGENT )
-	return;
+	return 0;
 
     snmp_sess_init( &sess );
     sess.version = AGENTX_VERSION_1;
     sess.retries = SNMP_DEFAULT_RETRIES;
     sess.timeout = SNMP_DEFAULT_TIMEOUT;
     sess.flags  |= SNMP_FLAGS_STREAM_SOCKET;
-    if ( ds_get_string(DS_APPLICATION_ID, DS_AGENT_X_SOCKET))
+    if ( ds_get_string(DS_APPLICATION_ID, DS_AGENT_X_SOCKET) )
 	sess.peername = strdup(ds_get_string(DS_APPLICATION_ID, DS_AGENT_X_SOCKET));
     else
 	sess.peername = strdup(AGENTX_SOCKET);
@@ -449,12 +441,12 @@ subagent_pre_init( void )
     if ( main_session == NULL ) {
       /* diagnose snmp_open errors with the input struct snmp_session pointer */
 	snmp_sess_perror("subagent_pre_init", &sess);
-	exit(1);
+	return -1;
     }
 
     if ( agentx_open_session( main_session ) < 0 ) {
 	snmp_close( main_session );
-	exit(1);
+	return -1;
     }
 
 
@@ -477,6 +469,8 @@ subagent_pre_init( void )
                            agentx_sysOR_callback, main_session);
 #endif
     DEBUGMSGTL(("agentx/subagent","initializing....  DONE\n"));
+    
+    return 0;
 }
 
 
