@@ -32,15 +32,47 @@ header_complex_generate_varoid(struct variable_list *var) {
         var->name[0] = *(var->val.integer);
         break;
 
-      case ASN_OPAQUE:
-      case ASN_OCTET_STR:
-        var->name_length = var->val_len+1;
-        var->name = (oid *) malloc(sizeof(oid) * (var->val_len + 1));
+      case ASN_PRIV_IMPLIED_OBJECT_ID:
+        var->name_length = var->val_len/sizeof(oid);
+        var->name = (oid *) malloc(sizeof(oid) * (var->name_length));
         if (var->name == NULL)
             return SNMPERR_GENERR;
-        var->name[0] = var->val_len;
-        for(i=0; i < var->val_len; i++)
-          var->name[i+1] = var->val.string[i];
+
+        for(i = 0; i < var->name_length; i++)
+          var->name[i] = var->val.objid[i];
+        break;
+
+      case ASN_OBJECT_ID:
+        var->name_length = var->val_len/sizeof(oid) + 1;
+        var->name = (oid *) malloc(sizeof(oid) * (var->name_length));
+        if (var->name == NULL)
+            return SNMPERR_GENERR;
+
+        var->name[0] = var->name_length-1;
+        for(i = 0; i < var->name_length-1; i++)
+          var->name[i+1] = var->val.objid[i];
+        break;
+        
+      case ASN_PRIV_IMPLIED_OCTET_STR:
+        var->name_length = var->val_len;
+        var->name = (oid *) malloc(sizeof(oid) * (var->name_length));
+        if (var->name == NULL)
+            return SNMPERR_GENERR;
+
+        for(i = 0; i < var->val_len; i++)
+          var->name[i] = (oid) var->val.string[i];
+        break;
+
+      case ASN_OPAQUE:
+      case ASN_OCTET_STR:
+        var->name_length = var->val_len + 1;
+        var->name = (oid *) malloc(sizeof(oid) * (var->name_length));
+        if (var->name == NULL)
+            return SNMPERR_GENERR;
+
+        var->name[0] = (oid) var->val_len;
+        for(i = 0; i < var->val_len; i++)
+          var->name[i+1] = (oid) var->val.string[i];
         break;
       
       default:
@@ -65,21 +97,7 @@ header_complex_generate_varoid(struct variable_list *var) {
 
    returns 1 if an error is encountered, or 0 if successful.
 */
-static int _header_complex_parse_oid(oid *oidIndex, size_t oidLen,
-                         struct variable_list *data);
-int
-header_complex_parse_oid(oid *oidIndex, size_t oidLen,
-                         struct variable_list *data)
-{
-    int rc;
-    memset(data, 0, sizeof(*data));
-    rc = _header_complex_parse_oid(oidIndex, oidLen, data);
-    if (rc == SNMPERR_GENERR)
-        snmp_free_varbind(data);
-    return (rc);
-}
-
-static int _header_complex_parse_oid(oid *oidIndex, size_t oidLen,
+int header_complex_parse_oid(oid *oidIndex, size_t oidLen,
                          struct variable_list *data)
 {
   struct variable_list *var = data;
@@ -102,13 +120,46 @@ static int _header_complex_parse_oid(oid *oidIndex, size_t oidLen,
                     "Parsed int(%d): %d\n", var->type, *var->val.integer));
         break;
 
+      case ASN_OBJECT_ID:
+      case ASN_PRIV_IMPLIED_OBJECT_ID:
+        if (var->type == ASN_PRIV_IMPLIED_OBJECT_ID) {
+            itmp = oidLen;
+        } else {
+            itmp = (long) *oidIndex++;
+            oidLen--;
+            if (itmp > oidLen)
+                return SNMPERR_GENERR;
+        }
+
+        if (itmp == 0)
+          break;        /* zero length strings shouldn't malloc */
+        
+        var->val_len = itmp*sizeof(oid);
+        var->val.objid = (oid *) calloc(1,var->val_len);
+        if (var->val.objid == NULL)
+            return SNMPERR_GENERR;
+
+        for(i = 0; i < itmp; i++)
+          var->val.objid[i] = (u_char) *oidIndex++;
+        oidLen -= itmp;
+
+        DEBUGMSGTL(("header_complex_parse_oid", "Parsed oid: "));
+        DEBUGMSGOID(("header_complex_parse_oid", var->val.objid, var->val_len/sizeof(oid)));
+        DEBUGMSG(("header_complex_parse_oid", "\n"));
+        break;
+
       case ASN_OPAQUE:
       case ASN_OCTET_STR:
-        itmp = (long) *oidIndex++;
-        oidLen--;
-        if (itmp > oidLen)
-          return SNMPERR_GENERR;
-          
+      case ASN_PRIV_IMPLIED_OCTET_STR:
+        if (var->type == ASN_PRIV_IMPLIED_OCTET_STR) {
+            itmp = oidLen;
+        } else {
+            itmp = (long) *oidIndex++;
+            oidLen--;
+            if (itmp > oidLen)
+                return SNMPERR_GENERR;
+        }
+
         if (itmp == 0)
           break;        /* zero length strings shouldn't malloc */
         
@@ -171,6 +222,25 @@ header_complex_generate_oid(oid *name, /* out */
   DEBUGMSG(("header_complex_generate_oid", "\n"));
 }
 
+/* finds the data in "datalist" stored at "index" */
+void *
+header_complex_get(struct header_complex_index *datalist,
+                    struct variable_list *index) {
+    oid searchfor[MAX_OID_LEN];
+    size_t searchfor_len;
+    struct header_complex_index *nptr;
+    
+    header_complex_generate_oid(searchfor,     /* out */
+                                &searchfor_len, /* out */
+                                NULL, 0, index);
+    for(nptr = datalist; nptr != NULL; nptr = nptr->next) {
+        if (snmp_oid_compare(searchfor, searchfor_len,
+                             nptr->name, nptr->namelen) == 0)
+            return nptr->data;
+    }
+    return NULL;
+}
+
 
 void *
 header_complex(struct header_complex_index *datalist,
@@ -193,10 +263,15 @@ header_complex(struct header_complex_index *datalist,
     *var_len = sizeof (long);
 
   for(nptr = datalist; nptr != NULL && found == NULL; nptr = nptr->next) {
-    memcpy(indexOid, vp->name, vp->namelen * sizeof(oid));
-    memcpy(indexOid + vp->namelen, nptr->name,
-           nptr->namelen * sizeof(oid));
-    len = vp->namelen + nptr->namelen;
+    if (vp) {
+        memcpy(indexOid, vp->name, vp->namelen * sizeof(oid));
+        memcpy(indexOid + vp->namelen, nptr->name,
+               nptr->namelen * sizeof(oid));
+        len = vp->namelen + nptr->namelen;
+    } else {
+        memcpy(indexOid, nptr->name, nptr->namelen * sizeof(oid));
+        len = nptr->namelen;
+    }
     result = snmp_oid_compare(name, *length, indexOid, len);
     DEBUGMSGTL(("header_complex", "Checking: "));
     DEBUGMSGOID(("header_complex", indexOid, len));
@@ -217,10 +292,15 @@ header_complex(struct header_complex_index *datalist,
     }
   }
   if (found) {
-      memcpy(name, vp->name, vp->namelen * sizeof(oid));
-      memcpy(name + vp->namelen, found->name,
-             found->namelen * sizeof(oid));
-      *length = vp->namelen + found->namelen;
+      if (vp) {
+          memcpy(name, vp->name, vp->namelen * sizeof(oid));
+          memcpy(name + vp->namelen, found->name,
+                 found->namelen * sizeof(oid));
+          *length = vp->namelen + found->namelen;
+      } else {
+          memcpy(name, found->name, found->namelen * sizeof(oid));
+          *length = found->namelen;
+      }
       return found->data;
   }
     
@@ -233,7 +313,6 @@ header_complex_add_data(struct header_complex_index **thedata,
   struct header_complex_index *hciptrn, *hciptrp, *ourself;
   oid newoid[MAX_OID_LEN];
   size_t newoid_len;
-  struct variable_list *vp, *vp2;
 
   if (thedata == NULL || var == NULL || data == NULL)
     return NULL;
