@@ -4,6 +4,11 @@
  */
 
 #include <config.h>
+#if HAVE_STRING_H
+#include <string.h>
+#else
+#include <strings.h>
+#endif
 #include "host_res.h"
 #include "hr_print.h"
 
@@ -18,8 +23,11 @@
 
 void  Init_HR_Print (void);
 int Get_Next_HR_Print (void);
-/* void  Save_HR_Print (void); */
-/* char*  describe_printer (void); */
+void  Save_HR_Print (void);
+const char *describe_printer (int);
+int printer_status(int);
+int printer_detail_status(int);
+int printer_errors(int);
 int header_hrprint (struct variable *,oid *, size_t *, int, size_t *, WriteMethod **);
 
 
@@ -45,10 +53,12 @@ void init_hr_print(void)
     next_device[ HRDEV_PRINTER ] = Get_Next_HR_Print;
 /*  save_device[ HRDEV_PRINTER ] = Save_HR_Print;	*/
 #ifdef HRPRINT_MONOTONICALLY_INCREASING
-    dev_idx_inc[ HRDEV_DISK ] = 1;
+    dev_idx_inc[ HRDEV_PRINTER ] = 1;
 #endif
 
-/*  device_descr[ HRDEV_PRINTER ] = &describe_printer;	*/
+    device_descr[ HRDEV_PRINTER ] = describe_printer;
+    device_status[ HRDEV_PRINTER ] = printer_status;
+    device_errors[ HRDEV_PRINTER ] = printer_errors;
 
     REGISTER_MIB("host/hr_print", hrprint_variables, variable4, hrprint_variables_oid);
 }
@@ -147,10 +157,7 @@ var_hrprint(struct variable *vp,
 
     switch (vp->magic){
 	case HRPRINT_STATUS:
-#if NO_DUMMY_VALUES
-	    return NULL;
-#endif
-	    long_return = 3;	/* idle */
+	    long_return = printer_detail_status(print_idx);
 	    return (u_char *)&long_return;
 	case HRPRINT_ERROR:
 #if NO_DUMMY_VALUES
@@ -159,7 +166,7 @@ var_hrprint(struct variable *vp,
 	    long_return = 0;	/* Null string */
 	    return (u_char *)&long_return;
 	default:
-	    DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_hrprint\n", vp->magic));
+	    DEBUGMSGTL(("host/hr_print", "unknown sub-id %d in var_hrprint\n", vp->magic));
     }
     return NULL;
 }
@@ -172,11 +179,76 @@ var_hrprint(struct variable *vp,
 	 *********************/
 
 static int HRP_index;
+static char **HRP_name;
+static int HRP_names, HRP_maxnames;
 
 void
 Init_HR_Print (void)
 {
-   HRP_index = 1;
+#if HAVE_LPSTAT || HAVE_CGETNEXT || HAVE_PRINTCAP
+    int i;
+#if HAVE_CGETNEXT
+    const char *caps[] = {"/etc/printcap", NULL};
+#elif HAVE_LPSTAT || HAVE_PRINTCAP
+    FILE *p;
+#endif
+
+    if (HRP_name) {
+        for (i = 0; i < HRP_names; i++) free(HRP_name[i]);
+	HRP_names = 0;
+    }
+    else {
+	HRP_maxnames = 5;
+	HRP_name = calloc(HRP_maxnames, sizeof( char *));
+    }
+  
+#if HAVE_LPSTAT
+    if ((p = popen(LPSTAT_PATH " -v", "r")) != NULL) {
+	char buf[BUFSIZ], ptr[BUFSIZ];
+	while (fgets(buf, sizeof buf, p)) {
+	    sscanf(buf, "%*s %*s %[^:]", ptr);
+#elif HAVE_CGETNEXT
+    {
+	char *buf, *ptr;
+	while (cgetnext(&buf, caps)) {
+	    if ((ptr = strchr(buf, ':'))) *ptr = 0;
+	    if ((ptr = strchr(buf, '|'))) *ptr = 0;
+	    ptr = buf;
+#elif HAVE_PRINTCAP
+    if ((p = fopen("/etc/printcap", "r")) != NULL) {
+	char buf[BUFSIZ], *ptr;
+	while (fgets(buf, sizeof buf, p)) {
+	    buf[strlen(buf)-1] = 0;
+	    if (buf[0] == '#' || buf[0] == 0 || buf[0] == ' ' || buf[0] == '\t')
+		continue;
+	    if ((ptr = strchr(buf, '\\'))) *ptr = 0;
+	    if ((ptr = strchr(buf, ':'))) *ptr = 0;
+	    if ((ptr = strchr(buf, '|'))) *ptr = 0;
+	    ptr = buf;
+#endif
+	    if (HRP_names == HRP_maxnames) {
+		char **tmp;
+		HRP_maxnames += 5;
+		tmp = calloc(HRP_maxnames, sizeof(char *));
+		memcpy(tmp, HRP_name, HRP_names*sizeof(char *));
+		HRP_name = tmp;
+	    }
+	    HRP_name[HRP_names++] = strdup(ptr);
+#if HAVE_CGETNEXT
+	    free(buf);
+#endif
+	}
+#if HAVE_LPSTAT
+	pclose(p);
+#elif HAVE_CGETNEXT
+	cgetclose();
+#elif HAVE_PRINTCAP
+	fclose(p);
+#endif
+    }
+#endif /* HAVE_anything */
+
+    HRP_index = 0;
 }
 
 int
@@ -188,8 +260,41 @@ Get_Next_HR_Print (void)
 			 *   no real idea how to detect them,
 			 *   so don't bother.
 			 */
-    if ( HRP_index < 1 ) 	/* No printer */
+    if ( HRP_index < HRP_names ) 	/* No printer */
         return ( HRDEV_PRINTER << HRDEV_TYPE_SHIFT ) + HRP_index++;
     else
         return -1;
 }
+
+const char *describe_printer(int idx)
+{
+    DEBUGMSGTL(("host/hr_print", "describe p: %d/%d %s\n", HRP_index, idx, HRP_name[HRP_index-1]));
+    return HRP_name[HRP_index-1];
+}
+
+int printer_status(int idx)
+{
+/*hrDeviceStatus OBJECT-TYPE
+    SYNTAX     INTEGER {
+                   unknown(1), running(2), warning(3), testing(4), down(5)
+               }
+*/
+
+    return 1;	/* unknown */
+}
+
+int printer_detail_status(int idx)
+{
+/*hrPrinterStatus OBJECT-TYPE
+    SYNTAX     INTEGER {
+                   other(1), unknown(2), idle(3), printing(4), warmup(5)
+               }
+*/
+
+    return 2;	/* unknown */
+}
+
+int printer_errors(int idx)
+{
+    return 0;
+}
