@@ -15,7 +15,6 @@
 #define _I_DEFINED_KERNEL
 #endif
 #include <sys/types.h>
-#include <sys/socket.h>
 #if TIME_WITH_SYS_TIME
 # include <sys/time.h>
 # include <time.h>
@@ -26,6 +25,9 @@
 #  include <time.h>
 # endif
 #endif
+#if HAVE_SYS_SOCKET_H
+#include <sys/socket.h>
+#endif
 
 #if HAVE_NETINET_IN_H
 #include <netinet/in.h>
@@ -464,6 +466,7 @@ static void ARP_Scan_Init (void)
 	else
 		at = NULL;
 	for (i = 0; i < arptab_size; i++) {
+		u_long tmp_a;
 		while (line == fgets (line, sizeof(line), in) &&
 			11 != sscanf (line, "%d.%d.%d.%d 0x%*x 0x%x %x:%x:%x:%x:%x:%x",
 			&za, &zb, &zc, &zd, &at[i].at_flags,
@@ -475,7 +478,11 @@ static void ARP_Scan_Init (void)
 		at [i].at_enaddr[3] = zh;
 		at [i].at_enaddr[4] = zi;
 		at [i].at_enaddr[5] = zj;
-		at [i].at_iaddr.s_addr = (zd << 24) | (zc << 16) | (zb << 8) | za;
+		tmp_a = ((u_long)za << 24) |
+			((u_long)zb << 16) |
+			((u_long)zc <<  8) |
+			((u_long)zd      ) ;
+		at [i].at_iaddr.s_addr = htonl(tmp_a);
 	}
 	fclose (in);
 #endif /* linux */
@@ -494,7 +501,7 @@ static void ARP_Scan_Init (void)
 		free(at);
 	if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
 		snmp_log_perror("route-sysctl-estimate");
-	if ((at = malloc(needed)) == NULL)
+	if ((at = malloc(needed ? needed : 1)) == NULL)
 		snmp_log_perror("malloc");
 	if (sysctl(mib, 6, at, &needed, NULL, 0) < 0)
 		snmp_log_perror("actual retrieval of routing table");
