@@ -141,13 +141,14 @@
 #   endif
 #  endif
 # endif
-#endif
 #endif /* defined(freebsd3) */
+#endif /* HAVE_SYS_SYSCTL_H */
 
 /* #include "../common_header.h" */
 
 #include "system.h"
 #include "snmp_logging.h"
+#include "read_config.h"
 
 #if HAVE_OSRELDATE_H
 #include <osreldate.h>
@@ -156,13 +157,18 @@
 #include <sys/sysctl.h>
 #endif
 #include "interfaces.h"
-#include "../struct.h"
-#include "../util_funcs.h"
+#include "struct.h"
+#include "util_funcs.h"
 #include "auto_nlist.h"
 #include "sysORTable.h"
 
+extern struct timeval starttime;
+
 static int Interface_Scan_Get_Count (void);
 
+static void parse_interface_config(const char *, char *);
+static void free_interface_config(void);
+
 struct variable4 interfaces_variables[] = {
     {IFNUMBER, ASN_INTEGER, RONLY, var_interfaces, 1, {1}},
     {IFINDEX, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 1}},
@@ -171,7 +177,7 @@ struct variable4 interfaces_variables[] = {
     {IFMTU, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 4}},
     {IFSPEED, ASN_GAUGE, RONLY, var_ifEntry, 3, {2, 1, 5}},
     {IFPHYSADDRESS, ASN_OCTET_STR, RONLY, var_ifEntry, 3, {2, 1, 6}},
-    {IFADMINSTATUS, ASN_INTEGER, RWRITE, var_ifEntry, 3, {2, 1, 7}},
+    {IFADMINSTATUS, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 7}},
     {IFOPERSTATUS, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 8}},
     {IFLASTCHANGE, ASN_TIMETICKS, RONLY, var_ifEntry, 3, {2, 1, 9}},
     {IFINOCTETS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 10}},
@@ -202,6 +208,9 @@ void init_interfaces(void)
   REGISTER_SYSOR_ENTRY(interfaces_module_oid,
 	"The MIB module to describe generic objects for network interface sub-layers");
   
+  snmpd_register_config_handler("interface", parse_interface_config,
+    				free_interface_config, "name type speed");
+
 #ifndef USE_SYSCTL_IFLIST
 #if HAVE_NET_IF_MIB_H
   init_interfaces_setup();
@@ -225,6 +234,7 @@ if_type_from_name( const char *pcch)
     static match_if lmatch_if[] = {
       { 24, "lo" },
       {  6, "eth" },
+      {  9, "tr" },
       { 23, "ppp" },
       { 28, "sl" },
       {  0, 0 }  /* end of list */
@@ -244,11 +254,82 @@ if_type_from_name( const char *pcch)
 }
 
 
-#ifdef USE_SYSCTL_IFLIST
+typedef struct _conf_if_list {
+    char *name;
+    int type;
+    u_long speed;
+    struct _conf_if_list *next;
+} conf_if_list;
+
+static conf_if_list *conf_list;
+static struct ifnet *ifnetaddr_list;
+
+static void parse_interface_config(const char *token, char *cptr)
+{
+    conf_if_list *if_ptr, *if_new;
+    char *name, *type, *speed, *ecp;
+
+    name = strtok(cptr, " \t");
+    if (!name) {
+	config_perror("Missing NAME parameter");
+	return;
+    }
+    type = strtok(NULL, " \t");
+    if (!type) {
+	config_perror("Missing TYPE parameter");
+	return;
+    }
+    speed = strtok(NULL, " \t");
+    if (!speed) {
+	config_perror("Missing SPEED parameter");
+	return;
+    }
+    if_ptr = conf_list;
+    while (if_ptr)
+	if (strcmp(if_ptr->name, name)) if_ptr = if_ptr->next;
+	else break;
+    if (if_ptr)
+	config_pwarn("Duplicate interface specification");
+    if_new = (conf_if_list *)malloc(sizeof(conf_if_list));
+    if (!if_new) {
+	config_perror("Out of memory");
+	return;
+    }
+    if_new->speed = strtoul(speed, &ecp, 0);
+    if (*ecp) {
+	config_perror("Bad SPEED value");
+	free(if_new);
+	return;
+    }
+    if_new->type = strtol(type, &ecp, 0);
+    if (*ecp || if_new->type < 0) {
+	config_perror("Bad TYPE");
+	free(if_new);
+	return;
+    }
+    if_new->name = strdup(name);
+    if (!if_new->name) {
+	config_perror("Out of memory");
+	free(if_new);
+	return;
+    }
+    if_new->next = conf_list;
+    conf_list = if_new;
+}
+
+static void free_interface_config(void)
+{
+    conf_if_list *if_ptr = conf_list, *if_next;
+    while (if_ptr) {
+	if_next = if_ptr->next;
+	free(if_ptr->name);
+	free(if_ptr);
+	if_ptr = if_next;
+    }
+    conf_list = NULL;
+}
+
 
-static u_char * if_list = 0;
-static const u_char * if_list_end;
-static size_t if_list_size = 0;
 
 /*
   header_ifEntry(...
@@ -307,46 +388,6 @@ header_ifEntry(struct variable *vp,
 }
 
 
-/*
-  header_interfaces(...
-  Arguments:
-  vp	  IN      - pointer to variable entry that points here
-  name    IN/OUT  - IN/name requested, OUT/name found
-  length  IN/OUT  - length of IN/OUT oid's 
-  exact   IN      - TRUE if an exact match was requested
-  var_len OUT     - length of variable or 0 if function returned
-  write_method
-  
-*/
-
-static int
-header_interfaces(struct variable *vp,
-		  oid *name,
-		  size_t *length,
-		  int exact,
-		  size_t *var_len,
-		  WriteMethod **write_method)
-{
-#define INTERFACES_NAME_LENGTH	8
-  oid newname[MAX_OID_LEN];
-  int result;
-
-    DEBUGMSGTL(("mibII/interfaces", "var_interfaces: "));
-    DEBUGMSGOID(("mibII/interfaces", name, *length));
-    DEBUGMSG(("mibII/interfaces"," %d\n", exact));
-
-  memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
-  newname[INTERFACES_NAME_LENGTH] = 0;
-  result = snmp_oid_compare(name, *length, newname, (int)vp->namelen + 1);
-  if ((exact && (result != 0)) || (!exact && (result >= 0)))
-    return MATCH_FAILED;
-  memcpy( (char *)name,(char *)newname, ((int)vp->namelen + 1) * sizeof(oid));
-  *length = vp->namelen + 1;
-
-  *write_method = 0;
-  *var_len = sizeof(long);	/* default to 'long' results */
-  return MATCH_SUCCEEDED;
-}
 
 u_char *
 var_interfaces(struct variable *vp,
@@ -356,7 +397,7 @@ var_interfaces(struct variable *vp,
 	       size_t *var_len,
 	       WriteMethod **write_method)
 {
-  if (header_interfaces(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
+  if (header_generic(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
     return NULL;
 
   switch (vp->magic)
@@ -370,6 +411,12 @@ var_interfaces(struct variable *vp,
   return NULL;
 }
 
+#ifdef USE_SYSCTL_IFLIST
+
+static u_char * if_list = 0;
+static const u_char * if_list_end;
+static size_t if_list_size = 0;
+
 struct small_ifaddr
 {
   struct in_addr	sifa_addr;
@@ -410,7 +457,7 @@ Interface_Scan_By_Index (int iindex,
 	      {
 		a = get_address (ifp+1, ifp->ifm_addrs, RTA_IFP);
                 if (a == NULL)
-                  return NULL;
+                  return 0;
 		strncpy (if_name,
 			 ((const struct sockaddr_in *) a)->sin_zero,
 			 ((const u_char *) a)[5]);
@@ -440,14 +487,14 @@ Interface_Scan_By_Index (int iindex,
 				     ifap->ifam_addrs &= ~RTA_NETMASK,
 				     RTA_IFA);
                 if (ia == NULL)
-                  return NULL;
+                  return 0;
                 
 		sifa->sifa_addr = *ia;
 		ia = get_in_address ((char *) (ifap+1)+8,
 				     ifap->ifam_addrs &= ~RTA_NETMASK,
 				     RTA_BRD);
                 if (ia == NULL)
-                  return NULL;
+                  return 0;
 
 		sifa->sifa_broadcast = *ia;
 		++have_addr;
@@ -569,7 +616,12 @@ var_ifEntry(struct variable *vp,
     long_return = (long) if_msg.ifm_data.ifi_mtu;
     return (u_char *) &long_return;
   case IFSPEED:
+#if STRUCT_IFNET_HAS_IF_BAUDRATE_IFS_VALUE 
+    long_return = (u_long) if_msg.ifm_data.ifi_baudrate.ifs_value <<
+                           if_msg.ifm_data.ifi_baudrate.ifs_log2;
+#else
     long_return = (u_long) if_msg.ifm_data.ifi_baudrate;
+#endif
     return (u_char *) &long_return;
   case IFPHYSADDRESS:
     /* XXX */
@@ -621,7 +673,18 @@ var_ifEntry(struct variable *vp,
   case IFOUTERRORS:
     long_return = (u_long) if_msg.ifm_data.ifi_oerrors;
     return (u_char *) &long_return;
-    /* ifOutQLen */
+  case IFLASTCHANGE:
+    if (if_msg.ifm_data.ifi_lastchange.tv_sec == 0 &&
+	if_msg.ifm_data.ifi_lastchange.tv_usec == 0)
+      long_return = 0;
+    else if (if_msg.ifm_data.ifi_lastchange.tv_sec < starttime.tv_sec)
+      long_return = 0;
+    else {
+      long_return = (u_long)
+	((if_msg.ifm_data.ifi_lastchange.tv_sec - starttime.tv_sec) * 100
+	 + (if_msg.ifm_data.ifi_lastchange.tv_usec - starttime.tv_usec) / 10000);
+    }
+    return (u_char *) &long_return;
   default:
     return 0;
   }
@@ -651,103 +714,6 @@ static int Interface_Scan_By_Index (int, char *, struct ifnet *, struct in_ifadd
 static int Interface_Get_Ether_By_Index (int, u_char *);
 #endif
 
-	/*********************
-	 *
-	 *  Initialisation & common implementation functions
-	 *
-	 *********************/
-
-
-
-
-
-#ifdef linux
-typedef struct _conf_if_list {
-    char *name;
-    int type;
-    int speed;
-    struct _conf_if_list *next;
-} conf_if_list;
-
-conf_if_list *if_list;
-struct ifnet *ifnetaddr_list;
-#endif /* linux */
-
-static int
-header_interfaces(struct variable *vp,
-		  oid *name,
-		  size_t *length,
-		  int exact,
-		  size_t *var_len,
-		  WriteMethod **write_method)
-{
-#define INTERFACES_NAME_LENGTH	8
-    oid newname[MAX_OID_LEN];
-    int result;
-
-    DEBUGMSGTL(("mibII/interfaces", "var_interfaces: "));
-    DEBUGMSGOID(("mibII/interfaces", name, *length));
-    DEBUGMSG(("mibII/interfaces"," %d\n", exact));
-
-    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
-    newname[INTERFACES_NAME_LENGTH] = 0;
-    result = snmp_oid_compare(name, *length, newname, vp->namelen + 1);
-    if ((exact && (result != 0)) || (!exact && (result >= 0)))
-        return MATCH_FAILED;
-    memcpy( (char *)name,(char *)newname, ((int)vp->namelen + 1) * sizeof(oid));
-    *length = vp->namelen + 1;
-
-    *write_method = 0;
-    *var_len = sizeof(long);	/* default to 'long' results */
-    return MATCH_SUCCEEDED;
-}
-
-
-
-static int
-header_ifEntry(struct variable *vp,
-	       oid *name,
-	       size_t *length,
-	       int exact,
-	       size_t *var_len,
-	       WriteMethod **write_method)
-{
-#define IFENTRY_NAME_LENGTH	10
-    oid newname[MAX_OID_LEN];
-    register int	interface;
-    int result, count;
-
-    DEBUGMSGTL(("mibII/interfaces", "var_ifEntry: "));
-    DEBUGMSGOID(("mibII/interfaces", name, *length));
-    DEBUGMSG(("mibII/interfaces"," %d\n", exact));
-
-    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
-    /* find "next" interface */
-    count = Interface_Scan_Get_Count();
-    for(interface = 1; interface <= count; interface++){
-	newname[IFENTRY_NAME_LENGTH] = (oid)interface;
-	result = snmp_oid_compare(name, *length, newname, vp->namelen + 1);
-	if ((exact && (result == 0)) || (!exact && (result < 0)))
-	    break;
-    }
-    if (interface > count) {
-        DEBUGMSGTL(("mibII/interfaces", "... index out of range\n"));
-        return MATCH_FAILED;
-    }
-
-
-    memcpy( (char *)name,(char *)newname, ((int)vp->namelen + 1) * sizeof(oid));
-    *length = vp->namelen + 1;
-    *write_method = 0;
-    *var_len = sizeof(long);	/* default to 'long' results */
-
-    DEBUGMSGTL(("mibII/interfaces", "... get I/F stats "));
-    DEBUGMSGOID(("mibII/interfaces", name, *length));
-    DEBUGMSG(("mibII/interfaces","\n"));
-
-    return interface;
-}
-
 
 
 	/*********************
@@ -756,27 +722,6 @@ header_ifEntry(struct variable *vp,
 	 *
 	 *********************/
 
-u_char	*
-var_interfaces(struct variable *vp,
-	       oid *name,
-	       size_t *length,
-	       int exact,
-	       size_t *var_len,
-	       WriteMethod **write_method)
-{
-    if (header_interfaces(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
-	return NULL;
-
-    switch (vp->magic){
-	case IFNUMBER:
-	    long_return = Interface_Scan_Get_Count();
-	    return (u_char *)&long_return;
-	default:
-	    DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_interfaces\n", vp->magic));
-    }
-    return NULL;
-}
-
 
 #ifndef solaris2
 #ifndef hpux
@@ -790,10 +735,11 @@ var_ifEntry(struct variable *vp,
 	    WriteMethod **write_method)
 {
     static struct ifnet ifnet;
-    register int interface;
+    int interface;
     static struct in_ifaddr in_ifaddr;
     static char Name[16];
-    register char *cp;
+    char *cp;
+    conf_if_list *if_ptr = conf_list;
 #if STRUCT_IFNET_HAS_IF_LASTCHANGE_TV_SEC
     struct timeval now;
 #endif
@@ -803,6 +749,7 @@ var_ifEntry(struct variable *vp,
 	return NULL;
 
     Interface_Scan_By_Index(interface, Name, &ifnet, &in_ifaddr);
+    while (if_ptr && strcmp(Name, if_ptr->name)) if_ptr = if_ptr->next;
 
     switch (vp->magic){
 	case IFINDEX:
@@ -816,7 +763,8 @@ var_ifEntry(struct variable *vp,
 #if STRUCT_IFNET_HAS_IF_TYPE
 	    long_return = ifnet.if_type;
 #else
-	    long_return = 1;	/* OTHER */
+	    if (if_ptr) long_return = if_ptr->type;
+	    else long_return = 1;	/* OTHER */
 #endif
 	    return (u_char *) &long_return;
 	case IFMTU: {
@@ -829,7 +777,8 @@ var_ifEntry(struct variable *vp,
 #elif STRUCT_IFNET_HAS_IF_SPEED
 	    long_return = ifnet.if_speed;
 #elif STRUCT_IFNET_HAS_IF_TYPE && defined(IFT_ETHER)
-	    if((long_return == 0) || (long_return == 1)) {
+	    if (if_ptr) long_return = if_ptr->speed;
+	    else {
 		if(ifnet.if_type == IFT_ETHER) long_return=10000000;
 		if(ifnet.if_type == IFT_P10) long_return=10000000;
 		if(ifnet.if_type == IFT_P80) long_return=80000000;
@@ -867,14 +816,15 @@ var_ifEntry(struct variable *vp,
  * a packet was last input or output.  In the 2.1.5 and later releases,
  * this is fixed, thus the 199607 comparison.
  */
-          if ((ifnet.if_lastchange.tv_sec == 0 ) &&
-              (ifnet.if_lastchange.tv_usec == 0))
+          if (ifnet.if_lastchange.tv_sec == 0 &&
+              ifnet.if_lastchange.tv_usec == 0)
             long_return = 0;
+	  else if (ifnet.if_lastchange.tv_sec < starttime.tv_sec)
+	    long_return = 0;
           else {
-            gettimeofday(&now, (struct timezone *)0);
             long_return = (u_long)
-              ((now.tv_sec - ifnet.if_lastchange.tv_sec) * 100
-               + (now.tv_usec - ifnet.if_lastchange.tv_usec) / 10000);
+              ((ifnet.if_lastchange.tv_sec - starttime.tv_sec) * 100
+               + (ifnet.if_lastchange.tv_usec - starttime.tv_usec) / 10000);
           }
 #else
 #if NO_DUMMY_VALUES
@@ -1271,13 +1221,12 @@ void
 Interface_Scan_Init (void)
 {
 #ifdef linux
-    char line [256], fullname [64], ifname_buf [64], *ifname, *ptr;
+    char line [256], ifname_buf [64], *ifname, *ptr;
     struct ifreq ifrq;
     struct ifnet **ifnetaddr_ptr;
     FILE *devin;
     unsigned long rec_pkt, rec_oct, rec_err, snd_pkt, snd_oct, snd_err, coll;
     int i, fd;
-    extern conf_if_list *if_list;
     conf_if_list *if_ptr;
     const char *scan_line_2_2="%lu %lu %lu %*lu %*lu %*lu %*lu %*lu %lu %lu %lu %*lu %*lu %lu";
     const char *scan_line_2_0="%lu %lu %*lu %*lu %*lu %lu %lu %*lu %*lu %lu";
@@ -1343,7 +1292,8 @@ Interface_Scan_Init (void)
 	char *stats, *ifstart = line;
 
 	while (*ifstart == ' ') ifstart++;
-	stats = strrchr(ifstart, ':');
+	if ((stats = strrchr(ifstart, ':')) == NULL)  /* /proc/net/dev bug */
+            continue;
 	*stats++ = 0;
 	strcpy(ifname_buf, ifstart);
 	while (*stats == ' ') stats++;
@@ -1385,7 +1335,7 @@ Interface_Scan_Init (void)
 	nnew->if_name = (char *) strdup (ifname);
 	for (ptr = nnew->if_name; *ptr && (*ptr < '0' || *ptr > '9'); 
 	     ptr++) ;
-	nnew->if_unit = strdup(*ptr ? ptr : "0");
+	nnew->if_unit = strdup(*ptr ? ptr : "");
 	*ptr = 0;
 
 	strcpy (ifrq.ifr_name, ifname);
@@ -1409,13 +1359,47 @@ Interface_Scan_Init (void)
 	strcpy (ifrq.ifr_name, ifname);
 	nnew->if_flags = ioctl (fd, SIOCGIFFLAGS, &ifrq) < 0 
 	  		? 0 : ifrq.ifr_flags;
-	
+
+	nnew->if_type = 0;
 	strcpy (ifrq.ifr_name, ifname);
 	if (ioctl(fd, SIOCGIFHWADDR, &ifrq) < 0)
 	  memset (nnew->if_hwaddr,(0), 6);
-	else
+	else {
 	  memcpy (nnew->if_hwaddr, ifrq.ifr_hwaddr.sa_data, 6);
-	    
+
+#ifdef ARPHRD_LOOPBACK
+	  switch (ifrq.ifr_hwaddr.sa_family) {
+	  case ARPHRD_ETHER:
+	      nnew->if_type = 6; break; 
+	  case ARPHRD_TUNNEL:
+	  case ARPHRD_TUNNEL6:
+#ifdef ARPHRD_IPGRE
+	  case ARPHRD_IPGRE:
+#endif
+	  case ARPHRD_SIT:
+	      nnew->if_type = 131; break; /* tunnel */
+	  case ARPHRD_SLIP:
+	  case ARPHRD_CSLIP:
+	  case ARPHRD_SLIP6:
+	  case ARPHRD_CSLIP6:
+	      nnew->if_type = 28; break; /* slip */
+	  case ARPHRD_PPP:
+	      nnew->if_type = 23; break; /* ppp */
+	  case ARPHRD_LOOPBACK:
+	      nnew->if_type = 24; break; /* softwareLoopback */
+	  case ARPHRD_FDDI:
+	      nnew->if_type = 15; break;
+	  case ARPHRD_ARCNET:
+	      nnew->if_type = 35; break;
+	  case ARPHRD_LOCALTLK:
+	      nnew->if_type = 42; break;
+	  case ARPHRD_HIPPI:
+	      nnew->if_type = 47; break;
+          /* XXX: more if_arp.h:ARPHDR_xxx to IANAifType mappings... */
+	  }
+#endif
+	}
+	
 	strcpy (ifrq.ifr_name, ifname);
 	nnew->if_metric = ioctl (fd, SIOCGIFMETRIC, &ifrq) < 0
 	  		? 0 : ifrq.ifr_metric;
@@ -1428,8 +1412,8 @@ Interface_Scan_Init (void)
 	nnew->if_mtu = 0;
 #endif
 
-	for (if_ptr = if_list; if_ptr; if_ptr = if_ptr->next)
-	    if (! strcmp (if_ptr->name, fullname))
+	for (if_ptr = conf_list; if_ptr; if_ptr = if_ptr->next)
+	    if (! strcmp (if_ptr->name, ifname))
 	      break;
 
 	if (if_ptr) {
@@ -1437,9 +1421,13 @@ Interface_Scan_Init (void)
 	    nnew->if_speed = if_ptr->speed;
 	}
 	else {
-	  nnew->if_type = if_type_from_name(nnew->if_name);
-	  nnew->if_speed = nnew->if_type == 6 ? 10000000 : 
-	    nnew->if_type == 24 ? 10000000 : 0;
+	    /* do only guess if_type from name, if we could not read
+	     * it before from SIOCGIFHWADDR */
+	    if (!nnew->if_type) 
+		nnew->if_type = if_type_from_name(nnew->if_name);
+		nnew->if_speed = nnew->if_type == 6 ? 10000000 : 
+		nnew->if_type == 24 ? 10000000 :
+		nnew->if_type ==  9 ?  4000000 : 0;
 	}
 
       } /* while (fgets ... */
@@ -1537,6 +1525,7 @@ int Interface_Scan_Next(short *Index,
 {
 	struct ifnet ifnet;
 	struct in_ifaddr *ia, in_ifaddr;
+	short has_ipaddr=0;
 #if !STRUCT_IFNET_HAS_IF_XNAME
 	register char *cp;
 #endif
@@ -1568,7 +1557,10 @@ int Interface_Scan_Next(short *Index,
 		auto_nlist(IFADDR_SYMBOL, (char *)&ia, sizeof(ia));
 		while (ia) {
 		    klookup((unsigned long)ia ,  (char *)&in_ifaddr, sizeof(in_ifaddr));
-		    if (in_ifaddr.ia_ifp == ifnetaddr) break;
+		    if (in_ifaddr.ia_ifp == ifnetaddr) {
+                        has_ipaddr = 1; /* this IF has IP-address */
+                        break;
+                    }
 		    ia = in_ifaddr.ia_next;
 		}
 
@@ -1583,6 +1575,9 @@ int Interface_Scan_Next(short *Index,
 		    *Retifnet = ifnet;
 		if (Retin_ifaddr)
 		    *Retin_ifaddr = in_ifaddr;
+		if (Retin_ifaddr && has_ipaddr) /* assign the in_ifaddr only
+                                                   if the IF has IP-address */
+		    *Retin_ifaddr = in_ifaddr;
 		if (Name)
 		    strcpy(Name, saveName);
 		saveifnet = ifnet;
@@ -1913,121 +1908,20 @@ get_phys_address(int iindex, char **ap, int *len)
 }
 
 static int
-header_interfaces(struct variable *vp,
-		  oid *name,
-		  size_t *length,
-		  int exact,
-		  size_t *var_len,
-		  WriteMethod **write_method)
+Interface_Scan_Get_Count (void)
 {
-#define INTERFACES_NAME_LENGTH	8
-    oid newname[MAX_OID_LEN];
-    int result;
-
-    DEBUGMSGTL(("mibII/interfaces", "var_interfaces: "));
-    DEBUGMSGOID(("mibII/interfaces", name, *length));
-    DEBUGMSG(("mibII/interfaces", " %d\n", exact));
-
-    memcpy(newname, vp->name, (int)vp->namelen * sizeof(oid));
-    newname[INTERFACES_NAME_LENGTH] = 0;
-    result = snmp_oid_compare(name, *length, newname, (int)vp->namelen + 1);
-    if ((exact && (result != 0)) || (!exact && (result >= 0)))
-        return MATCH_FAILED;
-    memcpy(name, newname, ((int)vp->namelen + 1) * sizeof(oid));
-    *length = vp->namelen + 1;
-
-    *write_method = 0;
-    *var_len = sizeof(long);	/* default to 'long' results */
-    return MATCH_SUCCEEDED;
-}
-
-static int count_oid[5] = { CTL_NET, PF_LINK, NETLINK_GENERIC, 
-			    IFMIB_SYSTEM, IFMIB_IFCOUNT };
-
-static int
-header_ifEntry(struct variable *vp,
-	       oid *name,
-	       size_t *length,
-	       int exact,
-	       size_t *var_len,
-	       WriteMethod **write_method)
-{
-#define IFENTRY_NAME_LENGTH	10
-    oid newname[MAX_OID_LEN];
-    register int	interface;
-    int result, count;
     static int count_oid[5] = { CTL_NET, PF_LINK, NETLINK_GENERIC, 
 				IFMIB_SYSTEM, IFMIB_IFCOUNT };
     size_t len;
-
-#ifdef DODEBUG
-    DEBUGMSGTL(("mibII/interfaces", "var_ifEntry: "));
-    DEBUGMSGOID(("mibII/interfaces", name, *length));
-    DEBUGMSG(("mibII/interfaces"," %d\n", exact));
-#endif
-
-    memcpy(newname, vp->name, (int)vp->namelen * sizeof(oid));
-    /* find "next" interface */
-    len = sizeof count;
-    if (sysctl(count_oid, 5, &count, &len, (void *)0, (size_t)0) < 0)
-	    return MATCH_FAILED;
-
-    for(interface = 1; interface <= count; interface++){
-	newname[IFENTRY_NAME_LENGTH] = (oid)interface;
-	result = snmp_oid_compare(name, *length, newname, (int)vp->namelen + 1);
-	if ((exact && (result == 0)) || (!exact && (result < 0)))
-	    break;
-    }
-    if (interface > count) {
-#ifdef DODEBUG
- DEBUGMSGTL(("interfaces", "... index out of range\n"));
-#endif
-        return MATCH_FAILED;
-    }
-
-
-    memcpy(name, newname, ((int)vp->namelen + 1) * sizeof(oid));
-    *length = vp->namelen + 1;
-    *write_method = 0;
-    *var_len = sizeof(long);	/* default to 'long' results */
-
-#ifdef DODEBUG
-    DEBUGMSGTL(("interfaces", "... get I/F stats "));
-    DEBUGMSGOID(("interfaces", name, *length));
-    DEBUGMSG(("interfaces","\n"));
-#endif
-
-    return interface;
-}
-
-u_char *
-var_interfaces(struct variable *vp,
-		oid *name,
-		size_t *length,
-		int exact,
-		size_t *var_len,
-		WriteMethod **write_method)
-{
-    size_t len;
     int count;
 
-    if (header_interfaces(vp, name, length, exact, var_len, write_method)
-	== MATCH_FAILED)
-	    return NULL;
-
-    switch (vp->magic){
-    case IFNUMBER:
-	    len = sizeof count;
-	    if (sysctl(count_oid, 5, &count, &len, 0, 0) < 0)
-		    return NULL;
-	    long_return = count;
-	    return (u_char *)&long_return;
-    default:
-	    DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_interfaces\n", vp->magic));
-    }
-    return NULL;
+    len = sizeof count;
+    if (sysctl(count_oid, 5, &count, &len, (void *)0, (size_t)0) < 0)
+	    return -1;
+    return count;
 }
 
+
 u_char *
 var_ifEntry(struct variable *vp,
 	    oid *name,
@@ -2085,19 +1979,15 @@ var_ifEntry(struct variable *vp,
 		long_return = ifmd.ifmd_flags & IFF_UP ? 1 : 2;
 		return (u_char *) &long_return;
 	case IFLASTCHANGE:
-		if ((ifmd.ifmd_data.ifi_lastchange.tv_sec == 0 ) &&
-		    (ifmd.ifmd_data.ifi_lastchange.tv_usec == 0)) {
+		if (ifmd.ifmd_data.ifi_lastchange.tv_sec == 0 &&
+		    ifmd.ifmd_data.ifi_lastchange.tv_usec == 0) {
 			long_return = 0;
+		else if (ifmd.ifmd_data.ifi_lastchange.tv_sec < starttime.tv_sec)
+		    long_return = 0;
 		} else {
-			struct timeval now;
-
-			gettimeofday(&now, (struct timezone *)0);
-			long_return = (u_long)
-				((now.tv_sec 
-				  - ifmd.ifmd_data.ifi_lastchange.tv_sec) * 100
-				 + ((now.tv_usec
-				     - ifmd.ifmd_data.ifi_lastchange.tv_usec)
-				    / 10000));
+		    long_return = (u_long)
+			((ifmd.ifmd_data.ifi_lastchange.tv_sec - starttime.tv_sec) * 100
+			 + ((ifmd.ifmd_data.ifi_lastchange.tv_usec - starttime.tv_usec) / 10000));
 		}
 		return (u_char *) &long_return;
 	case IFINOCTETS:
