@@ -4,6 +4,7 @@
  */
 
 #include <config.h>
+
 #if HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
@@ -15,6 +16,7 @@
 #else
 #include <strings.h>
 #endif
+#include <sys/types.h>
 #if HAVE_WINSOCK_H
 #include <winsock.h>
 #endif
@@ -40,9 +42,9 @@
 #endif
 
 #include "mibincl.h"
+#include "util_funcs.h"
 #include "system_mib.h"
-#include "../struct.h"
-#include "../util_funcs.h"
+#include "struct.h"
 #include "read_config.h"
 #include "agent_read_config.h"
 #include "system.h"
@@ -73,12 +75,15 @@ int sysServicesConfiged=0;
 extern oid version_id[];
 extern int version_id_len;
 
-extern struct timeval starttime;
 
 WriteMethod writeSystem;
 int header_system(struct variable *,oid *, size_t *, int, size_t *, WriteMethod **);
 
-/* snmpd.conf config parsing */
+	/*********************
+	 *
+	 *  snmpd.conf config parsing
+	 *
+	 *********************/
 
 void system_parse_config_sysloc(const char *token, 
 				char *cptr)
@@ -183,7 +188,7 @@ void init_system_mib(void)
 #endif /* HAVE_GETHOSTNAME */
 
   /* register ourselves with the agent to handle our mib tree */
-  REGISTER_MIB("mibII/system", system_variables, variable2, \
+  REGISTER_MIB("mibII/system", system_variables, variable2,
                system_variables_oid);
 
   if ( ++system_module_count == 3 )
@@ -200,51 +205,10 @@ void init_system_mib(void)
 
 }
 
-/*
-  header_system(...
-  Arguments:
-  vp	  IN      - pointer to variable entry that points here
-  name    IN/OUT  - IN/name requested, OUT/name found
-  length  IN/OUT  - length of IN/OUT oid's 
-  exact   IN      - TRUE if an exact match was requested
-  var_len OUT     - length of variable or 0 if function returned
-  write_method
-  
-*/
-
-int
-header_system(struct variable *vp,
-	      oid *name,
-	      size_t *length,
-	      int exact,
-	      size_t *var_len,
-	      WriteMethod **write_method)
-{
-#define SYSTEM_NAME_LENGTH	8
-    oid newname[MAX_OID_LEN];
-    int result;
-
-    DEBUGMSGTL(("mibII/system", "var_system: "));
-    DEBUGMSGOID(("mibII/system", name, *length));
-    DEBUGMSG(("mibII/system"," %d\n", exact));
-
-    memcpy((char *)newname, (char *)vp->name, vp->namelen * sizeof(oid));
-    newname[SYSTEM_NAME_LENGTH] = 0;
-    result = snmp_oid_compare(name, *length, newname, vp->namelen + 1);
-    if ((exact && (result != 0)) || (!exact && (result >= 0)))
-        return(MATCH_FAILED);
-    memcpy( (char *)name,(char *)newname, (vp->namelen + 1) * sizeof(oid));
-    *length = vp->namelen + 1;
-
-    *write_method = 0;
-    *var_len = sizeof(long);	/* default to 'long' results */
-    return(MATCH_SUCCEEDED);
-}
 
 	/*********************
 	 *
 	 *  System specific implementation functions
-	 *	(actually common!)
 	 *
 	 *********************/
 
@@ -261,9 +225,9 @@ var_system(struct variable *vp,
 	   WriteMethod **write_method)
 {
 
-  struct timeval now, diff;
+    struct timeval now;
 
-    if (header_system(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
+    if (header_generic(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
 	return NULL;
 
     switch (vp->magic){
@@ -276,15 +240,7 @@ var_system(struct variable *vp,
             return (u_char *)version_id;
         case UPTIME:
             gettimeofday(&now, NULL);
-            now.tv_sec--;
-            now.tv_usec += 1000000L;
-            diff.tv_sec = now.tv_sec - starttime.tv_sec;
-            diff.tv_usec = now.tv_usec - starttime.tv_usec;
-            if (diff.tv_usec > 1000000L){
-                diff.tv_usec -= 1000000L;
-                diff.tv_sec++;
-            }
-            long_return = ((diff.tv_sec * 100) + (diff.tv_usec / 10000));
+	    long_return = timeval_uptime( &now );
             return ((u_char *) &long_return);
         case SYSCONTACT:
             *var_len = strlen(sysContact);
@@ -308,17 +264,7 @@ var_system(struct variable *vp,
 
 #ifdef USING_MIBII_SYSORTABLE_MODULE
         case SYSORLASTCHANGE:
-              diff.tv_sec = sysOR_lastchange.tv_sec - 1 - starttime.tv_sec;
-              diff.tv_usec =
-                sysOR_lastchange.tv_usec + 1000000L - starttime.tv_usec;
-              if (diff.tv_usec > 1000000L){
-                diff.tv_usec -= 1000000L;
-                diff.tv_sec++;
-              }
-              if ((diff.tv_sec * 100) + (diff.tv_usec / 10000) < 0)
-                long_return = 0;
-              else
-                long_return = ((diff.tv_sec * 100) + (diff.tv_usec / 10000));
+	      long_return = timeval_uptime( &sysOR_lastchange );
               return ((u_char *) &long_return);
 #endif
               
