@@ -120,10 +120,10 @@ PERFORMANCE OF THIS SOFTWARE.
 #endif
 #if HAVE_KVM_OPENFILES
 #include <fcntl.h>
+#endif
 #if HAVE_KVM_H
 #include <kvm.h>
 #endif
-#endif
 
 #if HAVE_STRING_H
 #include <string.h>
@@ -165,18 +165,11 @@ PERFORMANCE OF THIS SOFTWARE.
 #include <dmalloc.h>
 #endif
 
-#include "asn1.h"
-#include "snmp_debug.h"
-#include "snmp_logging.h"
 
 #define CACHE_TIME (120)	    /* Seconds */
 
-#include "asn1.h"
-#include "snmp_api.h"
-#include "snmp_impl.h"
-#include "mib.h"
-#include "snmp.h"
-#include "../snmp_vars.h"
+#include "mibincl.h"
+
 #include "ip.h"
 #include "../kernel.h"
 #include "interfaces.h"
@@ -1315,19 +1308,12 @@ static int qsort_compare(RTENTRY **r1,
 
 #define CACHE_TIME (120)	    /* Seconds */
 
-#include "asn1.h"
-#include "snmp_api.h"
-#include "snmp_impl.h"
-#include "mib.h"
-#include "snmp.h"
-#include "../snmp_vars.h"
+#include "mibincl.h"
 #include "ip.h"
 #include "../kernel.h"
 #include "interfaces.h"
 #include "struct.h"
 #include "util_funcs.h"
-#include "snmp_logging.h"
-#include "snmp_debug.h"
 
 static TAILQ_HEAD(, snmprt) rthead;
 static char *rtbuf;
