@@ -58,6 +58,10 @@
 #include <netinet/in.h>
 #endif
 
+#if HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif
+
 #if HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
 #endif
@@ -210,6 +214,19 @@ init_smux(void)
 		return;
 	}
 
+#ifdef SO_REUSEADDR 
+       /* At least on Linux, when the master agent terminates, any
+          TCP connections for SMUX peers are put in the TIME_WAIT
+          state for about 60 seconds. If the master agent is started
+          during this time, the bind for the listening socket will
+          fail because the SMUX port is in use.
+       */
+       if (setsockopt (smux_listen_sd, SOL_SOCKET, SO_REUSEADDR, (char *)&one, 
+                       sizeof (one)) < 0) { 
+               snmp_log_perror("[init_smux] setsockopt(SO_REUSEADDR) failed"); 
+       } 
+#endif /* SO_REUSEADDR */ 
+
 	if (bind (smux_listen_sd, (struct sockaddr *) &lo_socket, 
 	    sizeof (lo_socket)) < 0) {
 		snmp_log_perror("[init_smux] bind failed");
@@ -253,7 +270,7 @@ var_smux(struct variable *vp,
 
 	/* search the active registration list */
 	for (rptr = ActiveRegs; rptr; rptr = rptr->sr_next) {
-		if (!compare_tree(name, *length, rptr->sr_name,
+		if (!compare_tree(vp->name, vp->namelen, rptr->sr_name,
 		    rptr->sr_name_len))
 			break;
 	}
@@ -295,7 +312,7 @@ var_smux_write(
 	smux_reg *rptr;
 	u_char buf[SMUXMAXPKTSIZE], *ptr, sout[3], type;
 	int reterr;
-	size_t len, var_len, datalen, name_length;
+	size_t len, var_len, datalen, name_length, packet_len;
 	long reqid, errsts, erridx;
 	u_char var_type, *dataptr;
 
@@ -360,37 +377,80 @@ var_smux_write(
 			return SNMP_ERR_GENERR;
 		}
 
-		if ((len = recv(rptr->sr_fd, buf,
-		    SMUXMAXPKTSIZE, 0)) <= 0) {
-			DEBUGMSGTL (("smux","[var_smux_write] recv failed or timed out\n"));
-			/* do we need to do a peer cleanup in this case?? */
-			smux_peer_cleanup(rptr->sr_fd);
-			return SNMP_ERR_GENERR;
-		}
+                while (1) { 
+                    /* 
+                     * peek at what's received 
+                     */ 
+                    if ((len = recv(rptr->sr_fd, buf, 
+                                    SMUXMAXPKTSIZE, MSG_PEEK)) <= 0) { 
+                        DEBUGMSGTL (("smux","[var_smux_write] peek failed or timed out\n")); 
+                        /* do we need to do a peer cleanup in this case?? */ 
+                        smux_peer_cleanup(rptr->sr_fd); 
+                        return SNMP_ERR_GENERR; 
+                    } 
 
-		ptr = buf;
-		ptr = asn_parse_header(ptr, &len, &type);
-		if ((ptr == NULL) || type != SNMP_MSG_RESPONSE) 
-			return SNMP_ERR_GENERR;
+                    DEBUGMSGTL (("smux","[var_smux_write] Peeked at %d bytes\n", len)); 
+                    DEBUGDUMPSETUP("var_smux_write", buf, len); 
 
-		ptr = asn_parse_int(ptr, &len, &type, &reqid, sizeof(reqid));
-		if ((ptr == NULL) || type != ASN_INTEGER) 
-			return SNMP_ERR_GENERR;
+                    /* 
+                     * determine if we received more than one packet 
+                     */ 
+                    packet_len = len; 
+                    ptr = asn_parse_header(buf, &packet_len, &type); 
+                    packet_len += (ptr - buf); 
+                    if (len > packet_len) { 
+                        /* set length to receive only the first packet */
+                        len = packet_len; 
+                    } 
 
-		ptr = asn_parse_int(ptr, &len, &type, &errsts, sizeof(errsts));
-		if ((ptr == NULL) || type != ASN_INTEGER) 
-			return SNMP_ERR_GENERR;
+                    /* 
+                     * receive the first packet 
+                     */ 
+                    if ((len = recv(rptr->sr_fd, buf, len, 0)) <= 0) { 
+                        DEBUGMSGTL (("smux","[var_smux_write] recv failed or timed out\n")); 
+                        /* do we need to do a peer cleanup in this case?? */ 
+                        smux_peer_cleanup(rptr->sr_fd); 
+                        return SNMP_ERR_GENERR; 
+                    } 
 
-		if (errsts) {
-			DEBUGMSGTL (("smux","[var_smux_write] errsts returned\n"));
-			return (errsts);
-		}
+                    DEBUGMSGTL (("smux","[var_smux_write] Received %d bytes\n", len)); 
 
-		ptr = asn_parse_int(ptr, &len, &type, &erridx, sizeof(erridx));
-		if ((ptr == NULL) || type != ASN_INTEGER) 
-			return SNMP_ERR_GENERR;
+                    if (buf[0] == SMUX_TRAP) { 
+                        DEBUGMSGTL (("smux","[var_smux_write] Received trap\n")); 
+                        snmp_log(LOG_INFO, "Got trap from peer on fd %d\n", rptr->sr_fd); 
+                        ptr = asn_parse_header(buf, &len, &type); 
+                        smux_trap_process(ptr, &len); 
 
-		reterr = SNMP_ERR_NOERROR;
+
+                        /* go and peek at received data again */
+                        /* we could receive the reply or another trap */
+                    } else { 
+                        ptr = buf; 
+                        ptr = asn_parse_header(ptr, &len, &type); 
+                        if ((ptr == NULL) || type != SNMP_MSG_RESPONSE) 
+                            return SNMP_ERR_GENERR; 
+
+                        ptr = asn_parse_int(ptr, &len, &type, &reqid, sizeof(reqid)); 
+                        if ((ptr == NULL) || type != ASN_INTEGER) 
+                            return SNMP_ERR_GENERR; 
+
+                        ptr = asn_parse_int(ptr, &len, &type, &errsts, sizeof(errsts)); 
+                        if ((ptr == NULL) || type != ASN_INTEGER) 
+                            return SNMP_ERR_GENERR; 
+
+                        if (errsts) { 
+                            DEBUGMSGTL (("smux","[var_smux_write] errsts returned\n")); 
+                            return (errsts); 
+                        } 
+
+                        ptr = asn_parse_int(ptr, &len, &type, &erridx, sizeof(erridx)); 
+                        if ((ptr == NULL) || type != ASN_INTEGER) 
+                            return SNMP_ERR_GENERR; 
+
+                        reterr = SNMP_ERR_NOERROR; 
+                        break; 
+                    } 
+                } /* while (1) */ 
 		break;	/* case Action == RESERVE1 */
 
 	case RESERVE2:
@@ -448,9 +508,10 @@ smux_accept(int sd)
 		snmp_log_perror("[smux_accept] accept failed");
 		return -1;
 	} else {
-		snmp_log(LOG_ERR, "[smux_accept] accepted fd %d - errno %d\n", fd, errno);
+		snmp_log(LOG_INFO, "[smux_accept] accepted fd %d from %s:%d\n",
+		    fd, inet_ntoa(in_socket.sin_addr), ntohs(in_socket.sin_port));
 		if (npeers + 1 == SMUXMAXPEERS) {
-			DEBUGMSGTL (("smux","[smux_accept] denied peer on fd %d, limit reached", fd));
+			snmp_log (LOG_ERR, "[smux_accept] denied peer on fd %d, limit %d reached", fd, SMUXMAXPEERS);
 			close(sd);
 			return -1;
 		}
@@ -487,7 +548,7 @@ smux_accept(int sd)
 #ifdef SO_RCVTIMEO
 		if (setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, (void *)&tv, sizeof(tv)) < 0) {
 			DEBUGMSGTL (("smux","[smux_accept] setsockopt(SO_RCVTIMEO) failed fd %d\n", fd));
-                        snmp_log_perror("smux/setsockopt");
+                        snmp_log_perror("smux_accept: setsockopt SO_RCVTIMEO");
                 }
 #endif
 		npeers++;
@@ -567,8 +628,8 @@ smux_pdu_process(int fd, u_char *data, size_t length)
 		case SMUX_TRAP:
 			snmp_log(LOG_INFO, "Got trap from peer on fd %d\n", fd);
 			ptr = smux_trap_process(ptr, &len);
-			// watch out for close on top of this...should return correct end
-			// debug this...
+			/* watch out for close on top of this...should return correct end */
+			/* debug this... */
 			ptr = NULL;
 			break;
 		default:
@@ -588,12 +649,13 @@ smux_open_process(int fd, u_char *ptr, size_t *len, int *fail)
 	u_char type;
 	long version;
 	oid oid_name[MAX_OID_LEN];
-	char string[SMUXMAXSTRLEN];
+	char passwd[SMUXMAXSTRLEN];
+	char descr[SMUXMAXSTRLEN];
+	char oid_print[SMUXMAXSTRLEN];
 	int i;
 	size_t oid_name_len, string_len;
 
-	if ((ptr = asn_parse_int(ptr, len, &type, &version, 
-	    sizeof(version))) == NULL) {
+	if (!(ptr = asn_parse_int(ptr, len, &type, &version, sizeof(version)))) {
 		DEBUGMSGTL (("smux","[smux_open_process] version parse failed\n"));
 		*fail = TRUE;
 		return((ptr += *len));
@@ -609,17 +671,15 @@ smux_open_process(int fd, u_char *ptr, size_t *len, int *fail)
 		*fail = TRUE;
 		return((ptr += *len));
 	}
+	sprint_objid(oid_print, oid_name, oid_name_len);
 
         if (snmp_get_do_debugging()) {
-          DEBUGMSGTL (("smux","[smux_open_process] smux peer:")); 
-          for (i=0; i < (int)oid_name_len; i++) 
-            DEBUGMSG (("smux",".%d", oid_name[i]));
-          DEBUGMSG (("smux"," \n"));
-          DEBUGMSGTL (("smux","[smux_open_process] len %d, type %d\n", *len, (int)type));
+	    DEBUGMSGTL (("smux","[smux_open_process] smux peer: %s\n", oid_print)); 
+	    DEBUGMSGTL (("smux","[smux_open_process] len %d, type %d\n", *len, (int)type));
         }
 
 	string_len = SMUXMAXSTRLEN;
-	if ((ptr = asn_parse_string(ptr, len, &type, (u_char *)string,
+	if ((ptr = asn_parse_string(ptr, len, &type, (u_char *)descr,
 	    &string_len)) == NULL) {
 		DEBUGMSGTL (("smux","[smux_open_process] descr parse failed\n"));
 		*fail = TRUE;
@@ -627,15 +687,16 @@ smux_open_process(int fd, u_char *ptr, size_t *len, int *fail)
 	}
 
         if (snmp_get_do_debugging()) {
-          DEBUGMSGTL (("smux","[smux_open_process] smux peer descr:")); 
+          DEBUGMSGTL (("smux","[smux_open_process] smux peer descr: ")); 
           for (i=0; i < (int)string_len; i++) 
-            DEBUGMSG (("smux","%c", string[i]));
-          DEBUGMSG (("smux"," \n"));
+            DEBUGMSG (("smux","%c", descr[i]));
+          DEBUGMSG (("smux","\n"));
           DEBUGMSGTL (("smux","[smux_open_process] len %d, type %d\n", *len, (int)type));
         }
+	descr[string_len] = 0;
 
 	string_len = SMUXMAXSTRLEN;
-	if ((ptr = asn_parse_string(ptr, len, &type, (u_char *)string,
+	if ((ptr = asn_parse_string(ptr, len, &type, (u_char *)passwd,
 	    &string_len)) == NULL) {
 		DEBUGMSGTL (("smux","[smux_open_process] passwd parse failed\n"));
 		*fail = TRUE;
@@ -643,23 +704,21 @@ smux_open_process(int fd, u_char *ptr, size_t *len, int *fail)
 	}
 
         if (snmp_get_do_debugging()) {
-          DEBUGMSGTL (("smux","[smux_open_process] smux peer passwd:")); 
+          DEBUGMSGTL (("smux","[smux_open_process] smux peer passwd: ")); 
           for (i=0; i < (int)string_len; i++) 
-            DEBUGMSG (("smux","%c", string[i]));
-          DEBUGMSG (("smux"," \n"));
+            DEBUGMSG (("smux","%c", passwd[i]));
+          DEBUGMSG (("smux","\n"));
           DEBUGMSGTL (("smux","[smux_open_process] len %d, type %d\n", *len, (int)type));
         }
-	string[string_len] = '\0';
-	if(!smux_auth_peer(oid_name, oid_name_len, string, fd)) {
-		if(snmp_get_do_debugging()) {
-		    DEBUGMSGTL (("smux","[smux_open_process] peer authentication failed for oid\n"));
-		    for (i = 0; i < (int)oid_name_len; i++) 
-			DEBUGMSG (("smux","\t.%d", oid_name[i]));
-		    DEBUGMSG (("smux"," password %s\n", string));
-		}
-		*fail = TRUE;
-		return ptr;
+	passwd[string_len] = '\0';
+	if(!smux_auth_peer(oid_name, oid_name_len, passwd, fd)) {
+	    snmp_log(LOG_WARNING, "refused smux peer: oid %s, password %s, descr %s\n",
+		     oid_print, passwd, descr);
+	    *fail = TRUE;
+	    return ptr;
 	}
+	snmp_log(LOG_INFO, "accepted smux peer: oid %s, password %s, descr %s\n",
+		 oid_print, passwd, descr);
 	*fail = FALSE;
 	return ptr;
 }
@@ -1043,6 +1102,7 @@ smux_snmp_process(int exact,
 	u_char packet[SMUXMAXPKTSIZE], *ptr, result[SMUXMAXPKTSIZE];
 	size_t length = SMUXMAXPKTSIZE;
 	u_char type;
+        size_t packet_len; 
 	
 	/* 
 	 * Send the query to the peer
@@ -1056,7 +1116,7 @@ smux_snmp_process(int exact,
 
 	if (smux_build(type, smux_reqid, objid, len, 0, NULL, 
 	    *len, packet, &length) < 0) {
-	 snmp_log(LOG_NOTICE, "[smux_snmp_process]: smux_build failed\n");
+	 snmp_log(LOG_ERR, "[smux_snmp_process]: smux_build failed\n");
 		return NULL;
 	}
     DEBUGMSGTL(("smux", "[smux_snmp_process] oid from build: "));
@@ -1070,24 +1130,60 @@ smux_snmp_process(int exact,
 	DEBUGMSGTL(("smux",
                     "[smux_snmp_process] Sent %d request to peer; %d bytes\n", 
                     (int)type, length));
-	/* 
-	 * receive 
-	 * XXX the RCVTIMEO could return a short result.
-	 */
-	length = recv(sd, (char *)result, SMUXMAXPKTSIZE, 0);
-	if (length < 0) {
-		snmp_log_perror("[smux_snmp_process] recv failed");
-		smux_peer_cleanup(sd);
-		return NULL;
-	}
 
-	DEBUGMSGTL (("smux","[smux_snmp_process] Recived %d bytes from gated\n", length));
+        while (1) { 
+            /* 
+             * peek at what's received 
+             */ 
+            length = recv(sd, (char *)result, SMUXMAXPKTSIZE, MSG_PEEK); 
+            if (length < 0) { 
+                snmp_log_perror("[smux_snmp_process] peek failed"); 
+                smux_peer_cleanup(sd); 
+                return NULL; 
+            } 
 
-	/* Interpret reply */
-	if ((ptr = smux_parse(result, objid, len, return_len, return_type)) == NULL) {
-		/* either error or parse failed, just return NULL */
-		return NULL;
-	}
+            DEBUGMSGTL (("smux","[smux_snmp_process] Peeked at %d bytes\n",
+                         length)); 
+            DEBUGDUMPSETUP("smux_snmp_process", result, length); 
+
+            /* 
+             * determine if we received more than one packet 
+             */ 
+            packet_len = length; 
+            ptr = asn_parse_header(result, &packet_len, &type); 
+            packet_len += (ptr - result); 
+            if (length > packet_len) { 
+                /* set length to receive only the first packet */
+                length = packet_len; 
+            } 
+
+            /* 
+             * receive the first packet 
+             */
+            length = recv(sd, (char *)result, length, 0); 
+            if (length < 0) { 
+                snmp_log_perror("[smux_snmp_process] recv failed"); 
+                smux_peer_cleanup(sd); 
+                return NULL; 
+            } 
+
+            DEBUGMSGTL (("smux","[smux_snmp_process] Received %d bytes\n", length)); 
+
+            if (result[0] == SMUX_TRAP) { 
+                DEBUGMSGTL (("smux","[smux_snmp_process] Received trap\n")); 
+                snmp_log(LOG_INFO, "Got trap from peer on fd %d\n", sd); 
+                ptr = asn_parse_header(result, &length, &type); 
+                smux_trap_process(ptr, &length); 
+
+                /* go and peek at received data again */
+                /* we could receive the reply or another trap */
+            } else { 
+                /* Interpret reply */ 
+                ptr = smux_parse(result, objid, len, return_len, return_type); 
+                /* ptr will point to query result or NULL if error */ 
+                break; 
+            } 
+        } /* while (1) */
 
 	return ptr;
 }
@@ -1240,7 +1336,7 @@ smux_parse_var(u_char *varbind,
 		objid_len = MAX_OID_LEN;
 		asn_parse_objid(var_val, &len, vartype, 
 				smux_objid, &objid_len);
-		*varlength = objid_len;
+		*varlength = objid_len*sizeof(oid);
 		return (u_char *)smux_objid;
 		break;
             case SNMP_NOSUCHOBJECT:
@@ -1373,7 +1469,10 @@ smux_peer_cleanup(int sd)
 	/* make his auth available again */
 	for (i = 0; i < nauths; i++) {
 		if (Auths[i]->sa_active_fd == sd) {
+		  	char oid_name[128];
 			Auths[i]->sa_active_fd = -1;
+			sprint_objid(oid_name, Auths[i]->sa_oid, Auths[i]->sa_oid_len);
+			snmp_log(LOG_INFO, "peer disconnected: %s\n", oid_name);
 		}
 	}
 }
@@ -1527,6 +1626,7 @@ smux_trap_process(u_char *rsp, size_t *len)
 				      var_val_len);
 				var_val = (u_char *)&(smux_sa.sin_addr.s_addr);
 				break;
+		    case ASN_OPAQUE:
 		    case ASN_OCTET_STR:
 				/* XXX */
 				if (len == 0)
@@ -1536,12 +1636,11 @@ smux_trap_process(u_char *rsp, size_t *len)
 						 smux_str, &var_val_len);
 				var_val = smux_str;
 				break;
-		    case ASN_OPAQUE:
-		    case ASN_NSAP:
 		    case ASN_OBJECT_ID:
 				var_val_len = MAX_OID_LEN;
 				asn_parse_objid(var_val, &maxlen, &vartype, 
 						smux_objid, &var_val_len);
+                                var_val_len *= sizeof(oid);
 				var_val = (u_char *)smux_objid;
 				break;
 		    case SNMP_NOSUCHOBJECT:
@@ -1559,6 +1658,7 @@ smux_trap_process(u_char *rsp, size_t *len)
 						 smux_str, &var_val_len);
 				var_val = (u_char *)smux_str;
 				break;
+		    case ASN_NSAP:
 		    default:
 				 snmp_log(LOG_ERR, "bad type returned (%x)\n", vartype);
 				var_val = NULL;
