@@ -344,8 +344,8 @@ int getswap(int rettype)
 #endif
   if ((fd = get_exec_output(&ex))) {
     file = fdopen(fd,"r");
-    for (i=1;i <= 2 && fgets(ex.output,STRMAX,file) != NULL; i++);
-    if (fgets(ex.output,STRMAX,file) != NULL) {
+    for (i=1;i <= 2 && fgets(ex.output,sizeof(ex.output),file) != NULL; i++);
+    if (fgets(ex.output,sizeof(ex.output),file) != NULL) {
       cp = skip_white(ex.output);  /* not there should be any */
       cp = skip_not_white(cp);     /* skip over "reserve" */
       cp = skip_white(cp);
@@ -443,7 +443,7 @@ unsigned char *var_extensible_mem(struct variable *vp,
       return((u_char *) (&long_ret));
     case MEMAVAILREAL:
 #ifdef linux
-	long_ret = memory(meminfo_used);
+	long_ret = memory(meminfo_free);
 #else
       long_ret = pagetok((int) total.t_arm);
 #endif
@@ -472,7 +472,7 @@ unsigned char *var_extensible_mem(struct variable *vp,
 #endif
     case MEMTOTALFREE:
 #ifdef linux
-	long_ret = memory(meminfo_free);
+	long_ret = memory(meminfo_free) + memswap(meminfo_free);
 #else
       long_ret = pagetok(total.t_free);
 #endif
