@@ -46,6 +46,46 @@ struct variable2 extensible_passthru_variables[] = {
   {MIBINDEX, ASN_INTEGER, RWRITE, var_extensible_pass, 0, {MIBINDEX}},
 };
 
+
+
+/* lexicographical compare two object identifiers.
+ * Returns -1 if name1 < name2,
+ *          0 if name1 = name2,
+ *          1 if name1 > name2
+ *
+ * This method differs from snmp_oid_compare
+ * in that the comparison stops at the length
+ * of the smallest object identifier.
+ */
+int
+snmp_oid_min_compare(const oid *in_name1, 
+		 size_t len1,
+		 const oid *in_name2, 
+		 size_t len2)
+{
+    register int len, res;
+    register const oid * name1 = in_name1;
+    register const oid * name2 = in_name2;
+
+    /* len = minimum of len1 and len2 */
+    if (len1 < len2)
+	len = len1;
+    else
+	len = len2;
+    /* find first non-matching OID */
+    while(len-- > 0){
+	res = *(name1++) - *(name2++);
+	if (res < 0)
+	    return -1;
+	if (res > 0)
+	    return 1;
+    }
+    /* both OIDs equal up to length of shorter OID */
+
+    return 0;
+}
+
+
 /*  This is also called from pass_persist.c */
 int asc2bin(char *p)
 {
@@ -179,11 +219,11 @@ u_char *var_extensible_pass(struct variable *vp,
 				   size_t *var_len,
 				   WriteMethod **write_method)
 {
-
   oid newname[MAX_OID_LEN];
-  int i, j, rtest=0, fd, newlen, last;
+  int i, rtest, fd, newlen;
   static long long_ret;
-  static char buf[SNMP_MAXBUF], buf2[SNMP_MAXBUF];
+  char buf[SNMP_MAXBUF];
+  static char buf2[SNMP_MAXBUF];
   static oid  objid[MAX_OID_LEN];
   struct extensible *passthru;
   FILE *file;
@@ -191,17 +231,8 @@ u_char *var_extensible_pass(struct variable *vp,
   long_ret = *length;
   for(i=1; i<= numpassthrus; i++) {
     passthru = get_exten_instance(passthrus,i);
-    last = passthru->miblen;
-    if (passthru->miblen > *length)
-      last = *length;
-    for(j=0,rtest=0; j < last && !rtest; j++) {
-      if (name[j] != passthru->miboid[j]) {
-        if (name[j] < passthru->miboid[j])
-          rtest = -1;
-        else
-          rtest = 1;
-      }
-    }
+    rtest = snmp_oid_min_compare(name, *length,
+                passthru->miboid, passthru->miblen);
     if ((exact && rtest == 0) || (!exact && rtest <= 0)) {
       /* setup args */
       if (passthru->miblen >= *length || rtest < 0)
@@ -247,6 +278,10 @@ u_char *var_extensible_pass(struct variable *vp,
           *var_len = strlen(buf2);
           vp->type = ASN_OCTET_STR;
           return((unsigned char *) buf2);
+        } else if (!strncasecmp(buf,"opaque",6)) {
+          *var_len = asc2bin(buf2);
+          vp->type = ASN_OPAQUE;
+          return((unsigned char *) buf2);
         } else if (!strncasecmp(buf,"integer",7)) {
           *var_len = sizeof(long_ret);
           long_ret = strtol(buf2, NULL, 10);
@@ -315,28 +350,18 @@ setPass(int action,
 	oid *name,
 	size_t name_len)
 {
-  int i, j, rtest, last;
+  int i, rtest;
   struct extensible *passthru;
 
-  static char buf[SNMP_MAXBUF], buf2[SNMP_MAXBUF];
-  static long tmp;
-  static unsigned long utmp;
+  char buf[SNMP_MAXBUF], buf2[SNMP_MAXBUF];
+  long tmp;
+  unsigned long utmp;
   size_t itmp;
-  static oid objid[MAX_OID_LEN];
   
   for(i=1; i<= numpassthrus; i++) {
     passthru = get_exten_instance(passthrus,i);
-    last = passthru->miblen;
-    if (passthru->miblen > name_len)
-      last = name_len;
-    for(j=0,rtest=0; j < last && !rtest; j++) {
-      if (name[j] != passthru->miboid[j]) {
-        if (name[j] < passthru->miboid[j])
-          rtest = -1;
-        else
-          rtest = 1;
-      }
-    }
+    rtest = snmp_oid_min_compare(name, name_len,
+                passthru->miboid, passthru->miblen);
     if (rtest <= 0) {
       if (action != COMMIT)
         return SNMP_ERR_NOERROR;
@@ -379,15 +404,13 @@ setPass(int action,
         case ASN_OCTET_STR:
           itmp = sizeof(buf2);
           memcpy(buf2, var_val, var_val_len);
-          if (bin2asc(buf2, var_val_len) == (int)itmp)
+          if (bin2asc(buf2, var_val_len) == (int)var_val_len)
               sprintf(buf,"string %s",buf2);
           else
               sprintf(buf,"octet %s",buf2);
           break;
         case ASN_OBJECT_ID:
-          itmp = var_val_len/sizeof(oid);
-          memcpy(objid, var_val, var_val_len);
-          sprint_mib_oid(buf2, objid, itmp);
+          sprint_mib_oid(buf2, (oid *)var_val, var_val_len);
           sprintf(buf,"objectid \"%s\"",buf2);
           break;
       }
