@@ -55,6 +55,7 @@ static int write_persist_pipe (int iindex, const char *data);
 /*  These are defined in pass.c */
 extern int asc2bin(char *p);
 extern int bin2asc(char *p, size_t n);
+extern int snmp_oid_min_compare(const oid *, size_t , const oid *, size_t );
 
 /* the relocatable extensible commands variables */
 struct variable2 extensible_persist_passthru_variables[] = {
@@ -156,11 +157,11 @@ u_char *var_extensible_pass_persist(struct variable *vp,
 					   size_t *var_len,
 					   WriteMethod **write_method)
 {
-
   oid newname[MAX_OID_LEN];
-  int i, j, rtest=0, newlen, last;
+  int i, rtest, newlen;
   static long long_ret;
-  static char buf[STRMAX], buf2[STRMAX];
+  char buf[SNMP_MAXBUF];
+  static char buf2[SNMP_MAXBUF];
   static oid  objid[MAX_OID_LEN];
   struct extensible *persistpassthru;
   FILE *file;
@@ -171,17 +172,8 @@ u_char *var_extensible_pass_persist(struct variable *vp,
   long_ret = *length;
   for(i=1; i<= numpersistpassthrus; i++) {
     persistpassthru = get_exten_instance(persistpassthrus,i);
-    last = persistpassthru->miblen;
-    if (persistpassthru->miblen > *length)
-      last = *length;
-    for(j=0,rtest=0; j < last && !rtest; j++) {
-      if (name[j] != persistpassthru->miboid[j]) {
-        if (name[j] < persistpassthru->miboid[j])
-          rtest = -1;
-        else
-          rtest = 1;
-      }
-    }
+    rtest = snmp_oid_min_compare(name, *length,
+                persistpassthru->miboid, persistpassthru->miblen);
     if ((exact && rtest == 0) || (!exact && rtest <= 0)) {
       /* setup args */
       if (persistpassthru->miblen >= *length || rtest < 0)
@@ -259,6 +251,10 @@ u_char *var_extensible_pass_persist(struct variable *vp,
           *var_len = asc2bin(buf2);
           vp->type = ASN_OCTET_STR;
           return((unsigned char *) buf2);
+        } else if (!strncasecmp(buf,"opaque",5)) {
+          *var_len = asc2bin(buf2);
+          vp->type = ASN_OPAQUE;
+          return((unsigned char *) buf2);
         } else if (!strncasecmp(buf,"gauge",5)) {
           *var_len = sizeof(long_ret);
           long_ret = strtoul(buf2, NULL, 10);
@@ -308,31 +304,21 @@ setPassPersist(int action,
 	       oid *name,
 	       size_t name_len)
 {
-  int i, j, rtest, tmplen=1000, last;
+  int i, rtest;
   struct extensible *persistpassthru;
 
-  static char buf[STRMAX], buf2[STRMAX];
-  static long tmp;
-  static unsigned long utmp;
-  static int itmp;
-  static oid objid[MAX_OID_LEN];
+  char buf[SNMP_MAXBUF], buf2[SNMP_MAXBUF];
+  long tmp;
+  unsigned long utmp;
+  int itmp;
 
   /* Make sure that our basic pipe structure is malloced */
   init_persist_pipes();
 
   for(i=1; i<= numpersistpassthrus; i++) {
     persistpassthru = get_exten_instance(persistpassthrus,i);
-    last = persistpassthru->miblen;
-    if (persistpassthru->miblen > name_len)
-      last = name_len;
-    for(j=0,rtest=0; j < last && !rtest; j++) {
-      if (name[j] != persistpassthru->miboid[j]) {
-        if (name[j] < persistpassthru->miboid[j])
-          rtest = -1;
-        else
-          rtest = 1;
-      }
-    }
+    rtest = snmp_oid_min_compare(name, name_len,
+                persistpassthru->miboid, persistpassthru->miblen);
     if (rtest <= 0) {
       if (action != COMMIT)
         return SNMP_ERR_NOERROR;
@@ -375,15 +361,13 @@ setPassPersist(int action,
         case ASN_OCTET_STR:
 	  itmp = sizeof(buf2);
           memcpy(buf2, var_val, var_val_len);
-          if (bin2asc(buf2, var_val_len) == (int)itmp)
+          if (bin2asc(buf2, var_val_len) == (int)var_val_len)
               sprintf(buf,"string %s",buf2);
           else
               sprintf(buf,"octet %s",buf2);
           break;
         case ASN_OBJECT_ID:
-          itmp = var_val_len/sizeof(oid);
-          memcpy(objid, var_val, var_val_len);
-          sprint_mib_oid(buf2, objid, itmp);
+          sprint_mib_oid(buf2, (oid *)var_val, var_val_len);
           sprintf(buf,"objectid \"%s\"",buf2);
           break;
       }
@@ -508,7 +492,7 @@ static int open_persist_pipe(int iindex, char *command)
 
   /* Send test packet always so we can self-catch */
   {
-    char buf[STRMAX];
+    char buf[SNMP_MAXBUF];
     /* Should catch SIGPIPE around this call! */
     if( ! write_persist_pipe( iindex, "PING\n" ) ) {
       DEBUGMSGTL(("ucd-snmp/pass_persist", "open_persist_pipe: Error writing PING\n"));
