@@ -66,28 +66,24 @@ SOFTWARE.
 #include <dmalloc.h>
 #endif
 
-#include "asn1.h"
 #define SNMP_NEED_REQUEST_LIST
-#include "snmp_api.h"
-#include "snmp_impl.h"
-#include "snmp.h"
-#include "mib.h"
+#include "mibincl.h"
 #include "snmp_client.h"
+#include "snmp_alarm.h"
 
-#include "snmp_vars.h"
 #include "snmpd.h"
 #include "mibgroup/struct.h"
 #include "mibgroup/util_funcs.h"
-#include "var_struct.h"
-#include "read_config.h"
-#include "snmp_logging.h"
-#include "snmp_debug.h"
 #include "mib_module_config.h"
 
 #include "default_store.h"
+#include "system.h"
 #include "ds_agent.h"
 #include "snmp_agent.h"
-#include "agent_trap.h"
+
+#ifdef USING_AGENTX_PROTOCOL_MODULE
+#include "agentx/protocol.h"
+#endif
 
 static int snmp_vars_inc;
 
@@ -95,8 +91,6 @@ static struct agent_snmp_session *agent_session_list = NULL;
 
 
 static void dump_var(oid *, size_t, int, void *, size_t);
-static int goodValue(u_char, size_t, u_char, size_t);
-static void setVariable(u_char *, u_char, size_t, u_char *, size_t);
 
 static void dump_var (
     oid *var_name,
@@ -136,7 +130,7 @@ agent_check_and_process(int block) {
   numfds = 0;
   FD_ZERO(&fdset);
   snmp_select_info(&numfds, &fdset, tvp, &fakeblock);
-  if (block == 1 && fakeblock == 1)
+  if (block == 1)
     tvp = NULL; /* block without timeout */
   else if (block == 0) {
       tvp->tv_sec = 0;
@@ -163,6 +157,10 @@ agent_check_and_process(int block) {
       snmp_log(LOG_ERR, "select returned %d\n", count);
       return -1;
   }  /* endif -- count>0 */
+
+  /* run requested alarms */
+  run_alarms();
+
   return count;
 }
 
@@ -184,36 +182,99 @@ agent_check_and_process(int block) {
 		for Index Allocation use initially. */
 struct snmp_session *main_session;
 
-void
+int
 init_master_agent(int dest_port, 
                   int (*pre_parse) (struct snmp_session *, snmp_ipaddr),
                   int (*post_parse) (struct snmp_session *, struct snmp_pdu *,int))
 {
     struct snmp_session sess, *session;
+    char *cptr, *cptr2;
+    char buf[SPRINT_MAX_LEN];
+    int flags;
 
     if ( ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_ROLE) != MASTER_AGENT )
-	return;
+	return 0; /* no error if ! MASTER_AGENT */
 
-    DEBUGMSGTL(("snmpd","installing master agent on port %d", dest_port));
+    /* has something been specified before? */
+    cptr = ds_get_string(DS_APPLICATION_ID, DS_AGENT_PORTS);
+                      
+    /* set the specification string up */
+    if (cptr && dest_port)
+        /* append to the older specification string */
+        sprintf(buf,"%d,%s", dest_port, cptr);
+    else if (cptr)
+        sprintf(buf,"%s",cptr);
+    else
+        sprintf(buf,"%d",SNMP_PORT);
 
-    snmp_sess_init( &sess );
+    DEBUGMSGTL(("snmpd_ports","final port spec: %s\n", buf));
+    cptr = strtok(buf, ",");
+    while(cptr) {
+        /* XXX: surely, this creates memory leaks */
+        /* specification format: [transport:]port[@interface/address],... */
+        DEBUGMSGTL(("snmpd_open","installing master agent on port %s\n", cptr));
+
+        flags = ds_get_int(DS_APPLICATION_ID, DS_AGENT_FLAGS);
+
+        /* transport type */
+        if ((cptr2 = strchr(cptr, ':'))) {
+            if (strncasecmp(cptr,"tcp",3) == 0)
+                flags |= SNMP_FLAGS_STREAM_SOCKET;
+            else if (strncasecmp(cptr,"udp",3) == 0)
+                flags ^= SNMP_FLAGS_STREAM_SOCKET; /* fix */
+            else {
+                snmp_log(LOG_ERR, "illegal port transport %s\n", buf);
+                return 1;
+            }
+            cptr = cptr2+1;
+        }
+
+        /* null? */
+        if (!cptr || !(*cptr)) {
+            snmp_log(LOG_ERR, "improper port specification\n");
+            return 1;
+        }
+
+        dest_port = strtol(cptr, &cptr2, 0);
+
+        if (dest_port <= 0 || (*cptr2 && *cptr2 != '@')) {
+            /* XXX: add getservbyname lookups */
+            snmp_log(LOG_ERR, "improper port specification %s\n", cptr);
+            return 1;
+        }
+
+        memset(&sess, 0, sizeof(sess));
+        snmp_sess_init( &sess );
     
-    sess.version = SNMP_DEFAULT_VERSION;
-    sess.peername = SNMP_DEFAULT_PEERNAME;
-    sess.community_len = SNMP_DEFAULT_COMMUNITY_LEN;
+        sess.version = SNMP_DEFAULT_VERSION;
+        /* XXX: add interface binding (ie, eth0) */
+        if (cptr2 && *cptr2 == '@' && *(cptr2+1))
+            sess.peername = strdup(cptr2+1);
+        else
+            sess.peername = SNMP_DEFAULT_PEERNAME;
+        
+        sess.community_len = SNMP_DEFAULT_COMMUNITY_LEN;
      
-    sess.local_port = dest_port;
-    sess.callback = handle_snmp_packet;
-    sess.authenticator = NULL;
-    sess.flags = ds_get_int(DS_APPLICATION_ID, DS_AGENT_FLAGS);
-    session = snmp_open_ex( &sess, pre_parse, 0, post_parse, 0, 0 );
+        sess.local_port = dest_port;
+        sess.callback = handle_snmp_packet;
+        sess.authenticator = NULL;
+        sess.flags = flags;
+        sess.isAuthoritative = SNMP_SESS_AUTHORITATIVE;
+        session = snmp_open_ex( &sess, pre_parse, 0, post_parse, 0, 0 );
 
-    if ( session == NULL ) {
-      /* diagnose snmp_open errors with the input struct snmp_session pointer */
-	snmp_sess_perror("init_master_agent", &sess);
-	exit(1);
+        if ( session == NULL ) {
+            /* diagnose snmp_open errors with the input struct
+               snmp_session pointer */
+            snmp_sess_perror("init_master_agent", &sess);
+            return 1;
+        }
+        if (!main_session)
+            main_session = session;
+
+        /* next */
+        cptr = strtok(NULL, ",");
     }
-    main_session = session;
+    return 0;
 }
 
 struct agent_snmp_session  *
@@ -224,23 +285,39 @@ init_agent_snmp_session( struct snmp_session *session, struct snmp_pdu *pdu )
     asp = malloc( sizeof( struct agent_snmp_session ));
     if ( asp == NULL )
 	return NULL;
-    asp->start = pdu->variables;
-    asp->end   = pdu->variables;
-    if ( asp->end != NULL )
-	while ( asp->end->next_variable != NULL )
-	    asp->end = asp->end->next_variable;
     asp->session = session;
-    asp->pdu     = pdu;
+    asp->pdu      = snmp_clone_pdu(pdu);
+    asp->orig_pdu = snmp_clone_pdu(pdu);
     asp->rw      = READ;
     asp->exact   = TRUE;
     asp->outstanding_requests = NULL;
     asp->next    = NULL;
     asp->mode    = RESERVE1;
     asp->status  = SNMP_ERR_NOERROR;
+    asp->index   = 0;
+
+    asp->start = asp->pdu->variables;
+    asp->end   = asp->pdu->variables;
+    if ( asp->end != NULL )
+	while ( asp->end->next_variable != NULL )
+	    asp->end = asp->end->next_variable;
 
     return asp;
 }
 
+void
+free_agent_snmp_session(struct agent_snmp_session *asp)
+{
+    if (!asp)
+	return;
+    if (asp->orig_pdu)
+	snmp_free_pdu(asp->orig_pdu);
+    if (asp->pdu)
+	snmp_free_pdu(asp->pdu);
+
+    free(asp);
+}
+
 int
 count_varbinds( struct snmp_pdu *pdu )
 {
@@ -263,7 +340,7 @@ handle_snmp_packet(int operation, struct snmp_session *session, int reqid,
     struct variable_list *var_ptr, *var_ptr2;
 
     if ( magic == NULL ) {
-	asp = init_agent_snmp_session( session, snmp_clone_pdu(pdu) );
+	asp = init_agent_snmp_session( session, pdu );
 	status = SNMP_ERR_NOERROR;
     }
     else {
@@ -274,18 +351,21 @@ handle_snmp_packet(int operation, struct snmp_session *session, int reqid,
     if (asp->outstanding_requests != NULL)
 	return 1;
 
-    if ((status = check_access(pdu)) != 0) {
+    if ( check_access(pdu) != 0) {
         /* access control setup is incorrect */
-       send_easy_trap(SNMP_TRAP_AUTHFAIL, 0);
+	send_easy_trap(SNMP_TRAP_AUTHFAIL, 0);
         if (asp->pdu->version != SNMP_VERSION_1 && asp->pdu->version != SNMP_VERSION_2c) {
             asp->pdu->errstat = SNMP_ERR_AUTHORIZATIONERROR;
             asp->pdu->command = SNMP_MSG_RESPONSE;
             snmp_increment_statistic(STAT_SNMPOUTPKTS);
-            snmp_send( asp->session, asp->pdu );
+            if (! snmp_send( asp->session, asp->pdu ))
+	        snmp_free_pdu(asp->pdu);
+	    asp->pdu = NULL;
+	    free_agent_snmp_session(asp);
             return 1;
         } else {
             /* drop the request */
-            free( asp );
+            free_agent_snmp_session( asp );
             return 0;
         }
     }
@@ -334,13 +414,17 @@ handle_snmp_packet(int operation, struct snmp_session *session, int reqid,
 	        break;
     
 	    while ( asp->pdu->errstat-- > 0 )	/* Skip non-repeaters */
-	        asp->start = asp->start->next_variable;
+	    {
+	        if ( NULL != asp->start )	/* if there are variables ... */
+		    asp->start = asp->start->next_variable;
+	    }
 	    asp->pdu->errindex--;           /* Handled first repetition */
 
 	    if ( asp->outstanding_requests != NULL )
 		return 1;
 	}
 
+	if ( NULL != asp->start )	/* if there are variables ... */
 	while ( asp->pdu->errindex-- > 0 ) {	/* Process repeaters */
 		/*
 		 * Add new variable structures for the
@@ -353,12 +437,11 @@ handle_snmp_packet(int operation, struct snmp_session *session, int reqid,
 		  var_ptr = var_ptr->next_variable ) {
 				/* XXX: we don't know the size of the next
 					OID, so assume the maximum length */
-		if ( var_ptr->type != SNMP_ENDOFMIBVIEW )
-		{
-		var_ptr2 = snmp_add_null_var(asp->pdu, var_ptr->name, MAX_OID_LEN);
-		for ( i=var_ptr->name_length ; i<MAX_OID_LEN ; i++)
-		    var_ptr2->name[i] = 0;
-		var_ptr2->name_length = var_ptr->name_length;
+		if ( var_ptr->type != SNMP_ENDOFMIBVIEW ) {
+		    var_ptr2 = snmp_add_null_var(asp->pdu, var_ptr->name, MAX_OID_LEN);
+		    for ( i=var_ptr->name_length ; i<MAX_OID_LEN ; i++)
+			var_ptr2->name[i] = 0;
+		    var_ptr2->name_length = var_ptr->name_length;
 
 		    allDone = FALSE;
 		}
@@ -434,17 +517,20 @@ handle_snmp_packet(int operation, struct snmp_session *session, int reqid,
 	if ( asp->mode == COMMIT ) {
 	    status = handle_next_pass( asp );
 
-	    if ( status != SNMP_ERR_NOERROR )
-	        asp->mode = FINISHED_FAILURE;	/* or UNDO ? */
+	    if ( status != SNMP_ERR_NOERROR ) {
+		status    = SNMP_ERR_COMMITFAILED;
+	        asp->mode = FINISHED_FAILURE;
+	    }
 	    else
-	        asp->mode = FINISHED_SUCCESS;	/* or FREE ? */
+	        asp->mode = FINISHED_SUCCESS;
 
 	    if ( asp->outstanding_requests != NULL )
 		return 1;
 	}
 
 	if ( asp->mode == UNDO ) {
-	    status = handle_next_pass( asp );
+	    if (handle_next_pass( asp ) != SNMP_ERR_NOERROR )
+		status = SNMP_ERR_UNDOFAILED;
 
 	    asp->mode = FINISHED_FAILURE;
 	    break;
@@ -459,50 +545,107 @@ handle_snmp_packet(int operation, struct snmp_session *session, int reqid,
 
     case SNMP_MSG_RESPONSE:
         snmp_increment_statistic(STAT_SNMPINGETRESPONSES);
-	free( asp );
+	free_agent_snmp_session( asp );
 	return 0;
     case SNMP_MSG_TRAP:
     case SNMP_MSG_TRAP2:
         snmp_increment_statistic(STAT_SNMPINTRAPS);
-	free( asp );
+	free_agent_snmp_session( asp );
 	return 0;
     default:
         snmp_increment_statistic(STAT_SNMPINASNPARSEERRS);
-	free( asp );
+	free_agent_snmp_session( asp );
 	return 0;
     }
-	
-    
-	
+
     if ( asp->outstanding_requests != NULL ) {
 	asp->status = status;
 	asp->next = agent_session_list;
 	agent_session_list = asp;
     }
     else {
+		/*
+		 * May need to "dumb down" a SET error status for a
+		 *  v1 query.  See RFC2576 - section 4.3
+		 */
+	if (( asp->pdu->command == SNMP_MSG_SET ) &&
+	    ( asp->pdu->version == SNMP_VERSION_1 )) {
+	    switch ( status ) {
+		case SNMP_ERR_WRONGVALUE:
+		case SNMP_ERR_WRONGENCODING:
+		case SNMP_ERR_WRONGTYPE:
+		case SNMP_ERR_WRONGLENGTH:
+		case SNMP_ERR_INCONSISTENTVALUE:
+			status = SNMP_ERR_BADVALUE;
+			break;
+		case SNMP_ERR_NOACCESS:
+		case SNMP_ERR_NOTWRITABLE:
+		case SNMP_ERR_NOCREATION:
+		case SNMP_ERR_INCONSISTENTNAME:
+		case SNMP_ERR_AUTHORIZATIONERROR:
+			status = SNMP_ERR_NOSUCHNAME;
+			break;
+		case SNMP_ERR_RESOURCEUNAVAILABLE:
+		case SNMP_ERR_COMMITFAILED:
+		case SNMP_ERR_UNDOFAILED:
+			status = SNMP_ERR_GENERR;
+			break;
+	    }
+	}
+		/*
+		 * Similarly we may need to "dumb down" v2 exception
+		 *  types to throw an error for a v1 query.
+		 *  See RFC2576 - section 4.1.2.3
+		 */
+	if (( asp->pdu->command != SNMP_MSG_SET ) &&
+	    ( asp->pdu->version == SNMP_VERSION_1 )) {
+		for ( var_ptr = asp->pdu->variables, i=1 ;
+			var_ptr != NULL ;
+			var_ptr = var_ptr->next_variable, i++ ) {
+	    	    switch ( var_ptr->type ) {
+			case SNMP_NOSUCHOBJECT:
+			case SNMP_NOSUCHINSTANCE:
+			case SNMP_ENDOFMIBVIEW:
+			case ASN_COUNTER64:
+				status = SNMP_ERR_NOSUCHNAME;
+				asp->index=i;
+				break;
+		    }
+	    }
+	}
 	if ( status == SNMP_ERR_NOERROR ) {
 	    snmp_increment_statistic_by(
 		(asp->pdu->command == SNMP_MSG_SET ?
 			STAT_SNMPINTOTALSETVARS : STAT_SNMPINTOTALREQVARS ),
 	    	count_varbinds( asp->pdu ));
 	}
-	asp->pdu->command = SNMP_MSG_RESPONSE;
-	asp->pdu->errstat = status;
-	snmp_send( asp->session, asp->pdu );
+	else {
+		/*
+		 * Use a copy of the original request
+		 *   to report failures.
+		 */
+	    snmp_free_pdu( asp->pdu );
+	    asp->pdu = asp->orig_pdu;
+	    asp->orig_pdu = NULL;
+	}
+	asp->pdu->command  = SNMP_MSG_RESPONSE;
+	asp->pdu->errstat  = status;
+	asp->pdu->errindex = asp->index;
+	if (! snmp_send( asp->session, asp->pdu ))
+	    snmp_free_pdu(asp->pdu);
 	snmp_increment_statistic(STAT_SNMPOUTPKTS);
 	snmp_increment_statistic(STAT_SNMPOUTGETRESPONSES);
-	free( asp );
+	asp->pdu = NULL;
+	free_agent_snmp_session( asp );
     }
 
     return 1;
 }
 
-
 int
 handle_next_pass(struct agent_snmp_session  *asp)
 {
     int status;
-    struct snmp_pdu *pdu = asp->pdu;
     struct request_list *req_p, *next_req;
 
 
@@ -513,13 +656,20 @@ handle_next_pass(struct agent_snmp_session  *asp)
 	    if ( status == SNMP_ERR_NOERROR ) {
 		/* Send out any subagent requests */
 		for ( req_p = asp->outstanding_requests ;
-			req_p != NULL ; req_p = req_p->next_request ) {
+			req_p != NULL ; req_p = next_req ) {
 
-		    snmp_async_send( req_p->session,  req_p->pdu,
-				      req_p->callback, req_p->cb_data );
-		    asp->pdu = snmp_clone_pdu( pdu );
-		    asp->pdu->variables = pdu->variables;
-		    pdu->variables = NULL;
+		    next_req = req_p->next_request;
+		    if ( snmp_async_send( req_p->session,  req_p->pdu,
+				      req_p->callback, req_p->cb_data ) == 0) {
+				/*
+				 * Send failed - call callback to handle this
+				 */
+			(void)req_p->callback( SEND_FAILED,
+					 req_p->session,
+					 req_p->pdu->reqid,
+					 req_p->pdu,
+					 req_p->cb_data );
+		    }
 		}
 	    }
 	    else {
@@ -536,20 +686,27 @@ handle_next_pass(struct agent_snmp_session  *asp)
 	return status;
 }
 
+	/*
+	 *  Private structure to save the results of a getStatPtr call.
+	 *  This data can then be used to avoid repeating this call on
+	 *  subsequent SET handling passes.
+	 */
+struct saved_var_data {
+    WriteMethod *write_method;
+    u_char	*statP;
+    u_char	statType;
+    size_t	statLen;
+    u_short	acl;
+};
+
 
 int
 handle_var_list(struct agent_snmp_session  *asp)
 {
     struct variable_list *varbind_ptr;
-    u_char  statType;
-    u_char *statP;
-    size_t  statLen;
-    u_short acl;
-    WriteMethod *write_method;
-    AddVarMethod *add_method;
-    int	    noSuchObject;
-    int     count, view;
-    
+    int     status;
+    int     count;
+
     count = 0;
     varbind_ptr = asp->start;
     if ( !varbind_ptr ) {
@@ -557,14 +714,85 @@ handle_var_list(struct agent_snmp_session  *asp)
     }
 
     while (1) {
-    
 	count++;
+	asp->index = count;
+	status = handle_one_var(asp, varbind_ptr);
+
+	if ( status != SNMP_ERR_NOERROR ) {
+	    return status;
+	}
+
+	if ( varbind_ptr == asp->end )
+	     break;
+	varbind_ptr = varbind_ptr->next_variable;
+	if ( asp->mode == RESERVE1 )
+	    snmp_vars_inc++;
+   }
+   asp->index = 0;
+   return SNMP_ERR_NOERROR;
+}
+
+int
+handle_one_var(struct agent_snmp_session  *asp, struct variable_list *varbind_ptr)
+{
+    u_char  statType;
+    u_char *statP;
+    size_t  statLen;
+    u_short acl;
+    struct saved_var_data *saved;
+    WriteMethod *write_method;
+    AddVarMethod *add_method;
+    int	    noSuchObject = TRUE;
+    int     view;
+    oid	    save[MAX_OID_LEN];
+    size_t  savelen = 0;
+    
 statp_loop:
-	statP = getStatPtr(  varbind_ptr->name,
+	if ( asp->rw == WRITE && varbind_ptr->data != NULL ) {
+		/*
+		 * SET handling is "multi-pass", so restore
+		 *  results from an earlier 'getStatPtr' call
+		 *  to avoid repeating this processing.
+		 */
+	    saved = (struct saved_var_data *) varbind_ptr->data;
+	    write_method = saved->write_method;
+	    statP        = saved->statP;
+	    statType     = saved->statType;
+	    statLen      = saved->statLen;
+	    acl          = saved->acl;
+	}
+	else {
+		/*
+		 *  For exact requests (i.e. GET/SET),
+		 *   check whether this object is accessible
+		 *   before actually doing any work
+		 */
+	    if ( asp->exact )
+		view = in_a_view(varbind_ptr->name, &varbind_ptr->name_length,
+				   asp->pdu, varbind_ptr->type);
+	    else
+		view = 0;	/* Assume accessible */
+	    
+            memcpy(save, varbind_ptr->name,
+			varbind_ptr->name_length*sizeof(oid));
+            savelen = varbind_ptr->name_length;
+	    if ( view == 0 )
+	        statP = getStatPtr(  varbind_ptr->name,
 			   &varbind_ptr->name_length,
 			   &statType, &statLen, &acl,
 			   asp->exact, &write_method, asp->pdu, &noSuchObject);
+	    else {
+		if (view != 5) send_easy_trap(SNMP_TRAP_AUTHFAIL, 0);
+		statP        = NULL;
+		write_method = NULL;
+	    }
+	}
 			   
+		/*
+		 * For a valid SET request, having no existing value
+		 *  is (possibly) acceptable (and implies creation).
+		 * In all other situations, this indicates failure.
+		 */
 	if (statP == NULL && (asp->rw != WRITE || write_method == NULL)) {
 	    	varbind_ptr->val.integer   = NULL;
 	    	varbind_ptr->val_len = 0;
@@ -578,143 +806,113 @@ statp_loop:
 	            statType = SNMP_ENDOFMIBVIEW;
 		}
 		if (asp->pdu->version == SNMP_VERSION_1) {
-		    asp->pdu->errstat = SNMP_ERR_NOSUCHNAME;
-		    asp->pdu->errindex = count;
 		    return SNMP_ERR_NOSUCHNAME;
 		}
+		else if (asp->rw == WRITE) {
+		    return
+			( noSuchObject	? SNMP_ERR_NOTWRITABLE
+					: SNMP_ERR_NOCREATION ) ;
+		}
 		else
 		    varbind_ptr->type = statType;
 	}
-                /* Delegated variables should be added to the
-                   relevant outgoing request */
+                /*
+		 * Delegated variables should be added to the
+                 *  relevant outgoing request
+		 */
         else if ( IS_DELEGATED(statType)) {
                 add_method = (AddVarMethod*)statP;
-                statType = (*add_method)( asp, varbind_ptr );
+                return (*add_method)( asp, varbind_ptr );
         }
-		/* GETNEXT/GETBULK should just skip inaccessible entries */
-	else if ((view = in_a_view(varbind_ptr->name, &varbind_ptr->name_length,
-				   asp->pdu, varbind_ptr->type))
-			 && !asp->exact) {
+		/*
+		 * GETNEXT/GETBULK should just skip inaccessible entries
+		 */
+	else if (!asp->exact &&
+		 (view = in_a_view(varbind_ptr->name, &varbind_ptr->name_length,
+				   asp->pdu, varbind_ptr->type))) {
+
 		if (view != 5) send_easy_trap(SNMP_TRAP_AUTHFAIL, 0);
 		goto statp_loop;
 	}
-		/* Other access problems are permanent */
-	else if (( asp->rw == WRITE && !(acl & 2)) || view) {
-	    if (asp->pdu->version == SNMP_VERSION_1 || asp->rw != WRITE) {
-		if (ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_VERBOSE))
-                  DEBUGMSGTL(("snmp_agent", "    >> noSuchName (read-only)\n"));
-		ERROR_MSG("read-only");
-		statType = SNMP_ERR_NOSUCHNAME;
+#ifdef USING_AGENTX_PROTOCOL_MODULE
+		/*
+		 * AgentX GETNEXT/GETBULK requests need to take
+		 *   account of the end-of-range value
+		 *
+		 * This doesn't really belong here, but it works!
+		 */
+	else if (!asp->exact && asp->pdu->version == AGENTX_VERSION_1 &&
+		 snmp_oid_compare( varbind_ptr->name,
+			  	   varbind_ptr->name_length,
+				   varbind_ptr->val.objid,
+				   varbind_ptr->val_len/sizeof(oid)) > 0 ) {
+            	memcpy(varbind_ptr->name, save, savelen*sizeof(oid));
+            	varbind_ptr->name_length = savelen;
+		varbind_ptr->type = SNMP_ENDOFMIBVIEW;
+	}
+#endif
+		/*
+		 * Other access problems are permanent
+		 *   (i.e. writing to non-writeable objects)
+		 */
+	else if ( asp->rw == WRITE && !((acl & 2) && write_method)) {
+	    send_easy_trap(SNMP_TRAP_AUTHFAIL, 0);
+	    if (asp->pdu->version == SNMP_VERSION_1 ) {
+		return SNMP_ERR_NOSUCHNAME;
 	    }
 	    else {
-		if (ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_VERBOSE))
-                  DEBUGMSGTL(("snmp_agent", "    >> notWritable\n"));
-		ERROR_MSG("Not Writable");
-		statType = SNMP_ERR_NOTWRITABLE;
+		return SNMP_ERR_NOTWRITABLE;
 	    }
-	    asp->pdu->errstat = statType;
-	    asp->pdu->errindex = count;
-	    return statType;
         }
+		
 	else {
-            /* dump verbose info */
+		/*
+		 *  Things appear to have worked (so far)
+		 *  Dump the current value of this object
+		 */
 	    if (ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_VERBOSE) && statP)
 	        dump_var(varbind_ptr->name, varbind_ptr->name_length,
 				statType, statP, statLen);
 
-		/*  FINALLY we can act on SET requests ....*/
+		/*
+		 *  FINALLY we can act on SET requests ....
+		 */
 	    if ( asp->rw == WRITE ) {
-	        if ( write_method != NULL ) {
-		    statType = (*write_method)(asp->mode,
+		    if ( varbind_ptr->data == NULL ) {
+				/*
+				 * Save the results from 'getStatPtr'
+				 *  to avoid repeating this call
+				 */
+			saved = (struct saved_var_data *)malloc(sizeof(struct saved_var_data));
+			if ( saved == NULL ) {
+			    return SNMP_ERR_GENERR;
+			}
+	    		saved->write_method = write_method;
+	    		saved->statP        = statP;
+	    		saved->statType     = statType;
+	    		saved->statLen      = statLen;
+	    		saved->acl          = acl;
+			varbind_ptr->data = (void *)saved;
+		    }
+				/*
+				 * Call the object's write method
+				 */
+		    return (*write_method)(asp->mode,
                                                varbind_ptr->val.string,
                                                varbind_ptr->type,
                                                varbind_ptr->val_len, statP,
                                                varbind_ptr->name,
                                                varbind_ptr->name_length);
-                    if (statType != SNMP_ERR_NOERROR) {
-                      asp->pdu->errstat = statType;
-                      asp->pdu->errindex = count;
-                      return statType;
-                    }
-		}
-		else {
-                    if (!goodValue(varbind_ptr->type, varbind_ptr->val_len,
-                                    statType, statLen)){
-                        if (asp->pdu->version == SNMP_VERSION_1)
-                            statType = SNMP_ERR_BADVALUE;
-                        else
-                            statType = SNMP_ERR_WRONGTYPE; /* poor approximation */
-			asp->pdu->errstat = statType;
-			asp->pdu->errindex = count;
-			return statType;
-                    }
-                    /* actually do the set if necessary */
-                    if (asp->mode == COMMIT)
-                        setVariable(varbind_ptr->val.string, varbind_ptr->type,
-                                    varbind_ptr->val_len, statP, statLen);
-                }
 	    }
-		/* ... or save the results from assorted GETs */
+		/*
+		 * ... or save the results from assorted GET requests
+		 */
 	    else {
-		     snmp_set_var_value(varbind_ptr, statP, statLen);
-		     varbind_ptr->type = statType;
+		snmp_set_var_value(varbind_ptr, statP, statLen);
+		varbind_ptr->type = statType;
 	    }
 	}
+
+	return SNMP_ERR_NOERROR;
 	
-	if ( varbind_ptr == asp->end )
-	     return SNMP_ERR_NOERROR;
-	varbind_ptr = varbind_ptr->next_variable;
-	if ( asp->mode == RESERVE1 )
-	    snmp_vars_inc++;
-    }
-}
-
-
-
-static int
-goodValue(u_char inType, 
-	  size_t inLen,
-	  u_char actualType,
-	  size_t actualLen)
-{
-    if (inLen > actualLen)
-	return FALSE;
-    return (inType == actualType);
-}
-
-static void
-setVariable(u_char *var_val,
-	    u_char var_val_type,
-	    size_t var_val_len,
-	    u_char *statP,
-	    size_t statLen)
-{
-    size_t buffersize = 1000;
-
-    switch(var_val_type){
-	case ASN_INTEGER:
-	    asn_parse_int(var_val, &buffersize, &var_val_type, (long *)statP, statLen);
-	    break;
-	case ASN_COUNTER:
-	case ASN_GAUGE:
-	case ASN_TIMETICKS:
-	    asn_parse_unsigned_int(var_val, &buffersize, &var_val_type, (u_long *)statP, statLen);
-	    break;
-	case ASN_COUNTER64:
-	    asn_parse_unsigned_int64(var_val, &buffersize, &var_val_type,
-				     (struct counter64 *)statP, statLen);
-	    break;
-	case ASN_OCTET_STR:
-	case ASN_IPADDRESS:
-	case ASN_OPAQUE:
-	case ASN_NSAP:
-	    asn_parse_string(var_val, &buffersize, &var_val_type, statP, &statLen);
-	    break;
-	case ASN_OBJECT_ID:
-	    asn_parse_objid(var_val, &buffersize, &var_val_type, (oid *)statP, &statLen);
-	    break;
-	case ASN_BIT_STR:
-	    asn_parse_bitstring(var_val, &buffersize, &var_val_type, statP, &statLen);
-	    break;
-    }
 }
