@@ -123,13 +123,13 @@ typedef long    fd_mask;
 #include "snmpd.h"
 #include "var_struct.h"
 #include "mibgroup/struct.h"
-#include "mibgroup/util_funcs.h"
 #include "snmp_debug.h"
 #include "mib_modules.h"
 
 #include "snmpusm.h"
 #include "tools.h"
 #include "lcd_time.h"
+#include "mibgroup/util_funcs.h"
 
 #include "transform_oids.h"
 
@@ -206,7 +206,7 @@ static void usage(char *prog)
 	printf("\n");
 	printf("\n\tVersion:  %s\n",VersionInfo);
 	printf("\tAuthor:   Wes Hardaker\n");
-	printf("\tEmail:    ucd-snmp-coders@ucd-snmp.ucdavis.edu\n");
+	printf("\tEmail:    net-snmp-coders@lists.sourceforge.net\n");
 	printf("\n-h\t\tThis usage message.\n");
 	printf("-H\t\tDisplay configuration file directives understood.\n");
 	printf("-v\t\tVersion information.\n");
@@ -218,13 +218,20 @@ static void usage(char *prog)
 	printf("-q\t\tPrint information in a more parsable format (quick-print)\n");
 	printf("-D\t\tTurn on debugging output\n");
 	printf("-p NUM\t\tRun on port NUM instead of the default:  161\n");
-	printf("-x SOCKADDR\tBind AgentX to this address\n");
+#if defined(USING_AGENTX_SUBAGENT_MODULE) || defined(USING_AGENTX_MASTER_MODULE)
+	printf("-x SOCKADDR\tBind AgentX port to this address\n");
+#endif
+#ifdef USING_AGENTX_SUBAGENT_MODULE
+	printf("-X\t\tRun as an AGENTX subagent rather than an SNMP master agent.\n");
+#endif
 	printf("-c CONFFILE\tRead CONFFILE as a configuration file.\n");
 	printf("-C\t\tDon't read the default configuration files.\n");
 	printf("-L\t\tPrint warnings/messages to stdout/err\n");
 	printf("-s\t\tLog warnings/messages to syslog\n");
 	printf("-A\t\tAppend to the logfile rather than truncating it.\n");
 	printf("-r Don't exit if root only accessible files can't be opened\n");
+	printf("-I [-]INITLIST\tList of mib modules to initialize (or not).\n");
+	printf("\t\t (run snmpd with -Dinit_mib for a list)\n");
 	printf("-l LOGFILE\tPrint warnings/messages to LOGFILE\n");
 	printf("\t\t(By default LOGFILE=%s)\n",
 #ifdef LOGFILE
@@ -294,17 +301,18 @@ main(int argc, char *argv[])
 {
 	int             arg, i;
 	int             ret;
-	u_short         dest_port = SNMP_PORT;
 	int             dont_fork = 0;
 	char            logfile[SNMP_MAXBUF_SMALL];
 	char           *cptr, **argvptr;
 	char           *pid_file = NULL;
+        char            buf[SPRINT_MAX_LEN];
 #if HAVE_GETPID
 	FILE           *PID;
 #endif
 	int             dont_zero_log = 0;
 	int             stderr_log=0, syslog_log=0;
 	int             uid=0, gid=0;
+        int             agent_mode=-1;
 
 	logfile[0]		= 0;
 
@@ -312,6 +320,10 @@ main(int argc, char *argv[])
 	strcpy(logfile, LOGFILE);
 #endif
 
+#ifdef NO_ROOT_ACCESS
+        /* default to no */
+        ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_NO_ROOT_ACCESS, 1);
+#endif
 
 	/*
 	 * usage: snmpd
@@ -374,20 +386,42 @@ main(int argc, char *argv[])
                 case 'p':
                   if (++arg == argc)
                     usage(argv[0]);
-                  dest_port = atoi(argv[arg]);
-                  if (dest_port <= 0)
-                    usage(argv[0]);
+
+                  /* has something been specified before? */
+                  cptr = ds_get_string(DS_APPLICATION_ID, DS_AGENT_PORTS);
+                      
+                  /* set the specification string up */
+                  if (cptr)
+                      /* append to the older specification string */
+                      sprintf(buf,"%s,%s", cptr, argv[arg]);
+                  else
+                      strcpy(buf,argv[arg]);
+
+                  DEBUGMSGTL(("snmpd_ports","port spec: %s\n", buf));
+                  ds_set_string(DS_APPLICATION_ID, DS_AGENT_PORTS, strdup(buf));
                   break;
 
+#if defined(USING_AGENTX_SUBAGENT_MODULE) || defined(USING_AGENTX_MASTER_MODULE)
                 case 'x':
                   if (++arg == argc)
                     usage(argv[0]);
                   ds_set_string(DS_APPLICATION_ID, DS_AGENT_X_SOCKET, argv[arg]);
                   break;
+#endif
+
+                case 'X':
+#if defined(USING_AGENTX_SUBAGENT_MODULE)
+                  agent_mode = SUB_AGENT;
+#else
+                  fprintf(stderr,"Illegal argument -X: AgentX support not compiled in.\n");
+                  usage(argv[0]);
+                  exit(1);
+#endif
+                  break;
 
 		case 'r':
-                    ds_set_boolean(DS_APPLICATION_ID,
-                                      DS_AGENT_NO_ROOT_ACCESS, 1);
+                    ds_toggle_boolean(DS_APPLICATION_ID,
+                                      DS_AGENT_NO_ROOT_ACCESS);
 		    break;
 
                 case 'P':
@@ -424,20 +458,27 @@ main(int argc, char *argv[])
                 case 'A':
                     dont_zero_log = 1;
                     break;
+
+                case 'I':
+                  if (++arg == argc) usage(argv[0]);
+                  add_to_init_list(argv[arg]);
+                  break;
+
 #if HAVE_UNISTD_H
                 case 'u':
                   if (++arg == argc) usage(argv[0]);
-                  uid = atoi(argv[arg]);
+                  ds_set_int(DS_APPLICATION_ID, DS_AGENT_USERID,atoi(argv[arg]));
                   break;
                 case 'g':
                   if (++arg == argc) usage(argv[0]);
-                  gid = atoi(argv[arg]);
+                  ds_set_int(DS_APPLICATION_ID, DS_AGENT_GROUPID, atoi(argv[arg]));
                   break;
 #endif
                 case 'h':
                   usage(argv[0]);
                   break;
                 case 'H':
+                  ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_NO_ROOT_ACCESS, 1);
                   init_agent("snmpd");   /* register our .conf handlers */
                   init_mib_modules();
                   init_snmp("snmpd");
@@ -481,10 +522,15 @@ main(int argc, char *argv[])
 	argvrestart = (char *) malloc(ret);
 	argvrestartname = (char *) malloc(strlen(argv[0]) + 1);
 	strcpy(argvrestartname, argv[0]);
-	if ( strstr(argvrestartname, "agentxd") != NULL)
-          ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_ROLE, SUB_AGENT);
-	else
-          ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_ROLE, MASTER_AGENT);
+        if (agent_mode == -1) {
+            if ( strstr(argvrestartname, "agentxd") != NULL )
+                ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_ROLE, SUB_AGENT);
+            else
+                ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_ROLE, MASTER_AGENT);
+        } else {
+            ds_set_boolean(DS_APPLICATION_ID, DS_AGENT_ROLE, agent_mode);
+        }
+        
 	for (cptr = argvrestart, i = 0; i < argc; i++) {
 		strcpy(cptr, argv[i]);
 		*(argvptr++) = cptr;
@@ -493,7 +539,6 @@ main(int argc, char *argv[])
 	*cptr = 0;
 	*argvptr = NULL;
 
-
 	/* 
 	 * Open the logfile if necessary.
 	 */
@@ -538,7 +583,7 @@ main(int argc, char *argv[])
     /* start library */
     init_snmp("snmpd");
 
-    ret = init_master_agent( dest_port,
+    ret = init_master_agent( 0,
                        snmp_check_packet,
                        snmp_check_parse );
 	if( ret != 0 )
@@ -556,27 +601,37 @@ main(int argc, char *argv[])
 #ifdef SIGUSR1
     signal(SIGUSR1, SnmpdDump);
 #endif
+#ifdef SIGPIPE
+    signal(SIGPIPE, SIG_IGN);	/* 'Inline' failure of wayward readers */
+#endif
+
+    /* store persistent data immediately in case we crash later */
+    snmp_store("snmpd");
 
     /* send coldstart trap via snmptrap(1) if possible */
     send_easy_trap (0, 0);
         
 #if HAVE_UNISTD_H
-	if (gid) {
+#ifdef HAVE_SETGID
+	if ((gid = ds_get_int(DS_APPLICATION_ID, DS_AGENT_GROUPID)) != 0) {
 		DEBUGMSGTL(("snmpd", "Changing gid to %d.\n", gid));
 		if (setgid(gid)==-1) {
 			snmp_log_perror("setgid failed");
-        		if (!ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_NO_ROOT_ACCESS))
-			  exit(1);
+			if (!ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_NO_ROOT_ACCESS))
+			    exit(1);
 		}
 	}
-	if (uid) {
+#endif
+#ifdef HAVE_SETUID
+	if ((uid = ds_get_int(DS_APPLICATION_ID, DS_AGENT_USERID)) != 0) {
 		DEBUGMSGTL(("snmpd", "Changing uid to %d.\n", uid));
 		if(setuid(uid)==-1) {
 			snmp_log_perror("setuid failed");
-        		if (!ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_NO_ROOT_ACCESS))
-			  exit(1);
+			if (!ds_get_boolean(DS_APPLICATION_ID, DS_AGENT_NO_ROOT_ACCESS))
+			    exit(1);
 		}
 	}
+#endif
 #endif
 
 	/* honor selection of standard error output */
@@ -618,13 +673,13 @@ main(int argc, char *argv[])
 receive(void)
 {
     int numfds;
-    fd_set fdset;
+    fd_set readfds, writefds, exceptfds;
     struct timeval	timeout, *tvp = &timeout;
     struct timeval	sched,   *svp = &sched,
 			now,     *nvp = &now;
-    int count, block;
+    int count, block, i;
 #ifdef	USING_SMUX_MODULE
-    int i, sd;
+    int sd;
 #endif	/* USING_SMUX_MODULE */
 
 
@@ -646,42 +701,116 @@ receive(void)
      */
     while (running) {
         if (reconfig) {
-          reconfig = 0;
-          snmp_log(LOG_INFO, "Reconfiguring daemon\n");
-          update_config();
+	    reconfig = 0;
+	    snmp_log(LOG_INFO, "Reconfiguring daemon\n");
+	    update_config();
         }
 	tvp =  &timeout;
 	tvp->tv_sec = 0;
 	tvp->tv_usec = TIMETICK;
 
 	numfds = 0;
-	FD_ZERO(&fdset);
+	FD_ZERO(&readfds);
+	FD_ZERO(&writefds);
+	FD_ZERO(&exceptfds);
         block = 0;
-        snmp_select_info(&numfds, &fdset, tvp, &block);
+        snmp_select_info(&numfds, &readfds, tvp, &block);
         if (block == 1)
             tvp = NULL; /* block without timeout */
 
 #ifdef	USING_SMUX_MODULE
-		if (smux_listen_sd >= 0) {
-			FD_SET(smux_listen_sd, &fdset);
-			numfds = smux_listen_sd >= numfds ? smux_listen_sd + 1 : numfds;
-			for (i = 0; i < sdlen; i++) {
-				FD_SET(sdlist[i], &fdset);
-				numfds = sdlist[i] >= numfds ? sdlist[i] + 1 : numfds;
-			}
-		}
+	if (smux_listen_sd >= 0) {
+	    FD_SET(smux_listen_sd, &readfds);
+	    numfds = smux_listen_sd >= numfds ? smux_listen_sd + 1 : numfds;
+	    for (i = 0; i < sdlen; i++) {
+		FD_SET(sdlist[i], &readfds);
+		numfds = sdlist[i] >= numfds ? sdlist[i] + 1 : numfds;
+	    }
+	}
 #endif	/* USING_SMUX_MODULE */
 
-	count = select(numfds, &fdset, 0, 0, tvp);
+	for (i = 0; i < external_readfdlen; i++) {
+	    FD_SET(external_readfd[i], &readfds);
+	    if (external_readfd[i] >= numfds)
+		numfds = external_readfd[i] + 1;
+	}
+	for (i = 0; i < external_writefdlen; i++) {
+	    FD_SET(external_writefd[i], &writefds);
+	    if (external_writefd[i] >= numfds)
+		numfds = external_writefd[i] + 1;
+	}
+	for (i = 0; i < external_exceptfdlen; i++) {
+	    FD_SET(external_exceptfd[i], &exceptfds);
+	    if (external_exceptfd[i] >= numfds)
+		numfds = external_exceptfd[i] + 1;
+	}
 
-	if (count > 0){
-	    snmp_read(&fdset);
-	} else switch(count){
+	for (i = 0; i < NUM_EXTERNAL_SIGS; i++) {
+	    if (external_signal_scheduled[i]) {
+		external_signal_scheduled[i]--;
+		external_signal_handler[i](i);
+	    }
+	}
+	
+	count = select(numfds, &readfds, &writefds, &exceptfds, tvp);
+
+	if (count > 0) {
+
+#ifdef USING_SMUX_MODULE 
+            /* handle the SMUX sd's */ 
+            if (smux_listen_sd >= 0) { 
+                for (i = 0; i < sdlen; i++) { 
+                    if (FD_ISSET(sdlist[i], &readfds)) { 
+                        if (smux_process(sdlist[i]) < 0) { 
+                            for (; i < (sdlen - 1); i++) { 
+                                sdlist[i] = sdlist[i+1]; 
+                            } 
+                            sdlen--; 
+                        } 
+                    } 
+                } 
+                /* new connection */ 
+                if (FD_ISSET(smux_listen_sd, &readfds)) { 
+                    if ((sd = smux_accept(smux_listen_sd)) >= 0) { 
+                        sdlist[sdlen++] = sd; 
+                    } 
+                } 
+            } 
+#endif /* USING_SMUX_MODULE */ 
+
+	    snmp_read(&readfds);
+
+	    for (i = 0; count && (i < external_readfdlen); i++) {
+		if (FD_ISSET(external_readfd[i], &readfds)) {
+		    external_readfdfunc[i](external_readfd[i],
+					   external_readfd_data[i]);
+		    FD_CLR(external_readfd[i], &readfds);
+		    count--;
+		}
+	    }
+	    for (i = 0; count && (i < external_writefdlen); i++) {
+		if (FD_ISSET(external_writefd[i], &writefds)) {
+		    external_writefdfunc[i](external_writefd[i],
+					    external_writefd_data[i]);
+		    FD_CLR(external_writefd[i], &writefds);
+		    count--;
+		}
+	    }
+	    for (i = 0; count && (i < external_exceptfdlen); i++) {
+		if (FD_ISSET(external_exceptfd[i], &exceptfds)) {
+		    external_exceptfdfunc[i](external_exceptfd[i],
+					     external_exceptfd_data[i]);
+		    FD_CLR(external_exceptfd[i], &exceptfds);
+		    count--;
+		}
+	    }
+	    
+	} else switch (count) {
 	    case 0:
                 snmp_timeout();
                 break;
 	    case -1:
-		if (errno == EINTR){
+		if (errno == EINTR) {
 		    continue;
 		} else {
                     snmp_log_perror("select");
@@ -692,27 +821,6 @@ receive(void)
 		return -1;
 	}  /* endif -- count>0 */
 
-#ifdef	USING_SMUX_MODULE
-		/* handle the SMUX sd's */
-		if (smux_listen_sd >= 0) {
-			for (i = 0; i < sdlen; i++) {
-				if (FD_ISSET(sdlist[i], &fdset)) {
-					if (smux_process(sdlist[i]) < 0) {
-						for (; i < (sdlen - 1); i++) {
-							sdlist[i] = sdlist[i+1];
-						}
-						sdlen--;
-					}
-				}
-			}
-			/* new connection */
-			if (FD_ISSET(smux_listen_sd, &fdset)) {
-				if ((sd = smux_accept(smux_listen_sd)) >= 0) {
-					sdlist[sdlen++] = sd;
-				}
-			}
-		}
-#endif	/* USING_SMUX_MODULE */
 
 
 
@@ -918,4 +1026,4 @@ snmp_input(int op,
     }
     return 1;
 
-}  /* end snmp_input() */
+} /* end snmp_input() */
