@@ -333,7 +333,16 @@ int main(int argc, char *argv[])
   int print = 1;
   int exit_code = 0;
     
-  arg = snmp_parse_args(argc, argv, &session, "C:", &optProc);
+  switch (arg = snmp_parse_args(argc, argv, &session, "C:", &optProc)) {
+  case -2:
+      exit(0);
+  case -1:
+      usage();
+      exit(1);
+  default:
+        break;
+  }
+  
   gateway = session.peername;
 
   /*
@@ -633,8 +642,8 @@ int main(int argc, char *argv[])
 	  }
 	  end = last_end;
 	  continue;
-	} else if (response->errstat == SNMP_ERR_NOSUCHNAME){
-	  fprintf(stderr, "This object doesn't exist on %s: ", gateway);
+	} else if (response->errindex != 0){
+	  fprintf(stderr, "Failed object: ");
 	  for(count = 1, vars = response->variables;
 	      vars && count != response->errindex;
 	      vars = vars->next_variable, count++)
@@ -666,10 +675,12 @@ int main(int argc, char *argv[])
     } else if (status == STAT_TIMEOUT){
       fprintf(stderr, "Timeout: No Response from %s\n", gateway);
       response = 0;
+      exit_code = 1;
       break;
     } else {    /* status == STAT_ERROR */
       snmp_sess_perror("snmpdelta", ss);
       response = 0;
+      exit_code = 1;
       break;
     }
 	
