@@ -124,9 +124,18 @@ int main(int argc, char *argv[])
     oid name[MAX_OID_LEN];
     size_t name_length;
     int status;
+    int exitval = 0;
 
     /* get the common command line arguments */
-    arg = snmp_parse_args(argc, argv, &session, "C:", optProc);
+    switch (arg = snmp_parse_args(argc, argv, &session, "C:", optProc)) {
+    case -2:
+	exit(0);
+    case -1:
+        usage();
+        exit(1);
+    default:
+        break;
+    }
 
     if (arg >= argc) {
       fprintf(stderr, "Missing object name\n");
@@ -188,8 +197,8 @@ retry:
         fprintf(stderr, "Error in packet\nReason: %s\n",
                 snmp_errstring(response->errstat));
 
-        if (response->errstat == SNMP_ERR_NOSUCHNAME){
-          fprintf(stderr, "This name doesn't exist: ");
+        if (response->errindex != 0){
+          fprintf(stderr, "Failed object: ");
           for(count = 1, vars = response->variables; 
                 vars && count != response->errindex;
                 vars = vars->next_variable, count++)
@@ -198,6 +207,7 @@ retry:
             fprint_objid(stderr, vars->name, vars->name_length);
           fprintf(stderr, "\n");
         }
+	exitval = 2;
 
         /* retry if the errored variable was successfully removed */
         if (!ds_get_boolean(DS_APPLICATION_ID, DS_APP_DONT_FIX_PDUS)) {
@@ -211,15 +221,11 @@ retry:
 
     } else if (status == STAT_TIMEOUT){
 	fprintf(stderr,"Timeout: No Response from %s.\n", session.peername);
-	snmp_close(ss);
-	SOCK_CLEANUP;
-	exit(1);
+	exitval = 1;
 
     } else {    /* status == STAT_ERROR */
       snmp_sess_perror("snmpget", ss);
-      snmp_close(ss);
-      SOCK_CLEANUP;
-      exit(1);
+      exitval = 1;
 
     }  /* endif -- STAT_SUCCESS */
 
@@ -228,7 +234,7 @@ retry:
       snmp_free_pdu(response);
     snmp_close(ss);
     SOCK_CLEANUP;
-    return 0;
+    return exitval;
 
 }  /* end main() */
 
