@@ -121,9 +121,18 @@ int main(int argc, char *argv[])
     size_t name_length;
     int   status;
     int   failures = 0;
+    int   exitval = 0;
 
     /* get the common command line arguments */
-    arg = snmp_parse_args(argc, argv, &session, "C:", &optProc);
+    switch (arg = snmp_parse_args(argc, argv, &session, "C:", &optProc)) {
+    case -2:
+    	exit(0);
+    case -1:
+        usage();
+        exit(1);
+    default:
+        break;
+    }
 
     if (arg >= argc) {
       fprintf(stderr, "Missing object name\n");
@@ -172,8 +181,8 @@ retry:
         } else {
           fprintf(stderr, "Error in packet.\nReason: %s\n",
                   snmp_errstring(response->errstat));
-          if (response->errstat == SNMP_ERR_NOSUCHNAME){
-            fprintf(stderr, "This name doesn't exist: ");
+          if (response->errindex != 0){
+            fprintf(stderr, "Failed object: ");
           for(count = 1, vars = response->variables;
                 vars && count != response->errindex;
                 vars = vars->next_variable, count++)
@@ -181,6 +190,7 @@ retry:
           if (vars)
             fprint_objid(stderr, vars->name, vars->name_length);
           fprintf(stderr, "\n");
+	  exitval = 2;
         }
 
         /* retry if the errored variable was successfully removed */
@@ -194,13 +204,15 @@ retry:
       }
     } else if (status == STAT_TIMEOUT){
       fprintf(stderr, "Timeout: No Response from %s.\n", session.peername);
+      exitval = 1;
     } else {    /* status == STAT_ERROR */
       snmp_sess_perror("snmpgetnext", ss);
+      exitval = 1;
     }
 
     if (response)
       snmp_free_pdu(response);
     snmp_close(ss);
     SOCK_CLEANUP;
-    return 0;
+    return exitval;
 }
