@@ -116,9 +116,18 @@ int main(int argc, char *argv[])
     oid name[MAX_OID_LEN];
     size_t name_length;
     int status;
+    int exitval = 0;
 
     /* get the common command line arguments */
-    arg = snmp_parse_args(argc, argv, &session, NULL, NULL);
+    switch (arg = snmp_parse_args(argc, argv, &session, NULL, NULL)) {
+    case -2:
+    	exit(0);
+    case -1:
+        usage();
+        exit(1);
+    default:
+        break;
+    }
 
     if (arg >= argc) {
       fprintf(stderr, "Missing object name\n");
@@ -139,6 +148,7 @@ int main(int argc, char *argv[])
         case 's':
         case 'x':
         case 'd':
+	case 'b':
 #ifdef OPAQUE_SPECIAL_TYPES
         case 'I':
         case 'U':
@@ -148,20 +158,17 @@ int main(int argc, char *argv[])
           types[current_type++] = *argv[arg++];
           break;
         default:
-          fprintf(stderr, "Bad object type: %c\n", *argv[arg]);
-          usage();
+          fprintf(stderr, "%s: Bad object type: %c\n", argv[arg-1], *argv[arg]);
           exit(1);
         }
       } else {
-        fprintf(stderr, "Need value type\n");
-        usage();
+        fprintf(stderr, "%s: Needs type and value\n", argv[arg-1]);
         exit(1);
       }
       if (arg < argc)
         values[current_value++] = argv[arg];
       else {
-        fprintf(stderr, "Need value\n");
-        usage();
+        fprintf(stderr, "%s: Needs value\n", argv[arg-2]);
         exit(1);
       }
     }
@@ -205,8 +212,8 @@ int main(int argc, char *argv[])
         } else {
           fprintf(stderr, "Error in packet.\nReason: %s\n",
                  snmp_errstring(response->errstat));
-          if (response->errstat == SNMP_ERR_NOSUCHNAME){
-            fprintf(stderr, "This name doesn't exist: ");
+          if (response->errindex != 0){
+            fprintf(stderr, "Failed object: ");
             for(count = 1, vars = response->variables;
                   vars && (count != response->errindex);
                   vars = vars->next_variable, count++)
@@ -215,22 +222,19 @@ int main(int argc, char *argv[])
               fprint_objid(stderr, vars->name, vars->name_length);
             fprintf(stderr, "\n");
           }
+	  exitval = 2;
         }
       } else if (status == STAT_TIMEOUT){
         fprintf(stderr,"Timeout: No Response from %s\n", session.peername);
-	snmp_close(ss);
-	SOCK_CLEANUP;
-	exit(1);
+	exitval = 1;
       } else {    /* status == STAT_ERROR */
         snmp_sess_perror("snmpset", ss);
-	snmp_close(ss);
-	SOCK_CLEANUP;
-	exit(1);
+	exitval = 1;
     }
 
     if (response)
       snmp_free_pdu(response);
     snmp_close(ss);
     SOCK_CLEANUP;
-    return 0;
+    return exitval;
 }
