@@ -144,9 +144,18 @@ int main(int argc, char *argv[])
     char buf[40];
     int interfaces;
     int count;
+    int exitval = 0;
 
     /* get the common command line arguments */
-    snmp_parse_args(argc, argv, &session, "C:", &optProc);
+    switch (snmp_parse_args(argc, argv, &session, "C:", &optProc)) {
+    case -2:
+    	exit(0);
+    case -1:
+        usage();
+        exit(1);
+    default:
+        break;
+    }
 
     SOCK_STARTUP;
 
@@ -194,8 +203,8 @@ retry:
       } else {
         fprintf(stderr, "Error in packet.\nReason: %s\n",
                 snmp_errstring(response->errstat));
-        if (response->errstat == SNMP_ERR_NOSUCHNAME){
-          fprintf(stderr, "This name doesn't exist: ");
+        if (response->errindex != 0){
+          fprintf(stderr, "Failed object: ");
           for(count = 1, vars = response->variables;
                 vars && count != response->errindex;
                 vars = vars->next_variable, count++)
@@ -287,8 +296,8 @@ retry:
         } else {
           fprintf(stderr, "Error in packet.\nReason: %s\n",
                   snmp_errstring(response->errstat));
-          if (response->errstat == SNMP_ERR_NOSUCHNAME){
-            fprintf(stderr, "This name doesn't exist: ");
+          if (response->errindex != 0){
+            fprintf(stderr, "Failed object: ");
             for(count = 1, vars = response->variables;
                   vars && count != response->errindex;
                   vars = vars->next_variable, count++)
@@ -297,11 +306,14 @@ retry:
               fprint_objid(stderr, vars->name, vars->name_length);
             fprintf(stderr, "\n");
           }
+	  exitval = 2;
         }
       } else if (status == STAT_TIMEOUT){
         fprintf(stderr,"Timeout: No Response from %s\n", session.peername);
+	exitval = 1;
       } else {    /* status == STAT_ERROR */
         snmp_sess_perror("snmpstatus", ss);
+	exitval = 1;
       }
 
       if (response)
@@ -316,5 +328,5 @@ retry:
 
     snmp_close(ss);
     SOCK_CLEANUP;
-    return 0;
+    return exitval;
 }
