@@ -70,6 +70,7 @@ SOFTWARE.
 #if HAVE_ARPA_INET_H
 #include <arpa/inet.h>
 #endif
+#include <getopt.h>
 
 #include "asn1.h"
 #include "snmp_api.h"
@@ -93,6 +94,9 @@ void usage(void)
     snmp_parse_args_usage(stderr);
     fprintf(stderr," [<trap parameters> ...]\n\n");
     snmp_parse_args_descriptions(stderr);
+    fprintf(stderr, "  -C <APPOPTS>  Toggle various application specific behaviour:\n");
+    fprintf(stderr, "\t\t  APPOPTS values:\n");
+    fprintf(stderr,"\t\t      i: Send an snmp INFORM instead of an snmp TRAP.\n");
     fprintf(stderr, "  -v 1 trap parameters:\n\t enterprise-oid agent trap-type specific-type uptime [ var ]...\n");
     fprintf(stderr, "  or\n");
     fprintf(stderr, "  -v 2 trap parameters:\n\t uptime trapoid [ var ] ...\n");
@@ -126,6 +130,26 @@ in_addr_t parse_address(char *address)
 
 }
 
+static void optProc(int argc, char *const *argv, int opt)
+{
+    switch (opt) {
+        case 'C':
+            while (*optarg) {
+                switch (*optarg++) {
+                    case 'i':
+                        inform = 1;
+                        break;
+                    default:
+                        fprintf(stderr,
+                                "Unknown flag passed to -C: %c\n", *optarg);
+                        usage();
+                        exit(1);
+                }
+            }
+            break;
+    }
+}
+
 int main(int argc, char *argv[])
 {
     struct snmp_session session, *ss;
@@ -137,13 +161,22 @@ int main(int argc, char *argv[])
     int status;
     char *trap = NULL, *specific = NULL, *description = NULL, *agent = NULL;
     char *prognam;
+    int exitval = 0;
 
     prognam = strrchr(argv[0], '/');
     if (prognam) prognam++;
     else prognam = argv[0];
 
     if (strcmp(prognam, "snmpinform") == 0) inform = 1;
-    arg = snmp_parse_args(argc, argv, &session, NULL, NULL);
+    switch (arg = snmp_parse_args(argc, argv, &session, "C:", optProc)) {
+    case -2:
+    	exit(0);
+    case -1:
+        usage();
+        exit(1);
+    default:
+        break;
+    }
 
     SOCK_STARTUP;
 
@@ -331,11 +364,12 @@ int main(int argc, char *argv[])
     else status = snmp_send(ss, pdu) == 0;
     if (status) {
         snmp_sess_perror(inform ? "snmpinform" : "snmptrap", ss);
-	snmp_free_pdu(pdu);
+	if (!inform) snmp_free_pdu(pdu);
+	exitval = 1;
     }
     else if (inform) snmp_free_pdu(response);
 
     snmp_close(ss);
     SOCK_CLEANUP;
-    return (0);
+    return exitval;
 }
