@@ -98,6 +98,7 @@ SOFTWARE.
 #include "system.h"
 #include "version.h"
 #include "snmptrapd_handlers.h"
+#include "snmptrapd_log.h"
 #include "read_config.h"
 #include "snmp_debug.h"
 #include "snmp_logging.h"
@@ -127,6 +128,7 @@ typedef long	fd_mask;
 #define FD_ZERO(p)      memset((p), 0, sizeof(*(p)))
 #endif
 
+char *logfile = 0;
 int Print = 0;
 int Syslog = 0;
 int Event = 0;
@@ -134,6 +136,11 @@ int dropauth = 0;
 int running = 1;
 int reconfig = 0;
 
+const char *trap1_std_str = "%.4y-%.2m-%.2l %.2h:%.2j:%.2k %B [%b] (via %A [%a]): %N\n\t%W Trap (%q) Uptime: %#T\n%v\n",
+	   *trap2_std_str = "%.4y-%.2m-%.2l %.2h:%.2j:%.2k %B [%b]:\n%v\n";
+char       *trap1_fmt_str = NULL,
+           *trap2_fmt_str = NULL; /* how to format logging to stdout */
+
 /*
  * These definitions handle 4.2 systems without additional syslog facilities.
  */
@@ -416,12 +423,11 @@ int snmp_input(int op,
 	       struct snmp_pdu *pdu,
 	       void *magic)
 {
+    char out_bfr[SPRINT_MAX_LEN];
     struct variable_list *vars;
     struct sockaddr_in *pduIp   = (struct sockaddr_in *)&(pdu->address);
     char buf[64], oid_buf [SPRINT_MAX_LEN], *cp;
     struct snmp_pdu *reply;
-    struct tm *tm;
-    time_t timer;
     struct hostent *host;
     int varbufidx;
     char varbuf[SPRINT_MAX_LEN];
@@ -447,43 +453,11 @@ int snmp_input(int op,
 		trapOid[trapOidLen++] = pdu->specific_type;
 	    }
 	    if (Print && (pdu->trap_type != SNMP_TRAP_AUTHFAIL || dropauth == 0)) {
-		time (&timer);
-		tm = localtime (&timer);
-                printf("%.4d-%.2d-%.2d %.2d:%.2d:%.2d %s [%s] ",
-		       tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
-		       tm->tm_hour, tm->tm_min, tm->tm_sec,
-                       host ? host->h_name : inet_ntoa(agentIp->sin_addr),
-                       inet_ntoa(agentIp->sin_addr));
-		if (agentIp->sin_addr.s_addr != pduIp->sin_addr.s_addr) {
-			if( ds_get_boolean(DS_APPLICATION_ID, DS_APP_NUMERIC_IP) )
-				host = NULL;
-			else
-		    host = gethostbyaddr ((char *)&pduIp->sin_addr,
-					  sizeof (pduIp->sin_addr), AF_INET);
-		    printf("(via %s [%s]) ",
-			    host ? host->h_name : inet_ntoa(pduIp->sin_addr),
-			    inet_ntoa(pduIp->sin_addr));
-		}
-		printf(" %s:\n",
- 		       sprint_objid (oid_buf, pdu->enterprise, pdu->enterprise_length));
-		if (pdu->trap_type == SNMP_TRAP_ENTERPRISESPECIFIC) {
-		    sprint_objid(oid_buf, trapOid, trapOidLen);
-		    cp = strrchr(oid_buf, '.');
-		    if (cp) cp++;
-		    else cp = oid_buf;
-		    printf("\t%s Trap (%s) Uptime: %s\n",
-			   trap_description(pdu->trap_type), cp,
-			   uptime_string(pdu->time, buf));
-		}
-		else
-		    printf("\t%s Trap (%ld) Uptime: %s\n",
-			   trap_description(pdu->trap_type), pdu->specific_type,
-			   uptime_string(pdu->time, buf));
-		for(vars = pdu->variables; vars; vars = vars->next_variable) {
-		    printf ("\t");
-		    print_variable(vars->name, vars->name_length, vars);
-		}
-                printf("\n");
+	        if (trap1_fmt_str == NULL || trap1_fmt_str[0] == '\0')
+	            (void) format_plain_trap (out_bfr, SPRINT_MAX_LEN, pdu);
+	        else
+		    (void) format_trap (out_bfr, SPRINT_MAX_LEN, trap1_fmt_str, pdu);
+                snmp_log(LOG_INFO, out_bfr);
 	    }
 	    if (Syslog && (pdu->trap_type != SNMP_TRAP_AUTHFAIL || dropauth == 0)) {
 	    	varbufidx=0;
@@ -536,24 +510,19 @@ int snmp_input(int op,
 	    host = gethostbyaddr ((char *)&pduIp->sin_addr,
 				  sizeof (pduIp->sin_addr), AF_INET);
 	    if (Print) {
-		time (&timer);
-		tm = localtime (&timer);
-                printf("%.4d-%.2d-%.2d %.2d:%.2d:%.2d %s [%s]:\n",
-		       tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
-		       tm->tm_hour, tm->tm_min, tm->tm_sec,
-                       host ? host->h_name : inet_ntoa(pduIp->sin_addr),
-                       inet_ntoa(pduIp->sin_addr));
-		for (vars = pdu->variables; vars; vars = vars->next_variable) {
-		    printf("\t");
-		    print_variable(vars->name, vars->name_length, vars);
-		}
-		printf("\n");
+	        if (trap2_fmt_str == NULL || trap2_fmt_str[0] == '\0')
+                    (void) format_trap (out_bfr, SPRINT_MAX_LEN,
+                                        trap2_std_str, pdu);
+                else
+                    (void) format_trap (out_bfr, SPRINT_MAX_LEN,
+                                        trap2_fmt_str, pdu);
+                snmp_log(LOG_INFO, out_bfr);
 	    }
 	    if (Syslog) {
 	    	varbufidx=0;
 	    	varbuf[varbufidx]='\0';
 
-	    	for(vars = pdu->variables; vars; vars = vars->next_variable) {
+	    	for (vars = pdu->variables; vars; vars = vars->next_variable) {
 		    sprint_variable(varbuf+varbufidx, vars->name,
 				    vars->name_length, vars);
 		    /* Update the length of the string with the
@@ -611,6 +580,32 @@ int snmp_input(int op,
 }
 
 
+static void parse_trap1_fmt(const char *token, char *line)
+{
+    trap1_fmt_str = strdup(line);
+}
+
+
+static void free_trap1_fmt(void)
+{
+    if (trap1_fmt_str != trap1_std_str) free ((char *)trap1_fmt_str);
+    trap1_fmt_str = NULL;
+}
+
+
+static void parse_trap2_fmt(const char *token, char *line)
+{
+    trap2_fmt_str = strdup(line);
+}
+
+
+static void free_trap2_fmt(void)
+{
+    if (trap2_fmt_str != trap2_std_str) free ((char *)trap2_fmt_str);
+    trap2_fmt_str = NULL;
+}
+
+
 void usage(void)
 {
     fprintf(stderr,"Usage: snmptrapd [-h|-H|-V] [-D] [-p #] [-P] [-s] [-f] [-l [d0-7]] [-e] [-d] [-n] [-a] [-m <MIBS>] [-M <MIBDIRS]\n");
@@ -623,6 +618,7 @@ void usage(void)
   -D[TOKEN,...] turn on debugging output, optionally by the list of TOKENs.\n\
   -p <port> Local port to listen from\n\
   -P        Print to standard output\n\
+  -F \"...\" Use custom format for logging to standard output\n\
   -u PIDFILE create PIDFILE with process id\n\
   -e        Print Event # (rising/falling alarm], etc.\n\
   -s        Log syslog\n\
@@ -662,6 +658,7 @@ int main(int argc, char *argv[])
     int local_port = SNMP_TRAP_PORT;
     int dofork=1;
     char *cp;
+    int tcp=0;
 #if HAVE_GETPID
 	FILE           *PID;
         char *pid_file = NULL;
@@ -683,6 +680,12 @@ int main(int argc, char *argv[])
 			    "username (MD5|SHA) passphrase [DES passphrase]");
     register_config_handler("snmptrapd", "usmUser",
                             usm_parse_config_usmUser, NULL, NULL);
+    register_config_handler("snmptrapd", "format1",
+			    parse_trap1_fmt, free_trap1_fmt,
+			    "format");
+    register_config_handler("snmptrapd", "format2",
+			    parse_trap2_fmt, free_trap2_fmt,
+			    "format");
 
   /* we need to be called back later */
   snmp_register_callback(SNMP_CALLBACK_LIBRARY, SNMP_CALLBACK_STORE_DATA,
@@ -696,7 +699,7 @@ int main(int argc, char *argv[])
     /*
      * usage: snmptrapd [-D] [-u PIDFILE] [-p #] [-P] [-s] [-l [d0-7]] [-d] [-e] [-a]
      */
-    while ((arg = getopt(argc, argv, "VdnqRD:p:m:M:PO:esSafl:Hu:c:C")) != EOF){
+    while ((arg = getopt(argc, argv, "VdnqRD:p:m:M:Po:O:esSafl:Hu:c:CF:T:")) != EOF){
 	switch(arg) {
 	case 'V':
             fprintf(stderr,"UCD-snmp version: %s\n", VersionInfo);
@@ -721,6 +724,16 @@ int main(int argc, char *argv[])
 	case 'M':
 	    setenv("MIBDIRS", optarg, 1);
 	    break;
+        case 'T':
+            if (strcasecmp(optarg,"TCP") == 0) {
+                tcp = 1;
+            } else if (strcasecmp(optarg,"UDP") == 0) {
+                tcp = 0;
+            } else {
+                fprintf(stderr,"Unknown transport \"%s\" after -T flag.\n", optarg);
+                exit(1);
+            }
+            break;
 	case 'O':
 	    cp = snmp_out_toggle_options(optarg);
 	    if (cp != NULL) {
@@ -730,8 +743,17 @@ int main(int argc, char *argv[])
 	    }
 	    break;
 	case 'P':
+            dofork = 0;
+            snmp_enable_stderrlog();
 	    Print++;
 	    break;
+
+        case 'o':
+	    Print++;
+            logfile = optarg;
+            snmp_enable_filelog(optarg, 0);
+            break;
+            
 	case 'e':
 	    Event++;
 	    break;
@@ -798,6 +820,10 @@ int main(int argc, char *argv[])
             ds_set_boolean(DS_APPLICATION_ID, DS_APP_NUMERIC_IP, 1);
             break;
 
+	case 'F':
+	    trap1_fmt_str = optarg;
+	    break;
+
 	default:
 	    fprintf(stderr,"invalid option: -%c\n", arg);
 	    usage();
@@ -814,14 +840,12 @@ int main(int argc, char *argv[])
     if (!Print) Syslog = 1;
 
     /* Initialize the world. Create initial user */
-    usm_set_reportErrorOnUnknownID(1);
-    ds_set_string(DS_LIBRARY_ID, DS_LIB_APPTYPE, "snmptrapd");
     init_snmp("snmptrapd");
 
 #ifndef WIN32
     /* fork the process to the background if we are not printing to stdout */
-    if (!Print && dofork) {
-      int fd, fdnum;
+    if (dofork) {
+      int fd;
 
       switch (fork()) {
 	case -1:
@@ -841,10 +865,6 @@ int main(int argc, char *argv[])
 		dup2(fd, STDOUT_FILENO);
 		dup2(fd, STDERR_FILENO);
 		close(fd);
-                /* Close all unnecessary file descriptors */
-		fdnum = getdtablesize();
-		for ( fd = (STDERR_FILENO + 1); fd < fdnum; fd++ )
-			close(fd);
 		break;
 
 	default:
@@ -872,10 +892,11 @@ int main(int argc, char *argv[])
 	time_t timer;
 	time (&timer);
 	tm = localtime (&timer);
-	printf("%.4d-%.2d-%.2d %.2d:%.2d:%.2d UCD-snmp version %s Started.\n",
-	       tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
-	       tm->tm_hour, tm->tm_min, tm->tm_sec,
-	       VersionInfo);
+        snmp_log(LOG_INFO,
+                 "%.4d-%.2d-%.2d %.2d:%.2d:%.2d UCD-snmp version %s Started.\n",
+                 tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
+                 tm->tm_hour, tm->tm_min, tm->tm_sec,
+                 VersionInfo);
     }
 
 
@@ -893,6 +914,9 @@ int main(int argc, char *argv[])
     session->callback = snmp_input;
     session->callback_magic = NULL;
     session->authenticator = NULL;
+    sess.isAuthoritative = SNMP_SESS_UNKNOWNAUTH;
+    if (tcp)
+        session->flags |= SNMP_FLAGS_STREAM_SOCKET;
 
     SOCK_STARTUP;
     ss = snmp_open( session );
