@@ -158,6 +158,7 @@ int main(int argc, char *argv[])
     int    count;
     int    running;
     int    status;
+    int    exitval = 0;
 
     ds_register_config(ASN_BOOLEAN, "snmpwalk", "includeRequested",
                        DS_APPLICATION_ID, DS_WALK_INCLUDE_REQUESTED);
@@ -165,7 +166,15 @@ int main(int argc, char *argv[])
                        DS_APPLICATION_ID, DS_WALK_PRINT_STATISTICS);
 
     /* get the common command line arguments */
-    arg = snmp_parse_args(argc, argv, &session, "C:", optProc);
+    switch (arg = snmp_parse_args(argc, argv, &session, "C:", optProc)) {
+    case -2:
+    	exit(0);
+    case -1:
+        usage();
+        exit(1);
+    default:
+        break;
+    }
 
     /* get the initial object and subtree */
     if (arg < argc) {
@@ -240,8 +249,8 @@ int main(int argc, char *argv[])
           } else {
             fprintf(stderr, "Error in packet.\nReason: %s\n",
 		    snmp_errstring(response->errstat));
-            if (response->errstat == SNMP_ERR_NOSUCHNAME){
-              fprintf(stderr, "The request for this object identifier failed: ");
+            if (response->errindex != 0){
+              fprintf(stderr, "Failed object: ");
               for(count = 1, vars = response->variables;
                     vars && count != response->errindex;
                     vars = vars->next_variable, count++)
@@ -250,14 +259,17 @@ int main(int argc, char *argv[])
                 fprint_objid(stderr, vars->name, vars->name_length);
               fprintf(stderr, "\n");
             }
+	    exitval = 2;
           }
         }
       } else if (status == STAT_TIMEOUT){
         fprintf(stderr, "Timeout: No Response from %s\n", session.peername);
         running = 0;
+	exitval = 1;
       } else {    /* status == STAT_ERROR */
         snmp_sess_perror("snmpwalk", ss);
         running = 0;
+	exitval = 1;
       }
       if (response)
         snmp_free_pdu(response);
@@ -276,5 +288,5 @@ int main(int argc, char *argv[])
     }
 
     SOCK_CLEANUP;
-    return 0;
+    return exitval;
 }
