@@ -179,7 +179,7 @@ int _asn_bitstring_check(const char * str, u_long asn_length, u_char datum)
 }
 
 /*
- * asn_parse_int - pulls a long out of an ASN int type.
+ * asn_parse_int - pulls a long out of an int type.
  *  On entry, datalength is input as the number of valid bytes following
  *   "data".  On exit, it is returned as the number of valid bytes
  *   following the end of this object.
@@ -229,12 +229,12 @@ asn_parse_int(u_char *data,
     if (*bufp & 0x80)
 	value = -1; /* integer is negative */
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data + asn_length);
+    DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);
 
     while(asn_length--)
 	value = (value << 8) | *bufp++;
 
-    DEBUGMSG(("dump_recv", "  ASN Integer:\t%ld (0x%.2X)\n", value, value));
+    DEBUGMSG(("dumpv_recv", "  Integer:\t%ld (0x%.2X)\n", value, value));
 
     *intp = value;
     return bufp;
@@ -291,12 +291,12 @@ asn_parse_unsigned_int(u_char *data,
     if (*bufp & 0x80)
 	value = ~value; /* integer is negative */
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data + asn_length);
+    DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);
 
     while(asn_length--)
 	value = (value << 8) | *bufp++;
 
-    DEBUGMSG(("dump_recv", "  ASN UInteger:\t%ld (0x%.2X)\n", value, value));
+    DEBUGMSG(("dumpv_recv", "  UInteger:\t%ld (0x%.2X)\n", value, value));
 
     *intp = value;
     return bufp;
@@ -330,10 +330,13 @@ asn_build_int(u_char *data,
 /*
  * ASN.1 integer ::= 0x02 asnlength byte {byte}*
  */
-	static const char *errpre = "build int";
+    static const char *errpre = "build int";
     register long integer;
     register u_long mask;
-
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
+    
     if (intsize != sizeof (long)){
 	_asn_size_err(errpre, intsize, sizeof(long));
 	return NULL;
@@ -363,6 +366,8 @@ asn_build_int(u_char *data,
 	*data++ = (u_char)((integer & mask) >> (8 * (sizeof(long) - 1)));
 	integer <<= 8;
     }
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
+    DEBUGMSG(("dumpv_send", "  Integer:\t%ld (0x%.2X)\n", *intp, *intp));
     return data;
 }
 
@@ -398,6 +403,9 @@ asn_build_unsigned_int(u_char *data,
     register u_long integer;
     register u_long mask;
     int add_null_byte = 0;
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
 
     if (intsize != sizeof (long)){
 	_asn_size_err(errpre, intsize, sizeof(long));
@@ -438,6 +446,8 @@ asn_build_unsigned_int(u_char *data,
 	*data++ = (u_char)((integer & mask) >> (8 * (sizeof(long) - 1)));
 	integer <<= 8;
     }
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
+    DEBUGMSG(("dumpv_send", "  UInteger:\t%ld (0x%.2X)\n", *intp, *intp));
     return data;
 }
 
@@ -487,7 +497,7 @@ asn_parse_string(u_char *data,
 	return NULL;
     }
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data + asn_length);
+    DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);
 
     memmove(string, bufp, asn_length);
     if (*strlength > (int)asn_length)
@@ -495,10 +505,10 @@ asn_parse_string(u_char *data,
     *strlength = (int)asn_length;
     *datalength -= (int)asn_length + (bufp - data);
 
-    DEBUGIF("dump_recv") {
+    DEBUGIF("dumpv_recv") {
       char *buf = (char *)malloc(1+asn_length);
       sprint_asciistring(buf, string, asn_length);
-      DEBUGMSG(("dump_recv", "  ASN String:\t%s\n", buf));
+      DEBUGMSG(("dumpv_recv", "  String:\t%s\n", buf));
       free (buf);
     }
         
@@ -536,6 +546,9 @@ asn_build_string(u_char *data,
  * cmpdstring ::= 0x24 asnlength string {string}*
  * This code will never send a compound string.
  */
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
     data = asn_build_header(data, datalength, type, strlength);
     if (_asn_build_header_check("build string", data, *datalength, strlength))
 	return NULL;
@@ -548,6 +561,13 @@ asn_build_string(u_char *data,
       }
     }
     *datalength -= strlength;
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap + strlength);
+    DEBUGIF("dumpv_send") {
+      char *buf = (char *)malloc(1+strlength);
+      sprint_asciistring(buf, string, strlength);
+      DEBUGMSG(("dumpv_send", "  String:\t%s\n", buf));
+      free (buf);
+    }
     return data + strlength;
 }
 
@@ -587,20 +607,25 @@ asn_parse_header(u_char	*data,
     }
     *type = *bufp;
     bufp = asn_parse_length(bufp + 1, &asn_length);
-    DEBUGDUMPSETUP("dump_recv", data, (bufp-data));
-    DEBUGMSG(("dump_recv", "  ASN Header: 0x%.2X, len = %d (0x%X)\n", *data,
-              asn_length, asn_length));
+
     if (_asn_parse_length_check("parse header", bufp, data, asn_length, *datalength))
 	return NULL;
 
+#ifdef DUMP_PRINT_HEADERS
+    DEBUGDUMPSETUP("recv", data, (bufp-data));
+    DEBUGMSG(("dumpv_recv", "  Header: 0x%.2X, len = %d (0x%X)\n", *data,
+              asn_length, asn_length));
+#else
+/*
+    DEBUGMSGHEXTLI(("recv",data,(bufp-data)));
+    DEBUGMSG(("dumpH_recv","\n"));
+*/
+#endif
+
 #ifdef OPAQUE_SPECIAL_TYPES
 
     if ((*type == ASN_OPAQUE) &&
         (*bufp == ASN_OPAQUE_TAG1)) {
-      DEBUGINDENTMORE();
-      DEBUGDUMPSETUP("dump_recv", data, 1);
-      DEBUGMSG(("dump_recv", "Opaque:\t%.2x\n", *bufp));
-      DEBUGINDENTLESS();
 
       /* check if 64-but counter */
       switch(*(bufp+1)) {
@@ -877,7 +902,7 @@ asn_parse_objid(u_char *data,
 
     *datalength -= (int)asn_length + (bufp - data);
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data + asn_length);
+    DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);
 
     /* Handle invalid object identifier encodings of the form 06 00 robustly */
     if (asn_length == 0)
@@ -929,9 +954,9 @@ asn_parse_objid(u_char *data,
 
     *objidlength = (int)(oidp - objid);
 
-    DEBUGMSG(("dump_recv", "  ASN ObjID: "));
-    DEBUGMSGOID(("dump_recv", objid, *objidlength));
-    DEBUGMSG(("dump_recv", "\n"));
+    DEBUGMSG(("dumpv_recv", "  ObjID: "));
+    DEBUGMSGOID(("dumpv_recv", objid, *objidlength));
+    DEBUGMSG(("dumpv_recv", "\n"));
     return bufp;
 }
 
@@ -972,6 +997,9 @@ asn_build_objid(u_char *data,
     register u_long objid_val;
     u_long first_objid_val;
     register int i;
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
 
     /* check if there are at least 2 sub-identifiers */
     if (objidlength == 0){
@@ -1063,6 +1091,10 @@ asn_build_objid(u_char *data,
 
     /* return the length and data ptr */
     *datalength -= asnlength;
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
+    DEBUGMSG(("dumpv_send", "  ObjID: "));
+    DEBUGMSGOID(("dumpv_send", objid, objidlength));
+    DEBUGMSG(("dumpv_send", "\n"));
     return data;
 }
 
@@ -1105,8 +1137,8 @@ asn_parse_null(u_char *data,
 
     *datalength -= (bufp - data);
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data);
-    DEBUGMSG(("dump_recv", "  ASN NULL\n"));
+    DEBUGDUMPSETUP("recv", data, bufp - data);
+    DEBUGMSG(("dumpv_recv", "  NULL\n"));
 
     return bufp + asn_length;
 }
@@ -1135,7 +1167,13 @@ asn_build_null(u_char *data,
 /*
  * ASN.1 null ::= 0x05 0x00
  */
-    return asn_build_header(data, datalength, type, 0);
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
+    data = asn_build_header(data, datalength, type, 0);
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
+    DEBUGMSG(("dumpv_send", "  NULL\n"));
+    return data;
 }
 
 /*
@@ -1184,9 +1222,10 @@ asn_parse_bitstring(u_char *data,
     if (_asn_bitstring_check(errpre, asn_length, *bufp))
 	return NULL;
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data);
-    DEBUGMSG(("dump_recv", "  ASN Bitstring: "));
-    DEBUGMSGHEX(("dump_recv", data, asn_length));
+    DEBUGDUMPSETUP("recv", data, bufp - data);
+    DEBUGMSG(("dumpv_recv", "  Bitstring: "));
+    DEBUGMSGHEX(("dumpv_recv", data, asn_length));
+    DEBUGMSG(("dumpv_recv", "\n"));
 
     memmove(string, bufp, asn_length);
     *strlength = (int)asn_length;
@@ -1233,9 +1272,47 @@ asn_build_bitstring(u_char *data,
 
     memmove(data, string, strlength);
     *datalength -= strlength;
+    DEBUGDUMPSETUP("send", data, strlength);
+    DEBUGMSG(("dumpv_send", "  Bitstring: "));
+    DEBUGMSGHEX(("dumpv_send", data, strlength));
+    DEBUGMSG(("dumpv_send", "\n"));
     return data + strlength;
 }
 
+#ifdef USE_REVERSE_ASNENCODING
+u_char *
+asn_rbuild_bitstring(u_char *data,
+		    size_t *datalength,
+		    u_char type,
+		    u_char *string,
+		    size_t strlength)
+{
+/*
+ * ASN.1 bit string ::= 0x03 asnlength unused {byte}*
+ */
+    static const char *errpre = "build bitstring";
+    u_char *initdatap = data;
+
+    if (strlength > *datalength)
+        return NULL;
+    data -= strlength;
+    memcpy(data+1, string, strlength);
+    *datalength -= strlength;
+    
+    data = asn_rbuild_header(data, datalength, type, strlength);
+    if (_asn_build_header_check(errpre, data+1, *datalength, strlength))
+	return NULL;
+
+    DEBUGDUMPSETUP("send", data+1, initdatap - data);
+    DEBUGIF("dumpv_send") {
+      char *buf = (char *)malloc(2*strlength);
+      sprint_asciistring(buf, string, strlength);
+      DEBUGMSG(("dumpv_send", "  Bitstring:\t%s\n", buf));
+      free (buf);
+    }
+    return data;
+}
+#endif /* USE_REVERSE_ASNENCODING */
 
 /*
  * asn_parse_unsigned_int64 - pulls a 64 bit unsigned long out of an ASN int
@@ -1280,7 +1357,7 @@ asn_parse_unsigned_int64(u_char *data,
     if (_asn_parse_length_check(errpre, bufp, data, asn_length, *datalength))
         return NULL;
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data);
+    DEBUGDUMPSETUP("recv", data, bufp - data);
 #ifdef OPAQUE_SPECIAL_TYPES
 /* 64 bit counters as opaque */
     if ((*type == ASN_OPAQUE) &&
@@ -1288,8 +1365,6 @@ asn_parse_unsigned_int64(u_char *data,
 	    (*bufp == ASN_OPAQUE_TAG1) &&
 	    ((*(bufp+1) == ASN_OPAQUE_COUNTER64) ||
              (*(bufp+1) == ASN_OPAQUE_U64))) {
-        DEBUGMSG(("dump_recv", "Opaque %.2x %.2x: ", *bufp, *(bufp+1)));
-
 	/* change type to Counter64 or U64 */
         *type = *(bufp+1);
         /* value is encoded as special format */
@@ -1318,9 +1393,10 @@ asn_parse_unsigned_int64(u_char *data,
     cp->low = low;
     cp->high = high;
 
-    DEBUGIF("dump_recv") {
+    DEBUGIF("dumpv_recv") {
       char i64buf[I64CHARSZ+1];
       printU64(i64buf, cp);
+      DEBUGMSG(("dumpv_recv", "Counter64: ", i64buf));
     }
 
     return bufp;
@@ -1359,6 +1435,9 @@ asn_build_unsigned_int64(u_char *data,
     register u_long mask, mask2;
     int add_null_byte = 0;
     size_t intsize;
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
 
   if (countersize != sizeof(struct counter64)){
     _asn_size_err("build uint64", countersize, sizeof(struct counter64));
@@ -1441,6 +1520,12 @@ asn_build_unsigned_int64(u_char *data,
 	low <<= 8;
 	
     }
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
+    DEBUGIF("dumpv_send") {
+      char i64buf[I64CHARSZ+1];
+      printU64(i64buf, cp);
+      DEBUGMSG(("dumpv_send", i64buf));
+    }
     return data;
 }
 
@@ -1479,12 +1564,11 @@ asn_parse_signed_int64(u_char *data,
   if (_asn_parse_length_check(errpre, bufp, data, asn_length, *datalength))
         return NULL;
 
-  DEBUGDUMPSETUP("dump_recv", data, bufp - data);
+  DEBUGDUMPSETUP("recv", data, bufp - data);
   if ((*type == ASN_OPAQUE) &&
       (asn_length <= ASN_OPAQUE_COUNTER64_MX_BER_LEN) &&
       (*bufp == ASN_OPAQUE_TAG1) &&
        (*(bufp+1) == ASN_OPAQUE_I64)) {
-      DEBUGMSG(("dump_recv", "Opaque %.2x %.2x: ", *bufp, *(bufp+1)));
     /* change type to Int64 */
     *type = *(bufp+1);
     /* value is encoded as special format */
@@ -1519,9 +1603,10 @@ asn_parse_signed_int64(u_char *data,
   cp->low = low;
   cp->high = high;
 
-  DEBUGIF("dump_recv") {
+  DEBUGIF("dumpv_recv") {
     char i64buf[I64CHARSZ+1];
     printI64(i64buf, cp);
+    DEBUGMSG(("dumpv_recv", "Integer64: %s", i64buf));
   }
 
   return bufp;
@@ -1552,6 +1637,9 @@ asn_build_signed_int64(u_char *data,
     register u_int mask, mask2;
     u_long low, high;
     size_t intsize;
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
 
   if (countersize != sizeof(struct counter64)){
     _asn_size_err("build int64", countersize, sizeof(struct counter64));
@@ -1596,9 +1684,16 @@ asn_build_signed_int64(u_char *data,
 	    | ((low & mask) >> (8 * (sizeof(u_int) - 1)));
 	low <<= 8;
     }
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
+    DEBUGIF("dumpv_send") {
+      char i64buf[I64CHARSZ+1];
+      printU64(i64buf, cp);
+      DEBUGMSG(("dumpv_send", i64buf));
+    }
     return data;
 }
 
+
 /*
  * asn_parse_float - pulls a single precision floating-point out of an opaque type.
  *
@@ -1642,13 +1737,12 @@ asn_parse_float(u_char *data,
                   asn_length, *datalength))
         return NULL;
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data + asn_length);
+    DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);
 /* the float is encoded as an opaque */
     if ((*type == ASN_OPAQUE) &&
             (asn_length == ASN_OPAQUE_FLOAT_BER_LEN) &&
 	    (*bufp == ASN_OPAQUE_TAG1) &&
 	    (*(bufp+1) == ASN_OPAQUE_FLOAT)) {
-      DEBUGMSG(("dump_recv", "Opaque %.2x %.2x: ", *bufp, *(bufp+1)));
 
         /* value is encoded as special format */
 	bufp = asn_parse_length(bufp + 2, &asn_length);
@@ -1673,7 +1767,7 @@ asn_parse_float(u_char *data,
 
     *floatp =  fu.floatVal;
 
-    DEBUGMSG(("dump_recv", "%f",*floatp));
+    DEBUGMSG(("dumpv_recv", "Opaque float: %f\n",*floatp));
     return bufp;
 }
 
@@ -1708,6 +1802,9 @@ asn_build_float(u_char *data,
         int    intVal;
 	u_char c[sizeof(float)];
     } fu;
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
 
     if (floatsize != sizeof (float)) {
 	_asn_size_err("build float", floatsize, sizeof(float));
@@ -1734,10 +1831,58 @@ asn_build_float(u_char *data,
     *datalength -= floatsize;
     memcpy(data, &fu.c[0], floatsize);
 
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
+    DEBUGMSG(("dumpv_send", "Opaque float: %f\n",*floatp));
     data += floatsize;
     return data;
 }
 
+#ifdef USE_REVERSE_ASNENCODING
+u_char *
+asn_rbuild_float(u_char *data,
+                 size_t *datalength,
+                 u_char type,
+                 float *floatp,
+                 size_t floatsize)
+{
+    u_char *initdatap = data;
+    union {
+        float  floatVal;
+        int    intVal;
+	u_char c[sizeof(float)];
+    } fu;
+
+    /* floatsize better not be larger than realistic */
+    if (floatsize != sizeof(float) ||
+        *datalength < (floatsize + 3) ||
+        floatsize > 122)
+        return NULL;
+
+    /* correct for endian differences */
+    fu.floatVal = *floatp;
+    fu.intVal = htonl(fu.intVal);	
+    *datalength -= floatsize + 3;
+    data -= floatsize;
+    memcpy(data+1, &fu.c[0], floatsize);
+
+    /* put the special tag and length (3 bytes) */
+    *data-- = (u_char)floatsize;
+    *data-- = ASN_OPAQUE_FLOAT;
+    *data-- = ASN_OPAQUE_TAG1;
+
+    /* put the tag and length for the Opaque wrapper */
+    data = asn_rbuild_header(data, datalength, ASN_OPAQUE, floatsize+3);
+    if (_asn_build_header_check("build float", data, *datalength, (floatsize+3)))
+	return NULL;
+
+    DEBUGDUMPSETUP("send", data+1, initdatap - data);
+    DEBUGMSG(("dumpv_send", "Opaque Float:\t%f\n", *floatp));
+    return data;
+}
+
+
+#endif /* USE_REVERSE_ASNENCODING */
+
 /*
 
   u_char * asn_parse_double(
@@ -1774,13 +1919,12 @@ asn_parse_double(u_char *data,
                   asn_length, *datalength))
         return NULL;
 
-    DEBUGDUMPSETUP("dump_recv", data, bufp - data + asn_length);
+    DEBUGDUMPSETUP("recv", data, bufp - data + asn_length);
 /* the double is encoded as an opaque */
     if ((*type == ASN_OPAQUE) &&
             (asn_length == ASN_OPAQUE_DOUBLE_BER_LEN) &&
 	    (*bufp == ASN_OPAQUE_TAG1) &&
 	    (*(bufp+1) == ASN_OPAQUE_DOUBLE)) {
-      DEBUGMSG(("dump_recv", "Opaque %.2x %.2x: ", *bufp, *(bufp+1)));
 
         /* value is encoded as special format */
 	bufp = asn_parse_length(bufp + 2, &asn_length);
@@ -1806,7 +1950,7 @@ asn_parse_double(u_char *data,
     fu.intVal[1] = tmp;
     	
     *doublep =  fu.doubleVal;
-    DEBUGMSG(("dump_recv", "%d",*doublep));
+    DEBUGMSG(("dumpv_recv", "  Opaque Double:\t%f\n",*doublep));
 
     return bufp;
 }
@@ -1833,6 +1977,9 @@ asn_build_double(u_char *data,
 	int    intVal[2];
 	u_char c[sizeof(double)];
     } fu;
+#ifndef SNMP_NO_DEBUGGING
+    u_char *initdatap = data;
+#endif
 
     if (doublesize != sizeof(double)){
 	_asn_size_err("build double", doublesize, sizeof(double));
@@ -1862,8 +2009,518 @@ asn_build_double(u_char *data,
     memcpy(data, &fu.c[0], doublesize);
 
     data += doublesize;
+    DEBUGDUMPSETUP("send", initdatap, data - initdatap);
+    DEBUGMSG(("dumpv_send", "  Opaque double: %f", *doublep));
     return data;
 }
 
+#ifdef USE_REVERSE_ASNENCODING
+u_char *
+asn_rbuild_double(u_char *data,
+                 size_t *datalength,
+                 u_char type,
+                 double *doublep,
+                 size_t doublesize)
+{
+    u_char *initdatap = data;
+    long  tmp;
+    union {
+        double doubleVal;
+	int    intVal[2];
+	u_char c[sizeof(double)];
+    } fu;
+
+    /* doublesize better not be larger than realistic */
+    if (doublesize != sizeof(double) ||
+        *datalength < (doublesize + 3) ||
+        doublesize > 122)
+        return NULL;
+
+    /* correct for endian differences */
+    fu.doubleVal = *doublep;
+    tmp = htonl(fu.intVal[0]);
+    fu.intVal[0] = htonl(fu.intVal[1]);	
+    fu.intVal[1] = tmp;
+
+    *datalength -= doublesize + 3;
+    data -= doublesize;
+    memcpy(data+1, &fu.c[0], doublesize);
+
+    /* put the special tag and length (3 bytes) */
+    *data-- = (u_char)doublesize;
+    *data-- = ASN_OPAQUE_DOUBLE;
+    *data-- = ASN_OPAQUE_TAG1;
+
+    /* put the tag and length for the Opaque wrapper */
+    data = asn_rbuild_header(data, datalength, ASN_OPAQUE, doublesize+3);
+    if (_asn_build_header_check("build double", data, *datalength, (doublesize+3)))
+	return NULL;
+
+    DEBUGDUMPSETUP("send", data+1, initdatap - data);
+    DEBUGMSG(("dumpv_send", "  Opaque Double:\t%f\n", *doublep));
+    return data;
+}
+
+#endif /* USE_REVERSE_ASNENCODING */
 #endif /* OPAQUE_SPECIAL_TYPES */
 
+#ifdef USE_REVERSE_ASNENCODING
+u_char	*
+asn_rbuild_int (u_char *data,
+                size_t *datalength,
+                u_char type,
+                long *intp,
+                size_t intsize)
+{
+    static const char *errpre = "build int";
+    register long integer;
+    u_char *initdatap = data;
+    int testvalue = (*intp < 0) ? -1 : 0;
+    
+    if (intsize != sizeof (long)){
+	_asn_size_err(errpre, intsize, sizeof(long));
+	return NULL;
+    }
+    integer = *intp;
+
+    if (((*datalength)--) < 1) return NULL;
+    *data-- = integer & 0xff;
+    integer >>= 8;
+    while (integer != testvalue) {
+        if (((*datalength)--) < 1) return NULL;
+        *data-- = integer & 0xff;
+        integer >>= 8;
+    }
+    if ((*(data+1) & 0x80) != (testvalue & 0x80)) {
+        /* make sure left most bit is representational of the rest of
+           the bits that aren't encoded */
+        if (((*datalength)--) < 1) return NULL;
+        *data-- = testvalue & 0xff;
+    }
+
+    intsize = initdatap-data;
+    data = asn_rbuild_header(data, datalength, type, intsize);
+    if (_asn_build_header_check(errpre, data+1, *datalength, intsize))
+	return NULL;
+
+    DEBUGDUMPSETUP("send", data+1, intsize);
+    DEBUGMSG(("dumpv_send", "  Integer:\t%ld (0x%.2X)\n", *intp, *intp));
+    return data;
+}
+
+u_char	*
+asn_rbuild_string (u_char *data,
+                   size_t *datalength,
+                   u_char	type,
+                   const u_char *string,
+                   size_t strlength)
+{
+    static const char *errpre = "build string";
+    u_char *initdatap = data;
+
+    if (strlength > *datalength)
+        return NULL;
+    data -= strlength;
+    memcpy(data+1, string, strlength);
+    *datalength -= strlength;
+    
+    data = asn_rbuild_header(data, datalength, type, strlength);
+    if (_asn_build_header_check(errpre, data+1, *datalength, strlength))
+	return NULL;
+
+    DEBUGDUMPSETUP("send", data+1, initdatap - data);
+    DEBUGIF("dumpv_send") {
+        if (strlength == 0) {
+            DEBUGMSG(("dumpv_send", "  String: [NULL]\n"));
+        } else {
+            char *buf = (char *)malloc(2*strlength);
+            sprint_asciistring(buf, string, strlength);
+            DEBUGMSG(("dumpv_send", "  String:\t%s\n", buf));
+            free (buf);
+        }
+    }
+    return data;
+}
+
+u_char	*
+asn_rbuild_unsigned_int (u_char *data,
+		       size_t *datalength,
+		       u_char type,
+		       u_long *intp,
+		       size_t intsize)
+{
+    static const char *errpre = "build uint";
+    register u_long integer;
+    u_char *initdatap = data;
+    
+    if (intsize != sizeof (long)){
+	_asn_size_err(errpre, intsize, sizeof(long));
+	return NULL;
+    }
+    integer = *intp;
+
+    if (((*datalength)--) < 1) return NULL;
+    *data-- = integer & 0xff;
+    integer >>= 8;
+    while (integer != 0) {
+        if (((*datalength)--) < 1) return NULL;
+        *data-- = integer & 0xff;
+        integer >>= 8;
+    }
+    if ((*(data+1) & 0x80) != (0 & 0x80)) {
+        /* make sure left most bit is representational of the rest of
+           the bits that aren't encoded */
+        if (((*datalength)--) < 1) return NULL;
+        *data-- = 0;
+    }
+
+    intsize = initdatap-data;
+    data = asn_rbuild_header(data, datalength, type, intsize);
+    if (_asn_build_header_check(errpre, data+1, *datalength, intsize))
+	return NULL;
+
+    DEBUGDUMPSETUP("send", data+1, intsize);
+    DEBUGMSG(("dumpv_send", "  UInteger:\t%ld (0x%.2X)\n", *intp, *intp));
+    return data;
+}
+
+u_char	*
+asn_rbuild_header (u_char *data,
+                   size_t *datalength,
+                   u_char type,
+                   size_t length)
+{
+    char ebuf[128];
+    
+    data = asn_rbuild_length(data, datalength, length);
+    if (*datalength < 1){
+	sprintf(ebuf, "bad header length < 1 :%d, %d", *datalength, length);
+	ERROR_MSG(ebuf);
+	return NULL;
+    }
+    *data-- = type;
+    (*datalength)--;
+    return data;
+}
+
+u_char	*
+asn_rbuild_sequence (u_char *data,
+                     size_t *datalength,
+                     u_char type,
+                     size_t length)
+{
+    return asn_rbuild_header(data, datalength, type, length);
+}
+
+u_char	*
+asn_rbuild_length (u_char *data,
+		 size_t *datalength,
+		 size_t length)
+{
+    static const char *errpre = "build length";
+    char ebuf[128];
+    
+    u_char    *start_data = data;
+
+    if (length <= 0x7f) {
+        if ((*datalength)-- == 0) {
+	    sprintf(ebuf, "%s: bad length < 1 :%d, %d",errpre,*datalength,length);
+	    ERROR_MSG(ebuf);
+            return NULL;
+        }
+        *data-- = length;
+    } else {
+        while(length > 0xff) {
+            if ((*datalength)-- == 0) {
+                sprintf(ebuf, "%s: bad length < 1 :%d, %d",errpre,*datalength,length);
+                ERROR_MSG(ebuf);
+                return NULL;
+            }
+            *data-- = length & 0xff;
+            length >>= 8;
+        }
+        if (*datalength <= 1) {
+            sprintf(ebuf, "%s: bad length < 1 :%d, %d",errpre,*datalength,length);
+            ERROR_MSG(ebuf);
+            return NULL;
+        }
+        *data-- = length & 0xff;
+        *data-- = (start_data - data) | 0x80;
+        *datalength -= 2;
+    }
+    return data;
+}
+
+u_char	*
+asn_rbuild_objid (u_char *data,
+		size_t *datalength,
+		u_char type,
+		oid *objid,
+		size_t objidlength)
+{
+/*
+ * ASN.1 objid ::= 0x06 asnlength subidentifier {subidentifier}*
+ * subidentifier ::= {leadingbyte}* lastbyte
+ * leadingbyte ::= 1 7bitvalue
+ * lastbyte ::= 0 7bitvalue
+ */
+    register oid *op = objid;
+    register int i, tmpint;
+    u_char *initdatap = data;
+    const char *errpre = "build objid";
+
+    /* check if there are at least 2 sub-identifiers */
+    if (objidlength == 0) {
+        /* there are not, so make OID have two with value of zero */
+        if (*datalength < 2) {
+            return NULL;
+        }
+        *data-- = 0;
+        *data-- = 0;
+        *datalength -= 2;
+    } else if (objidlength == 1) {
+        /* encode the first value */
+        if ((*datalength)-- < 1) return NULL;
+        *data-- = objid[0];
+    } else {
+        for(i = objidlength; i > 2; i--) {
+            tmpint = objid[i-1];
+            if ((*datalength)-- < 1) return NULL;
+            *data-- = tmpint & 0x7f;
+            tmpint >>= 7;
+
+            while (tmpint > 0) {
+                if ((*datalength)-- < 1) return NULL;
+                *data-- = ((tmpint & 0x7f) | 0x80);
+                tmpint >>= 7;
+            }
+        }
+
+        /* combine the first two values */
+	if ( op[1] > 40 ) {
+	    ERROR_MSG("build objid: bad second subidentifier");
+	    return NULL;
+	}
+        if ((*datalength)-- < 1) return NULL;
+	*data-- = (op[0] * 40) + op[1];
+    }
+    
+    tmpint = initdatap-data;
+    data = asn_rbuild_header(data, datalength, type, tmpint);
+    if (_asn_build_header_check(errpre, data+1, *datalength, tmpint))
+	return NULL;
+
+    /* return the length and data ptr */
+    DEBUGDUMPSETUP("send", data+1, tmpint);
+    DEBUGMSG(("dumpv_send", "  ObjID: "));
+    DEBUGMSGOID(("dumpv_send", objid, objidlength));
+    DEBUGMSG(("dumpv_send", "\n"));
+    return data;
+}
+
+u_char	*asn_rbuild_null (u_char *data,
+                          size_t *datalength,
+                          u_char type)
+{
+/*
+ * ASN.1 null ::= 0x05 0x00
+ */
+    u_char *initdatap = data;
+    
+    data = asn_rbuild_header(data, datalength, type, 0);
+    DEBUGDUMPSETUP("send", data+1, initdatap - data);
+    DEBUGMSG(("dumpv_send", "  NULL\n"));
+    return data;
+}
+
+
+u_char	*asn_rbuild_unsigned_int64 (u_char *data,
+			 size_t *datalength,
+			 u_char type,
+			 struct counter64 *cp,
+			 size_t countersize)
+{
+/*
+ * ASN.1 integer ::= 0x02 asnlength byte {byte}*
+ */
+
+    register u_long low, high;
+    size_t intsize;
+    u_char *initdatap = data;
+    int count;
+    
+  if (countersize != sizeof(struct counter64)){
+    _asn_size_err("build uint64", countersize, sizeof(struct counter64));
+    return NULL;
+  }
+
+  low = cp->low;
+  high = cp->high;
+  
+  /* encode the low 4 bytes first */
+  if (((*datalength)--) < 1) return NULL;
+  *data-- = low & 0xff;
+  low >>= 8;
+  count = 1;
+  while (low != 0) {
+      count++;
+      if (((*datalength)--) < 1) return NULL;
+      *data-- = low & 0xff;
+      low >>= 8;
+  }
+
+  /* then the high byte if present */
+  if (high) {
+      /* do the rest of the low byte */
+      for(; count < 4; count++) {
+          if (((*datalength)--) < 1) return NULL;
+          *data-- = 0;
+      }
+
+      /* do high byte */
+      if (((*datalength)--) < 1) return NULL;
+      *data-- = high & 0xff;
+      high >>= 8;
+      while (high != 0) {
+          if (((*datalength)--) < 1) return NULL;
+          *data-- = high & 0xff;
+          high >>= 8;
+      }
+  }
+
+  if ((*(data+1) & 0x80) != (0 & 0x80)) {
+      /* make sure left most bit is representational of the rest of
+         the bits that aren't encoded */
+      if (((*datalength)--) < 1) return NULL;
+      *data-- = 0;
+  }
+  
+  intsize = initdatap-data;
+
+#ifdef OPAQUE_SPECIAL_TYPES
+  if (*datalength < 5) return NULL;
+  *datalength -= 3;
+  *data-- = (u_char)intsize;
+
+  /* encode a Counter64 as an opaque (it also works in SNMPv1) */
+  /* turn into Opaque holding special tagged value */
+  if (type == ASN_OPAQUE_COUNTER64) {
+      *data-- = ASN_OPAQUE_COUNTER64;
+      *data-- = ASN_OPAQUE_TAG1;
+      
+      /* put the tag and length for the Opaque wrapper */
+      data = asn_rbuild_header(data, datalength, ASN_OPAQUE, intsize+3);
+      if (_asn_build_header_check("build counter u64", data+1, *datalength, intsize+3))
+          return NULL;
+
+  } else if (type == ASN_OPAQUE_U64) {
+      /* Encode the Unsigned int64 in an opaque */
+      /* turn into Opaque holding special tagged value */
+
+      *data-- = ASN_OPAQUE_U64;
+      *data-- = ASN_OPAQUE_TAG1;
+      
+      /* put the tag and length for the Opaque wrapper */
+      data = asn_rbuild_header(data, datalength, ASN_OPAQUE, intsize+3);
+      if (_asn_build_header_check("build opaque u64", data+1, *datalength, intsize+3))
+          return NULL;
+    } else {
+#endif /* OPAQUE_SPECIAL_TYPES */
+    data = asn_rbuild_header(data, datalength, type, intsize);
+    if (_asn_build_header_check("build uint64", data+1, *datalength, intsize))
+	return NULL;
+#ifdef OPAQUE_SPECIAL_TYPES
+    }
+#endif /* OPAQUE_SPECIAL_TYPES */
+
+  DEBUGDUMPSETUP("send", data+1, intsize);
+  DEBUGMSG(("dumpv_send", "  U64:\t%ld %ld\n", cp->high, cp->low));
+  return data;
+}
+
+#ifdef OPAQUE_SPECIAL_TYPES
+u_char	*asn_rbuild_signed_int64 (u_char *data,
+                                  size_t *datalength,
+                                  u_char type,
+                                  struct counter64 *cp,
+                                  size_t countersize)
+{
+/*
+ * ASN.1 integer ::= 0x02 asnlength byte {byte}*
+ */
+
+    register u_long low, high;
+    size_t intsize;
+    u_char *initdatap = data;
+    int count;
+    int testvalue;
+    
+  if (countersize != sizeof(struct counter64)){
+    _asn_size_err("build uint64", countersize, sizeof(struct counter64));
+    return NULL;
+  }
+
+  low = cp->low;
+  high = cp->high;
+
+  testvalue = (high & 0x80000000) ? -1 : 0;
+  
+  /* encode the low 4 bytes first */
+  if (((*datalength)--) < 1) return NULL;
+  *data-- = low & 0xff;
+  low >>= 8;
+  count = 1;
+  while (low != testvalue) {
+      count++;
+      if (((*datalength)--) < 1) return NULL;
+      *data-- = low & 0xff;
+      low >>= 8;
+  }
+
+  /* then the high byte if present */
+  if (high) {
+      /* do the rest of the low byte */
+      for(; count < 4; count++) {
+          if (((*datalength)--) < 1) return NULL;
+          *data-- = (testvalue == 0) ? 0 : 0xff;
+      }
+
+      /* do high byte */
+      if (((*datalength)--) < 1) return NULL;
+      *data-- = high & 0xff;
+      high >>= 8;
+      while (high != testvalue) {
+          if (((*datalength)--) < 1) return NULL;
+          *data-- = high & 0xff;
+          high >>= 8;
+      }
+  }
+
+  if ((*(data+1) & 0x80) != (0 & 0x80)) {
+      /* make sure left most bit is representational of the rest of
+         the bits that aren't encoded */
+      if (((*datalength)--) < 1) return NULL;
+      *data-- = (testvalue == 0) ? 0 : 0xff;
+  }
+  
+  intsize = initdatap-data;
+
+  if (*datalength < 5) return NULL;
+  *datalength -= 3;
+  *data-- = (u_char)intsize;
+  *data-- = ASN_OPAQUE_I64;
+  *data-- = ASN_OPAQUE_TAG1;
+      
+  /* put the tag and length for the Opaque wrapper */
+  data = asn_rbuild_header(data, datalength, ASN_OPAQUE, intsize+3);
+  if (_asn_build_header_check("build counter u64", data+1, *datalength, intsize+3))
+      return NULL;
+
+  DEBUGDUMPSETUP("send", data+1, intsize);
+  DEBUGMSG(("dumpv_send", "  UInt64:\t%ld %ld\n", cp->high, cp->low)); /* WWW */
+  return data;
+}
+#endif /* OPAQUE_SPECIAL_TYPES */
+
+u_char	*asn_rbuild_bitstring (u_char *, size_t *, u_char, u_char *, size_t);
+#endif
