@@ -76,7 +76,7 @@
 #include "read_config.h"
 #include "tools.h"
 
-int config_errors;
+static int config_errors;
 
 struct config_files *config_files = NULL;
 
@@ -138,13 +138,11 @@ register_config_handler(const char *type_param,
 
   if (*ctmp == NULL) {
     *ctmp = (struct config_files *)
-      malloc(sizeof(struct config_files));
+      calloc(1,sizeof(struct config_files));
     if ( !*ctmp ) {
       return NULL;
     }
 
-    (*ctmp)->next		 = NULL;
-    (*ctmp)->start		 = NULL;
     (*ctmp)->fileHeader	 = strdup(type);
   }
 
@@ -160,19 +158,15 @@ register_config_handler(const char *type_param,
 
   if (*ltmp == NULL) {
     *ltmp = (struct config_line *)
-      malloc(sizeof(struct config_line));
+      calloc(1,sizeof(struct config_line));
     if ( !*ltmp ) {
       return NULL;
     }
 
-    (*ltmp)->next		 = NULL;
     (*ltmp)->config_time	 = NORMAL_CONFIG;
-    (*ltmp)->parse_line	 = 0;
-    (*ltmp)->free_func	 = 0;
     (*ltmp)->config_token	 = strdup(token);
     if (help != NULL)
       (*ltmp)->help = strdup(help);
-
   }
 
   /* 
@@ -223,10 +217,11 @@ unregister_config_handler(const char *type_param,
   }
   if (strcmp((*ltmp)->config_token,token) == 0) {
     /* found it at the top of the list */
-    (*ctmp)->start = (*ltmp)->next;
+    ltmp2 = (*ltmp)->next;
     free((*ltmp)->config_token);
     SNMP_FREE((*ltmp)->help);
     free(*ltmp);
+    (*ctmp)->start = ltmp2;
     return;
   }
   while ((*ltmp)->next != NULL && strcmp((*ltmp)->next->config_token,token)) {
@@ -247,6 +242,27 @@ unregister_app_config_handler(const char *token)
   unregister_config_handler( NULL, token );
 }
 
+void 
+unregister_all_config_handlers()
+{
+  struct config_files *ctmp, *save;
+  struct config_line *ltmp;
+
+  free_config();
+
+  /* Keep using config_files until there are no more! */
+  for (ctmp = config_files; ctmp;) {
+      for (ltmp = ctmp->start; ltmp; ltmp = ctmp->start) {
+          unregister_config_handler (ctmp->fileHeader, ltmp->config_token);
+      }
+      free (ctmp->fileHeader);
+      save = ctmp->next;
+      free (ctmp);
+      ctmp = save;
+      config_files = save;
+  }
+}
+
 #ifdef TESTING
 void print_config_handlers (void)
 {
@@ -264,13 +280,21 @@ void print_config_handlers (void)
 int linecount;
 const char *curfilename;
 
+struct config_line *
+read_config_get_handlers(const char *type) {
+    struct config_files *ctmp = config_files;
+    for(;ctmp != NULL && strcmp(ctmp->fileHeader,type); ctmp = ctmp->next);
+    if (ctmp)
+        return ctmp->start;
+    return NULL;
+}
+
 void read_config_with_type(const char *filename, 
 			   const char *type)
 {
-  struct config_files *ctmp = config_files;
-  for(;ctmp != NULL && strcmp(ctmp->fileHeader,type); ctmp = ctmp->next);
+  struct config_line *ctmp = read_config_get_handlers(type);
   if (ctmp)
-    read_config(filename, ctmp->start, EITHER_CONFIG);
+    read_config(filename, ctmp, EITHER_CONFIG);
   else
     DEBUGMSGTL(("read_config", "read_config: I have no registrations for type:%s,file:%s\n",
            type, filename));
@@ -344,12 +368,34 @@ void read_config(const char *filename,
       if ((cptr = skip_white(cptr)))
 	{
           cptr = copy_word(cptr,token);
+          if (token[0] == '[') {
+              token[strlen(token)-1] = '\0';
+              lptr = read_config_get_handlers(&token[1]);
+              if (lptr == NULL) {
+                  sprintf(tmpbuf,"No handlers regestered for type %s.",
+                          &token[1]);
+                  config_perror(tmpbuf);
+                  continue;
+              }
+              DEBUGMSGTL(("read_config","Switching to new context: %s%s\n",
+                          ((cptr)?"(this line only) ":""),&token[1]));
+              if (cptr == NULL) {
+                  /* change context permanently */
+                  line_handler = lptr;
+                  continue;
+              } else {
+                  /* the rest of this line only applies. */
+                  cptr = copy_word(cptr,token);
+              }
+          } else {
+              lptr = line_handler;
+          }
           if (cptr == NULL) {
             sprintf(tmpbuf,"Blank line following %s token.", token);
             config_perror(tmpbuf);
           } else {
-            for(lptr = line_handler, done=0;
-                lptr != NULL && !done;
+              
+            for(done=0; lptr != NULL && !done;
                 lptr = lptr->next) {
               if (!strcasecmp(token,lptr->config_token)) {
                 if (when == EITHER_CONFIG || lptr->config_time == when) {
@@ -400,8 +446,18 @@ read_configs (void)
     read_config_files(NORMAL_CONFIG);
 
  /* do this even when the normal above wasn't done */
-  if (optional_config && type)
-    read_config_with_type(optional_config, type);
+  if (optional_config && type) {
+      struct stat statbuf;
+      if (stat(optional_config, &statbuf)) {
+          DEBUGMSGTL(("read_config","Optional File \"%s\" does not exist.\n",
+                      optional_config));
+          snmp_log_perror(optional_config);
+      } else {
+          DEBUGMSGTL(("read_config","Reading optional config file: \"%s\"\n",
+                      optional_config));
+          read_config_with_type(optional_config, type);
+      }
+  }
 
   snmp_call_callbacks(SNMP_CALLBACK_LIBRARY, SNMP_CALLBACK_POST_READ_CONFIG,
                       NULL);
@@ -464,6 +520,8 @@ read_config_files (int when)
   struct config_files *ctmp = config_files;
   struct config_line *ltmp;
   struct stat statbuf;
+
+  config_errors = 0;
   
   if (when == PREMIB_CONFIG)
     free_config();
@@ -531,7 +589,7 @@ read_config_files (int when)
   }
   
   if (config_errors) {
-    snmp_log(LOG_ERR, "ucd-snmp: %d errors in config file\n", config_errors);
+    snmp_log(LOG_ERR, "ucd-snmp: %d error(s) in config file(s)\n", config_errors);
 /*    exit(1); */
   }
 }
@@ -548,9 +606,21 @@ void read_config_print_usage(const char *lead)
     snmp_log(LOG_INFO, "%sIn %s.conf and %s.local.conf:\n", lead, ctmp->fileHeader,
             ctmp->fileHeader);
     for(ltmp = ctmp->start; ltmp != NULL; ltmp = ltmp->next) {
-      if (ltmp->help != NULL)
-        snmp_log(LOG_INFO, "%s%s%-15s %s\n", lead, lead, ltmp->config_token,
-                 ltmp->help);
+        DEBUGIF("read_config_usage") {
+            if (ltmp->config_time == PREMIB_CONFIG)
+                DEBUGMSG(("read_config_usage", "*"));
+            else
+                DEBUGMSG(("read_config_usage", " "));
+        }
+        if (ltmp->help) {
+            snmp_log(LOG_INFO, "%s%s%-15s %s\n", lead, lead,
+                     ltmp->config_token, ltmp->help);
+        } else {
+            DEBUGIF("read_config_usage") {
+                snmp_log(LOG_INFO, "%s%s%-15s [NO HELP]\n", lead, lead,
+                         ltmp->config_token);
+            }
+        }
     }
   }
 }
@@ -644,7 +714,7 @@ read_app_config_store(const char *line)
 void
 snmp_save_persistent(const char *type)
 {
-  char file[512], fileold[512];
+  char file[512], fileold[SPRINT_MAX_LEN];
   struct stat statbuf;
   int j;
 
@@ -663,6 +733,9 @@ snmp_save_persistent(const char *type)
       }
     }
   }
+  /* save a warning header to the top of the new file */
+  sprintf(fileold, "#\n# net-snmp (or ucd-snmp) persistent data file.\n#\n# DO NOT STORE CONFIGURATION ENTRIES HERE.\n# Please save normal configuration tokens for %s in SNMPCONFPATH/%s.conf.\n# Only \"createUser\" tokens should be placed here by %s administrators.\n#\n", type, type, type);
+  read_config_store(type, fileold);
 }
 
 
@@ -710,20 +783,19 @@ snmp_clean_persistent(const char *type)
    line number of a .conf file and increments the error count. */
 void config_perror(const char *string)
 {
-  config_pwarn(string);
+  snmp_log(LOG_ERR, "%s: line %d: Error: %s\n", curfilename, linecount, string);
   config_errors++;
 }
 
 void config_pwarn(const char *string)
 {
-  snmp_log(LOG_WARNING, "%s: line %d: %s\n", curfilename, linecount, string);
+  snmp_log(LOG_WARNING, "%s: line %d: Warning: %s\n", curfilename, linecount, string);
 }
 
 /* skip all white spaces and return 1 if found something either end of
    line or a comment character */
 char *skip_white(char *ptr)
 {
-
   if (ptr == NULL) return (NULL);
   while (*ptr != 0 && isspace(*ptr)) ptr++;
   if (*ptr == 0 || *ptr == '#') return (NULL);
@@ -732,7 +804,6 @@ char *skip_white(char *ptr)
 
 char *skip_not_white(char *ptr)
 {
-  
   if (ptr == NULL) return (NULL);
   while (*ptr != 0 && !isspace(*ptr)) ptr++;
   if (*ptr == 0 || *ptr == '#') return (NULL);
@@ -792,17 +863,27 @@ char *copy_word(char *from, char *to)
    followed by a string of hex */
 char *read_config_save_octet_string(char *saveto, u_char *str, size_t len) {
   int i;
-  if (str != NULL) {
-    sprintf(saveto, "0x");
-    saveto += 2;
-    for(i = 0; i < (int)len; i++) {
-      sprintf(saveto,"%02x", str[i]);
-      saveto = saveto + 2;
-    }
-    return saveto;
+  u_char *cp;
+
+  /* is everything easily printable */
+  for(i=0, cp=str; i < (int)len && cp &&
+          (isalpha(*cp) || isdigit(*cp) || *cp == ' '); cp++, i++);
+
+  if (len != 0 && i == (int)len) {
+      sprintf(saveto, "\"%s\"", str);
+      saveto += strlen(saveto);
   } else {
-    sprintf(saveto,"\"\"");
-    saveto += 2;
+      if (str != NULL) {
+          sprintf(saveto, "0x");
+          saveto += 2;
+          for(i = 0; i < (int)len; i++) {
+              sprintf(saveto,"%02x", str[i]);
+              saveto = saveto + 2;
+          }
+      } else {
+          sprintf(saveto,"\"\"");
+          saveto += 2;
+      }
   }
   return saveto;
 }
@@ -833,13 +914,9 @@ char *read_config_read_octet_string(char *readfrom, u_char **str, size_t *len) {
     }
     *len = *len / 2;
 
-    /* malloc data space if needed */
+    /* malloc data space if needed (+1 for good measure) */
     if (*str == NULL) {
-      if (*len == 0) {
-        /* null length string found */
-        cptr = NULL;
-
-      } else if (*len > 0 && (str == NULL || (cptr = (u_char *)malloc(*len)) == NULL)) {
+      if ((cptr = (u_char *)malloc(*len + 1)) == NULL) {
         return NULL;
       }
       *str = cptr;
@@ -847,26 +924,28 @@ char *read_config_read_octet_string(char *readfrom, u_char **str, size_t *len) {
       cptr = *str;
     }
 
-    /* copy data */
+    /* copy validated data */
     for(i = 0; i < (int)*len; i++) {
-      sscanf(readfrom,"%2x",&tmp);
-      *cptr++ = (u_char) tmp;
+      if (1 == sscanf(readfrom,"%2x",&tmp))
+        *cptr++ = (u_char) tmp;
+      else {
+        /* we may lose memory, but don't know caller's buffer XX free(cptr); */
+        return (NULL);
+      }
       readfrom += 2;
     }
+    *cptr++ = '\0';
     readfrom = skip_white(readfrom);
   } else {
     /* Normal string */
 
-    /* malloc data space if needed */
+    /* malloc string space if needed (including NULL terminator) */
     if (*str == NULL) {
       char buf[SNMP_MAXBUF];
       readfrom = copy_word(readfrom, buf);
 
       *len = strlen(buf);
-      /* malloc an extra space to add a null */
-      if (*len > 0 && (str == NULL ||
-                       (cptr = (u_char *) malloc(*len + 1))
-                       == NULL))
+      if (*len > 0 && ((cptr = (u_char *) malloc(*len + 1)) == NULL))
         return NULL;
       *str = cptr;
       if (cptr)
@@ -904,48 +983,29 @@ char *read_config_read_objid(char *readfrom, oid **objid, size_t *len) {
   if (objid == NULL || readfrom == NULL)
     return NULL;
 
-  if (*objid != NULL) {
-    char buf[SPRINT_MAX_LEN];
+  if (*objid == NULL) {
+      *len = 0;
+      if ((*objid = (oid*)malloc(MAX_OID_LEN * sizeof(oid))) == NULL)
+        return NULL;
+      *len = MAX_OID_LEN;
+  }
 
-    if (strncmp(readfrom,"NULL",4) == 0) {
+  if (strncmp(readfrom,"NULL",4) == 0) {
       /* null length oid */
       *len = 0;
-    } else {
-      /* read_objid is touchy with trailing stuff */
+  } else {
+      /* qualify the string for read_objid */
+      char buf[SPRINT_MAX_LEN];
       copy_word(readfrom, buf);
 
-      /* read the oid into the buffer passed to us */
       if (!read_objid(buf, *objid, len)) {
         DEBUGMSGTL(("read_config_read_objid","Invalid OID"));
+        *len = 0;
         return NULL;
       }
-    }
-    
-    readfrom = skip_token(readfrom);
-  } else {
-    if (strncmp(readfrom,"NULL",4) == 0) {
-      /* null length oid */
-      *len = 0;
-      readfrom = skip_token(readfrom);
-    } else {
-      /* space needs to be malloced.  Call ourself recursively to figure
-       out how long the oid actually is */
-      oid obuf[MAX_OID_LEN];
-      size_t obuflen = MAX_OID_LEN;
-      oid *oidp = obuf;
-      oid **oidpp = &oidp;   /* done this way for odd, untrue, gcc warnings */
-
-      readfrom = read_config_read_objid(readfrom, oidpp, &obuflen);
-
-      /* Then malloc and copy the results */
-      *len = obuflen;
-      if (*len > 0 && (*objid = (oid*)malloc(*len * sizeof(oid))) == NULL)
-        return NULL;
-
-      if (obuflen > 0)
-        memcpy(*objid, obuf, obuflen*sizeof(oid));
-    }
   }
+    
+  readfrom = skip_token(readfrom);
   return readfrom;
 }
 
@@ -957,14 +1017,11 @@ char *read_config_read_objid(char *readfrom, oid **objid, size_t *len) {
             NULL if an unknown type.
 */
 char *read_config_read_data(int type, char *readfrom, void *dataptr, size_t *len) {
-
   int *intp;
   char **charpp;
   oid  **oidpp;
 
-  if (dataptr == NULL || readfrom == NULL)
-    return NULL;
-  
+  if (dataptr && readfrom)
   switch(type) {
     case ASN_INTEGER:
       intp = (int *) dataptr;
@@ -995,14 +1052,11 @@ char *read_config_read_data(int type, char *readfrom, void *dataptr, size_t *len
             NULL if an unknown type.
 */
 char *read_config_store_data(int type, char *storeto, void *dataptr, size_t *len) {
-
   int *intp;
   u_char **charpp;
   oid  **oidpp;
 
-  if (dataptr == NULL || storeto == NULL)
-    return NULL;
-  
+  if (dataptr && storeto)
   switch(type) {
     case ASN_INTEGER:
       intp = (int *) dataptr;
@@ -1010,10 +1064,12 @@ char *read_config_store_data(int type, char *storeto, void *dataptr, size_t *len
       return (storeto + strlen(storeto));
       
     case ASN_OCTET_STR:
+      *storeto++ = ' ';
       charpp = (u_char **) dataptr;
       return read_config_save_octet_string(storeto, *charpp, *len);
 
     case ASN_OBJECT_ID:
+      *storeto++ = ' ';
       oidpp = (oid **) dataptr;
       return read_config_save_objid(storeto, *oidpp, *len);
 
