@@ -200,6 +200,7 @@ struct snmp_session {
 	/*
 	 * SNMPv3 fields
 	 */
+    u_char  isAuthoritative;    /* are we the authoritative engine? */
     u_char  *contextEngineID;	/* authoritative snmpEngineID */
     size_t  contextEngineIDLen; /* Length of contextEngineID */
     u_int   engineBoots;        /* initial engineBoots for remote engine */
@@ -269,14 +270,21 @@ extern void snmp_perror (const char *);
 extern void snmp_set_detail (const char *);
 
 #define SNMP_MAX_MSG_SIZE          1472 /* ethernet MTU minus IP/UDP header */
+#define SNMP_MAX_MSG_V3_HDRS       (4+3+4+7+7+3+7+16) /* fudge factor=16 */
 #define SNMP_MAX_ENG_SIZE          32
 #define SNMP_MAX_SEC_NAME_SIZE     256
 #define SNMP_MAX_SEC_NAME_SIZE     256
 #define SNMP_MAX_CONTEXT_SIZE      256
-#define SNMP_SEC_PARAM_BUF_SIZE     256
+#define SNMP_SEC_PARAM_BUF_SIZE    256
 
 /* set to one to ignore unauthenticated Reports */
 #define SNMPV3_IGNORE_UNAUTH_REPORTS 0
+
+/* authoritative engine definitions */
+#define SNMP_SESS_NONAUTHORITATIVE 0 /* should be 0 to default to this */
+#define SNMP_SESS_AUTHORITATIVE    1 /* don't learn engineIDs */
+#define SNMP_SESS_UNKNOWNAUTH      2 /* sometimes (like NRs) */
+
 /* to determine type of Report from varbind_list */
 #define REPORT_STATS_LEN 9
 #define REPORT_snmpUnknownSecurityModels_NUM 1
@@ -290,9 +298,25 @@ extern void snmp_set_detail (const char *);
 
 #define SNMP_DETAIL_SIZE        512
 
+#define SNMP_FLAGS_DONT_PROBE      0x100    /* don't probe for an engineID */
 #define SNMP_FLAGS_STREAM_SOCKET   0x80
 #define SNMP_FLAGS_LISTENING       0x40     /* Server stream sockets only */
 #define SNMP_FLAGS_SUBSESSION      0x20
+#define SNMP_FLAGS_STRIKE2         0x02
+#define SNMP_FLAGS_STRIKE1         0x01
+
+#define CLEAR_SNMP_STRIKE_FLAGS(x) \
+	x &= ~(SNMP_FLAGS_STRIKE2|SNMP_FLAGS_STRIKE1)
+
+	/*
+	 * returns '1' if the session is to be regarded as dead,
+	 * otherwise set the strike flags appropriately, and return 0
+	 */
+#define SET_SNMP_STRIKE_FLAGS(x) \
+	((   x & SNMP_FLAGS_STRIKE2 ) ? 1 :				\
+	 ((( x & SNMP_FLAGS_STRIKE1 ) ? ( x |= SNMP_FLAGS_STRIKE2 ) :	\
+	                                ( x |= SNMP_FLAGS_STRIKE1 )),	\
+	                                0))
 
 /*
  * Error return values.
@@ -392,6 +416,8 @@ struct variable_list {
     size_t	    val_len;
     oid name_loc[MAX_OID_LEN];  /* 90 percentile < 24. */
     u_char buf[40];             /* 90 percentile < 40. */
+    void *data;			/* (Opaque) hook for additional data */
+    int  index;
 };
 
 
@@ -480,9 +506,9 @@ void snmp_read (fd_set *);
  */
 void snmp_free_pdu (struct snmp_pdu *);
 
-void snmp_free_var (struct variable_list *);
+void snmp_free_var (struct variable_list *); /* frees just this one */
 
-void snmp_free_varbind(struct variable_list *var);
+void snmp_free_varbind(struct variable_list *var); /* frees all in list */
 
 /*
  * int snmp_select_info(numfds, fdset, timeout, block)
@@ -563,36 +589,41 @@ void snmp_timeout (void);
 
 #define RECEIVED_MESSAGE   1
 #define TIMED_OUT	   2
+#define SEND_FAILED	   3
 
 long snmp_get_next_msgid(void);
 long snmp_get_next_reqid(void);
 long snmp_get_next_sessid(void);
 long snmp_get_next_transid(void);
 /* provide for backwards compatibility */
-#define snmp_set_dump_packet(x) ds_set_boolean(DS_LIBRARY_ID, DS_LIB_DUMP_PACKET, x)
-#define snmp_get_dump_packet() (ds_get_boolean(DS_LIBRARY_ID, DS_LIB_DUMP_PACKET))
-#define snmp_set_quick_print(x) ds_set_boolean(DS_LIBRARY_ID, DS_LIB_QUICK_PRINT, x);
-#define snmp_get_quick_print() (ds_get_boolean(DS_LIBRARY_ID, DS_LIB_QUICK_PRINT))
-#define snmp_set_suffix_only(x) ds_set_int(DS_LIBRARY_ID, DS_LIB_PRINT_SUFFIX_ONLY, x);
-#define snmp_get_suffix_only() (ds_get_int(DS_LIBRARY_ID, DS_LIB_PRINT_SUFFIX_ONLY))
-#define snmp_set_full_objid(x) ds_set_boolean(DS_LIBRARY_ID, DS_LIB_PRINT_FULL_OID, x);
-#define snmp_get_full_objid() (ds_get_boolean(DS_LIBRARY_ID, DS_LIB_PRINT_SUFFIX_ONLY));
-#define snmp_set_random_access(x) ds_set_boolean(DS_LIBRARY_ID, DS_LIB_RANDOM_ACCESS, x);
-#define snmp_get_random_access() (ds_get_boolean(DS_LIBRARY_ID, DS_LIB_RANDOM_ACCESS))
-
+void snmp_set_dump_packet(int);
+int snmp_get_dump_packet(void);
+void snmp_set_quick_print(int);
+int snmp_get_quick_print(void);
+void snmp_set_suffix_only(int);
+int snmp_get_suffix_only(void);
+void snmp_set_full_objid(int);
+int snmp_get_full_objid(void);
+void snmp_set_random_access(int);
+int snmp_get_random_access(void);
 
 int snmp_oid_compare (const oid *, size_t, const oid *, size_t);
 void init_snmp (const char *);
 u_char *snmp_pdu_build (struct snmp_pdu *, u_char *, size_t *);
-int snmpv3_parse(struct snmp_pdu *, u_char *, size_t *, u_char  **);
+#ifdef USE_REVERSE_ASNENCODING
+u_char *snmp_pdu_rbuild (struct snmp_pdu *, u_char *, size_t *);
+#endif
+int snmpv3_parse(struct snmp_pdu *, u_char *, size_t *, u_char  **, struct snmp_session *);
 int snmpv3_dparse(struct snmp_pdu *, u_char *, size_t *, u_char  **, int);
 int snmpv3_packet_build(struct snmp_pdu *pdu, u_char *packet, size_t *out_length, u_char *pdu_data, size_t pdu_data_len);
+int snmpv3_packet_rbuild(struct snmp_pdu *pdu, u_char *packet, size_t *out_length, u_char *pdu_data, size_t pdu_data_len);
 int snmpv3_make_report(struct snmp_pdu *pdu, int error);
 int snmpv3_get_report_type(struct snmp_pdu *pdu);
 int snmp_pdu_parse(struct snmp_pdu *pdu, u_char *data, size_t *length);
 int snmp_pdu_dparse(struct snmp_pdu *pdu, u_char *data, size_t *length, int);
 u_char* snmpv3_scopedPDU_parse(struct snmp_pdu *pdu, u_char *cp, size_t *length);
 u_char* snmpv3_scopedPDU_dparse(struct snmp_pdu *pdu, u_char *cp, size_t *length, int);
+void snmp_store(const char *type);
 void snmp_shutdown(const char *type);
 struct variable_list *snmp_pdu_add_variable (struct snmp_pdu *, oid *, size_t, u_char, u_char *, size_t);
 struct variable_list *snmp_varlist_add_variable(struct variable_list **varlist,
