@@ -20,6 +20,9 @@
 #include <errno.h>
 #if HAVE_SYSLOG_H
 #include <syslog.h>
+#ifndef LOG_CONS	/* Interesting Ultrix feature */
+#include <sys/syslog.h>
+#endif
 #endif
 #if TIME_WITH_SYS_TIME
 # ifdef WIN32
@@ -35,6 +38,9 @@
 #  include <time.h>
 # endif
 #endif
+#if HAVE_NETINET_IN_H
+#include <netinet/in.h>
+#endif
 
 #if HAVE_STDARG_H
 #include <stdarg.h>
@@ -54,6 +60,7 @@
 #include "default_store.h"
 #include "snmp_logging.h"
 #include "callback.h"
+#include "system.h"
 #define LOGLENGTH 1024
 
 static int do_syslogging=0;
@@ -193,7 +200,7 @@ snmp_log_string (int priority, const char *string)
   if (do_filelogging || do_stderrlogging) {
 
     if (ds_get_boolean(DS_LIBRARY_ID, DS_LIB_LOG_TIMESTAMP) && newline) {
-      sprintf_stamp(NULL, (char *)&sbuf);
+      sprintf_stamp(NULL, sbuf);
     } else {
       strcpy(sbuf, "");
     }
@@ -212,15 +219,11 @@ snmp_vlog (int priority, const char *format, va_list ap)
 {
   char buffer[LOGLENGTH];
   int length;
-#if HAVE_VSNPRINTF
   char *dynamic;
 
   length=vsnprintf(buffer, LOGLENGTH, format, ap);
-#else
-  length=vsprintf(buffer, format, ap);
-#endif
 
-  if (length == 0)
+  if (length == 0) 
     return(0);		/* Empty string */
 
   if (length == -1) {
@@ -233,8 +236,7 @@ snmp_vlog (int priority, const char *format, va_list ap)
     return(0);
   }
 
-#if HAVE_VSNPRINTF
-  dynamic=malloc(length+1);
+  dynamic = (char *)malloc(length+1);
   if (dynamic==NULL) {
     snmp_log_string(LOG_ERR, "Could not allocate memory for log-message\n");
     snmp_log_string(priority, buffer);
@@ -244,13 +246,7 @@ snmp_vlog (int priority, const char *format, va_list ap)
   vsnprintf(dynamic, length+1, format, ap);
   snmp_log_string(priority, dynamic);
   free(dynamic);
-  return(0);
-
-#else
-  snmp_log_string(priority, buffer);
-  snmp_log_string(LOG_ERR, "Log-message too long!\n");
-  return(-3);
-#endif
+  return 0;
 }
 
 
