@@ -60,6 +60,8 @@
 #include "snmp-tc.h"
 #include "lcd_time.h"
 #include "scapi.h"
+#include "callback.h"
+#include "default_store.h"
 #include "snmpusm.h"
 
 #include "transform_oids.h"
@@ -73,11 +75,6 @@ static u_int salt_integer;
 	/* 1/2 of seed for the salt.   Cf. RFC2274, Sect 8.1.1.1.
 	 */
 
-int reportErrorOnUnknownID = 0;
-	/* Should be determined based on msg type.
-	 */
-
-static struct usmUser *initialUser = NULL;
 static struct usmUser *noNameUser = NULL;
 
 /*
@@ -120,13 +117,6 @@ usm_check_secLevel_vs_protocols(int level,
 }
 
 
-void
-usm_set_reportErrorOnUnknownID (int value)
-{
-	reportErrorOnUnknownID = value;
-}
-
-
 struct usmStateReference *
 usm_malloc_usmStateReference(void)
 {
@@ -535,9 +525,7 @@ usm_calc_offsets (
 	 */
 	if (secLevel == SNMP_SEC_LEVEL_AUTHPRIV)
 	{
-		scopedPduLen = ( scopedPduLen % 8 )
-					? ROUNDUP8(scopedPduLen)
-					: scopedPduLen + 8;
+                scopedPduLen = ROUNDUP8(scopedPduLen);
 
 		if ((*datalen = 
 			asn_predict_length (ASN_OCTET_STR,0,scopedPduLen))==-1)
@@ -605,7 +593,7 @@ usm_set_salt (	u_char		*iv,
 	/*
 	 * Sanity check.
 	 */
-	if ( !iv || !iv_length || !priv_salt || !msgSalt
+	if ( !iv || !iv_length || !priv_salt
 		|| (*iv_length != propersize_salt)
              	|| (priv_salt_length < propersize_salt) )
 	{
@@ -621,7 +609,8 @@ usm_set_salt (	u_char		*iv,
 	memcpy(iv,			&net_boots,	propersize_salt/2);
 	memcpy(iv+(propersize_salt/2), &net_salt_int,	propersize_salt/2);
 
-	memcpy(msgSalt, iv, propersize_salt);
+        if (msgSalt)
+            memcpy(msgSalt, iv, propersize_salt);
 
 
 	/* 
@@ -763,7 +752,6 @@ usm_generate_out_msg (
 						 * value (other then err).
 						 */
 
-
 	DEBUGMSGTL(("usm","USM processing has begun.\n"));
 
 	if (secStateRef != NULL)
@@ -807,7 +795,7 @@ usm_generate_out_msg (
 				== NULL &&
                      			secLevel != SNMP_SEC_LEVEL_NOAUTH)
 		{
-			DEBUGMSGTL(("usm","Unknown User\n"));
+			DEBUGMSGTL(("usm","Unknown User(%s)\n",secName));
 			usm_free_usmStateReference (secStateRef);
 			return SNMPERR_USM_UNKNOWNSECURITYNAME;
 		}
@@ -852,7 +840,7 @@ usm_generate_out_msg (
 	if (usm_check_secLevel_vs_protocols(
 		theSecLevel,
 		theAuthProtocol, theAuthProtocolLength,
-		theAuthProtocol, theAuthProtocolLength) == 1)
+		thePrivProtocol, thePrivProtocolLength) == 1)
 	{
 		DEBUGMSGTL(("usm","Unsupported Security Level\n"));
 		usm_free_usmStateReference (secStateRef);
@@ -920,7 +908,6 @@ usm_generate_out_msg (
 	memset (&ptr[globalDataLen], 0xFF, theTotalLength-globalDataLen);
 #endif /* SNMP_TESTING_CODE */
 
-
 	/* 
 	 * Do the encryption.
 	 */
@@ -932,10 +919,11 @@ usm_generate_out_msg (
 
 		/* XXX  Hardwired to seek into a 1DES private key!
 		 */
-		if ( usm_set_salt(	salt,		&salt_length,
+		if ( !thePrivKey ||
+		     ( usm_set_salt(    salt,	   &salt_length,
 					thePrivKey+8,	thePrivKeyLength-8,
                         		&ptr[privParamsOffset])
-						== -1 )
+						== -1 ) )
 		{
 			DEBUGMSGTL(("usm","Can't set DES-CBC salt.\n"));
 			usm_free_usmStateReference (secStateRef);
@@ -1016,24 +1004,32 @@ usm_generate_out_msg (
 		(u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR), seq_len);
 	
 	offSet = ptr_len - remaining;
+        DEBUGDUMPHEADER("send", "msgAuthoritativeEngineID");
 	asn_build_string (&ptr[offSet], &remaining,
 		(u_char)(ASN_UNIVERSAL|ASN_PRIMITIVE|ASN_OCTET_STR),
 		theEngineID, theEngineIDLength);
+        DEBUGINDENTLESS();
 	
 	offSet = ptr_len - remaining;
+        DEBUGDUMPHEADER("send", "msgAuthoritativeEngineBoots");
 	asn_build_int (&ptr[offSet], &remaining,
 		(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
 		&boots_long, sizeof(long));
+        DEBUGINDENTLESS();
 	
 	offSet = ptr_len - remaining;
+        DEBUGDUMPHEADER("send", "msgAuthoritativeEngineTime");
 	asn_build_int (&ptr[offSet], &remaining,
 		(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
 		&time_long, sizeof(long));
-	
+        DEBUGINDENTLESS();
+
 	offSet = ptr_len - remaining;
+        DEBUGDUMPHEADER("send", "msgUserName");
 	asn_build_string (&ptr[offSet], &remaining,
 		(u_char)(ASN_UNIVERSAL|ASN_PRIMITIVE|ASN_OCTET_STR),
 		(u_char *)theName, theNameLength);
+        DEBUGINDENTLESS();
 
 
 	/*
@@ -1158,6 +1154,463 @@ usm_generate_out_msg (
 
 }  /* end usm_generate_out_msg() */
 
+#ifdef USE_REVERSE_ASNENCODING
+int
+usm_rgenerate_out_msg (
+     int      msgProcModel,	/* (UNUSED) */
+
+     u_char  *globalData,	/* IN */
+		/* Pointer to msg header data will point to the beginning
+		 * of the entire packet buffer to be transmitted on wire,
+		 * memory will be contiguous with secParams, typically
+		 * this pointer will be passed back as beginning of
+		 * wholeMsg below.  asn seq. length is updated w/ new length.
+		 *
+		 * While this points to a buffer that should be big enough
+		 * for the whole message, only the first two parts
+		 * of the message are completed, namely SNMPv3Message and
+		 * HeaderData.  globalDataLen (next parameter) represents
+		 * the length of these two completed parts.
+		 */
+
+     size_t   globalDataLen,	/* IN - Length of msg header data.	*/
+     int      maxMsgSize,	/* (UNUSED) */
+     int      secModel,		/* (UNUSED) */
+     u_char  *secEngineID,	/* IN - Pointer snmpEngineID.		*/
+     size_t   secEngineIDLen,	/* IN - SnmpEngineID length.		*/
+     char    *secName,		/* IN - Pointer to securityName.	*/
+     size_t   secNameLen,	/* IN - SecurityName length.		*/
+     int      secLevel,		/* IN - AuthNoPriv, authPriv etc.	*/
+
+     u_char  *scopedPdu,	/* IN */
+		/* Pointer to scopedPdu will be encrypted by USM if needed
+		 * and written to packet buffer immediately following
+		 * securityParameters, entire msg will be authenticated by
+		 * USM if needed.
+		 */
+
+     size_t   scopedPduLen,	/* IN - scopedPdu length. */
+
+     void    *secStateRef,	/* IN */
+		/* secStateRef, pointer to cached info provided only for
+		 * Response, otherwise NULL.
+		 */
+
+     u_char *wholeMsg,         /* OUT */
+		/* Complete authenticated/encrypted message - typically
+		 * the pointer to start of packet buffer provided in
+		 * globalData is returned here, could also be a separate
+		 * buffer.
+		 */
+
+     size_t *wholeMsgLen)          /* IN/OUT - Len available, len returned. */
+{
+    size_t msgAuthParmLen = 0;
+#ifdef SNMP_TESTING_CODE
+    size_t theTotalLength;
+#endif
+
+    u_int           boots_uint;
+    u_int           time_uint;
+    long            boots_long;
+    long            time_long;
+
+    /*
+		Indirection because secStateRef values override parameters.
+
+		None of these are to be free'd - they are either pointing to
+		what's in the secStateRef or to something either in the
+		actual prarmeter list or the user list.
+	*/
+
+    char   *theName		 	= NULL;
+    u_int   theNameLength		= 0;
+    u_char *theEngineID		= NULL;
+    u_int   theEngineIDLength	= 0;
+    u_char *theAuthKey		= NULL;
+    u_int   theAuthKeyLength	= 0;
+    oid    *theAuthProtocol		= NULL;
+    u_int   theAuthProtocolLength	= 0;
+    u_char *thePrivKey		= NULL;
+    u_int   thePrivKeyLength	= 0;
+    oid    *thePrivProtocol		= NULL;
+    u_int   thePrivProtocolLength	= 0;
+    int     theSecLevel		= 0;	/* No defined const for bad
+                                         * value (other then err). */
+    size_t  salt_length	 = 0;
+    u_char  salt[BYTESIZE(USM_MAX_SALT_LENGTH)];
+    u_char  authParams[USM_MAX_AUTHSIZE];
+    u_char *authParamsPtr;
+    u_char *startcp;
+    u_char *endp = wholeMsg;
+    u_char iv[BYTESIZE(USM_MAX_SALT_LENGTH)];
+
+    DEBUGMSGTL(("usm","USM processing has begun.\n"));
+
+    if (secStateRef != NULL)
+	{
+            /* To hush the compiler for now.  XXX */
+            struct usmStateReference *ref
+                = (struct usmStateReference *)secStateRef;
+
+            theName		 	= ref->usr_name;
+            theNameLength		= ref->usr_name_length;
+            theEngineID		= ref->usr_engine_id;
+            theEngineIDLength	= ref->usr_engine_id_length;
+
+            if (!theEngineIDLength) {
+                theEngineID		= secEngineID;
+                theEngineIDLength	= secEngineIDLen;
+            }
+
+            theAuthProtocol		= ref->usr_auth_protocol;
+            theAuthProtocolLength	= ref->usr_auth_protocol_length;
+            theAuthKey		= ref->usr_auth_key;
+            theAuthKeyLength	= ref->usr_auth_key_length;
+            thePrivProtocol		= ref->usr_priv_protocol;
+            thePrivProtocolLength	= ref->usr_priv_protocol_length;
+            thePrivKey		= ref->usr_priv_key;
+            thePrivKeyLength	= ref->usr_priv_key_length;
+            theSecLevel		= ref->usr_sec_level;
+	}
+
+    /* 
+	 * Identify the user record.
+	 */
+    else
+	{
+            struct usmUser *user;
+
+            /* we do allow an unknown user name for
+               unauthenticated requests. */
+            if ( (user = 
+                  usm_get_user(secEngineID, secEngineIDLen, secName))
+                 == NULL &&
+                 secLevel != SNMP_SEC_LEVEL_NOAUTH)
+		{
+                    DEBUGMSGTL(("usm","Unknown User\n"));
+                    usm_free_usmStateReference (secStateRef);
+                    return SNMPERR_USM_UNKNOWNSECURITYNAME;
+		}
+
+            theName		 	= secName;
+            theNameLength		= secNameLen;
+            theEngineID		= secEngineID;
+            theSecLevel		= secLevel;
+            theEngineIDLength	= secEngineIDLen;
+            if (user) {
+                theAuthProtocol	= user->authProtocol;
+                theAuthProtocolLength	= user->authProtocolLen;
+                theAuthKey		= user->authKey;
+                theAuthKeyLength	= user->authKeyLen;
+                thePrivProtocol	= user->privProtocol;
+                thePrivProtocolLength	= user->privProtocolLen;
+                thePrivKey		= user->privKey;
+                thePrivKeyLength	= user->privKeyLen;
+            } else {
+                /* unknown users can not do authentication (obviously) */
+                theAuthProtocol	= usmNoAuthProtocol;
+                theAuthProtocolLength	= sizeof(usmNoAuthProtocol)/sizeof(oid);
+                theAuthKey		= NULL;
+                theAuthKeyLength	= 0;
+                thePrivProtocol	= usmNoPrivProtocol;
+                thePrivProtocolLength	= sizeof(usmNoPrivProtocol)/sizeof(oid);
+                thePrivKey		= NULL;
+                thePrivKeyLength	= 0;
+            }
+	}  /* endif -- secStateRef==NULL */
+
+
+    /*
+		From here to the end of the function, avoid reference to
+		secName, secEngineID, secLevel, and associated lengths.
+	*/
+
+
+	/* 
+	 * Check to see if the user can use the requested sec services.
+	 */
+    if (usm_check_secLevel_vs_protocols(
+        theSecLevel,
+        theAuthProtocol, theAuthProtocolLength,
+        thePrivProtocol, thePrivProtocolLength) == 1)
+	{
+            DEBUGMSGTL(("usm","Unsupported Security Level\n"));
+            usm_free_usmStateReference (secStateRef);
+            return SNMPERR_USM_UNSUPPORTEDSECURITYLEVEL;
+	}
+
+
+    /* 
+	 * Retrieve the engine information.
+	 *
+	 * XXX	No error is declared in the EoP when sending messages to
+         * 	unknown engines, processing continues w/ boots/time == (0,0).
+	 */
+    if (get_enginetime (theEngineID, theEngineIDLength, 
+                        &boots_uint, &time_uint, FALSE) == -1)
+	{
+            DEBUGMSGTL(("usm","%s\n", "Failed to find engine data."));
+	}
+
+    boots_long = boots_uint;
+    time_long  = time_uint;
+	
+    /* 
+	 * Do the encryption.
+	 */
+    if (theSecLevel == SNMP_SEC_LEVEL_AUTHPRIV)
+	{
+            u_char cyphertext[SNMP_MAX_MSG_SIZE];
+            size_t cyphertextLen = SNMP_MAX_MSG_SIZE;
+            
+            /* XXX  Hardwired to seek into a 1DES private key!
+             */
+            salt_length = BYTESIZE(USM_MAX_SALT_LENGTH);
+            if ( !thePrivKey ||
+                 ( usm_set_salt(    salt,	   &salt_length,
+                                    thePrivKey+8,	thePrivKeyLength-8,
+                                    iv)
+                   == -1 ) )
+		{
+                    DEBUGMSGTL(("usm","Can't set DES-CBC salt.\n"));
+                    usm_free_usmStateReference (secStateRef);
+                    return SNMPERR_USM_GENERICERROR;
+		}
+
+#ifdef SNMP_TESTING_CODE
+            if ( debug_is_token_registered("usm/dump") == SNMPERR_SUCCESS) {
+                dump_chunk("usm/dump", "This data was encrypted:",
+                           scopedPdu, scopedPduLen);
+            }
+#endif
+                
+            if ( sc_encrypt(
+                thePrivProtocol,	 thePrivProtocolLength,
+                thePrivKey,		 thePrivKeyLength,
+                salt,			 salt_length,
+                scopedPdu,		 scopedPduLen,
+                cyphertext,	&cyphertextLen)
+                 != SNMP_ERR_NOERROR )
+		{
+                    DEBUGMSGTL(("usm","DES-CBC error.\n"));
+                    usm_free_usmStateReference (secStateRef);
+                    return SNMPERR_USM_ENCRYPTIONERROR;
+		}
+
+            if (cyphertextLen > *wholeMsgLen) {
+                    DEBUGMSGTL(("usm","encrypted size too long.\n"));
+                    usm_free_usmStateReference (secStateRef);
+                    return SNMPERR_TOO_LONG;
+            }
+            
+            /* copy it back onto the outgoing packet */
+#ifdef SNMP_TESTING_CODE
+            theTotalLength = *wholeMsgLen;
+#endif
+            wholeMsg = asn_rbuild_string (wholeMsg, wholeMsgLen,
+                                          (u_char) (ASN_UNIVERSAL|ASN_PRIMITIVE|
+                                                    ASN_OCTET_STR),
+                                          cyphertext, cyphertextLen);
+
+#ifdef SNMP_TESTING_CODE
+            if ( debug_is_token_registered("usm/dump") == SNMPERR_SUCCESS) {
+                dump_chunk("usm/dump", "salt + Encrypted form:",
+                           salt, salt_length);
+                dump_chunk("usm/dump", "wholeMsg:",
+                           wholeMsg+1, theTotalLength - *wholeMsgLen);
+            }
+#endif
+
+            DEBUGMSGTL(("usm","Encryption successful.\n"));
+	}
+
+    /* 
+	 * No encryption for you!
+	 */
+    else
+	{
+            /* mark the scoped pdu beginning of the whole msg */
+            wholeMsg -= scopedPduLen;
+            *wholeMsgLen -= scopedPduLen;
+	}
+
+    /* Start encoding the msgSecurityParameters. */
+
+    /* msgPrivacyParameters (warning: assumes DES salt) */
+    startcp = wholeMsg;
+    wholeMsg = asn_rbuild_string (wholeMsg, wholeMsgLen,
+                                 (u_char)
+                                 (ASN_UNIVERSAL|ASN_PRIMITIVE|ASN_OCTET_STR),
+                                 iv, salt_length);
+    if (wholeMsg == NULL) {
+        DEBUGMSGTL(("usm","building privParams failed.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+
+
+    /* msgAuthenticationParameters (warnings assumes 0x00 by 12) */
+    if (theSecLevel == SNMP_SEC_LEVEL_AUTHNOPRIV
+        || theSecLevel == SNMP_SEC_LEVEL_AUTHPRIV) {
+        memset(authParams, 0, USM_MD5_AND_SHA_AUTH_LEN);
+        msgAuthParmLen = USM_MD5_AND_SHA_AUTH_LEN;
+    }
+    authParamsPtr = wholeMsg - msgAuthParmLen +1; /* remember this spot */
+    wholeMsg =
+        asn_rbuild_string (wholeMsg, wholeMsgLen,
+                          (u_char) (ASN_UNIVERSAL|ASN_PRIMITIVE|ASN_OCTET_STR),
+                          authParams, msgAuthParmLen);
+    if (wholeMsg == NULL) {
+        DEBUGMSGTL(("usm","building authParams failed.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+    
+    /* msgUserName */
+    DEBUGDUMPHEADER("send", "msgUserName");
+    wholeMsg =
+        asn_rbuild_string (wholeMsg, wholeMsgLen,
+                          (u_char)(ASN_UNIVERSAL|ASN_PRIMITIVE|ASN_OCTET_STR),
+                          (u_char *)theName, theNameLength);
+    DEBUGINDENTLESS();
+    if (wholeMsg == NULL) {
+        DEBUGMSGTL(("usm","building authParams failed.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+
+    
+    /* msgAuthoritativeEngineTime */
+    DEBUGDUMPHEADER("send", "msgAuthoritativeEngineTime");
+    wholeMsg =
+        asn_rbuild_int (wholeMsg, wholeMsgLen,
+                       (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
+                       &time_long, sizeof(long));
+    DEBUGINDENTLESS();
+    if (wholeMsg == NULL) {
+        DEBUGMSGTL(("usm","building msgAuthoritativeEngineTime failed.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+
+
+    /* msgAuthoritativeEngineBoots */
+    DEBUGDUMPHEADER("send", "msgAuthoritativeEngineBoots");
+    wholeMsg =
+        asn_rbuild_int (wholeMsg, wholeMsgLen,
+                       (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
+                       &boots_long, sizeof(long));
+    DEBUGINDENTLESS();
+    if (wholeMsg == NULL) {
+        DEBUGMSGTL(("usm","building msgAuthoritativeEngineBoots failed.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+
+
+    DEBUGDUMPHEADER("send", "msgAuthoritativeEngineID");
+    wholeMsg =
+        asn_rbuild_string (wholeMsg, wholeMsgLen,
+                          (u_char)(ASN_UNIVERSAL|ASN_PRIMITIVE|ASN_OCTET_STR),
+                          theEngineID, theEngineIDLength);
+    DEBUGINDENTLESS();
+    if (wholeMsg == NULL) {
+        DEBUGMSGTL(("usm","building msgAuthoritativeEngineID failed.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+
+    /* USM msgSecurityParameters sequence header */
+    wholeMsg =
+        asn_rbuild_sequence (wholeMsg, wholeMsgLen, 
+                            (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR),
+                            startcp - wholeMsg);
+    if (wholeMsg == NULL) {
+        DEBUGMSGTL(("usm","building usm security parameters failed.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+
+    /* msgSecurityParameters OCTET STRING wrapper */
+    wholeMsg =
+        asn_rbuild_header(wholeMsg, wholeMsgLen,
+                          (u_char)(ASN_UNIVERSAL|ASN_PRIMITIVE|ASN_OCTET_STR),
+                          startcp - wholeMsg);
+
+    /* copy in the rest of the headers */
+    if (*wholeMsgLen < globalDataLen) {
+        DEBUGMSGTL(("usm","building global data failed.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+    wholeMsg -= globalDataLen;
+    *wholeMsgLen -= globalDataLen;
+    memcpy(wholeMsg+1, globalData, globalDataLen);
+
+
+    /* total packet sequence */
+    wholeMsg =
+        asn_rbuild_sequence (wholeMsg, wholeMsgLen, 
+                            (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR),
+                            endp - wholeMsg);
+    if (wholeMsg == NULL) {
+        DEBUGMSGTL(("usm","building master packet sequence.\n"));
+        usm_free_usmStateReference (secStateRef);
+        return SNMPERR_TOO_LONG;
+    }
+
+    /* 
+     * Now, time to consider / do authentication.
+     */
+    if (theSecLevel == SNMP_SEC_LEVEL_AUTHNOPRIV
+        || theSecLevel == SNMP_SEC_LEVEL_AUTHPRIV)
+	{
+            size_t	temp_sig_len	= msgAuthParmLen;
+            u_char *temp_sig	= (u_char *) malloc (temp_sig_len);
+            if (temp_sig == NULL)
+		{
+                    DEBUGMSGTL(("usm","Out of memory.\n"));
+                    usm_free_usmStateReference (secStateRef);
+                    return SNMPERR_USM_GENERICERROR;
+		}
+
+            if ( sc_generate_keyed_hash (
+                theAuthProtocol,	 theAuthProtocolLength,
+                theAuthKey,		 theAuthKeyLength,
+                wholeMsg+1,		 endp - wholeMsg,
+                temp_sig,		&temp_sig_len)
+                 != SNMP_ERR_NOERROR )
+		{
+                    SNMP_FREE(temp_sig);
+                    DEBUGMSGTL(("usm","Signing failed.\n"));
+                    usm_free_usmStateReference (secStateRef);
+                    return SNMPERR_USM_AUTHENTICATIONFAILURE;
+		}
+
+            if (temp_sig_len != msgAuthParmLen)
+		{
+                    SNMP_FREE(temp_sig);
+                    DEBUGMSGTL(("usm","Signing lengths failed.\n"));
+                    usm_free_usmStateReference (secStateRef);
+                    return SNMPERR_USM_AUTHENTICATIONFAILURE;
+		}
+
+            memcpy (authParamsPtr, temp_sig, msgAuthParmLen);
+
+            SNMP_FREE(temp_sig);
+	}  /* endif -- create keyed hash */
+
+
+    usm_free_usmStateReference (secStateRef);
+
+    DEBUGMSGTL(("usm","USM processing completed.\n"));
+	
+    return SNMPERR_SUCCESS;
+
+}  /* end usm_generate_out_msg() */
+
+#endif /* */
 
 
 
@@ -1240,7 +1693,7 @@ usm_parse_security_parameters (
 	parse_ptr 	= value_ptr;
 	remaining_bytes = sequence_length;
 
-        DEBUGDUMPHEADER("dump_recv", "Parsing msgAuthoritativeEngineID\n");
+        DEBUGDUMPHEADER("recv", "msgAuthoritativeEngineID");
 	if ( (next_ptr
 		= asn_parse_string (parse_ptr, &remaining_bytes, &type_value,
 			secEngineID, secEngineIDLen)) == NULL )
@@ -1260,7 +1713,7 @@ usm_parse_security_parameters (
 	 * Retrieve the engine boots, notice switch in the way next_ptr and
 	 * remaining_bytes are used (to accomodate the asn code).
 	 */
-        DEBUGDUMPHEADER("dump_recv", "Parsing msgAuthoritativeEngineBoots\n");
+        DEBUGDUMPHEADER("recv", "msgAuthoritativeEngineBoots");
 	if ((next_ptr = asn_parse_int (next_ptr, &remaining_bytes, &type_value,
 		&boots_long, sizeof(long))) == NULL)
 	{
@@ -1281,16 +1734,16 @@ usm_parse_security_parameters (
 	/* 
 	 * Retrieve the time value.
 	 */
-        DEBUGDUMPHEADER("dump_recv", "Parsing msgAuthoritativeEngineTime\n");
+        DEBUGDUMPHEADER("recv", "msgAuthoritativeEngineTime");
 	if ((next_ptr = asn_parse_int (next_ptr, &remaining_bytes, &type_value,
 		&time_long, sizeof(long))) == NULL)
 	{
 		/* RETURN parse error */ return -1;
 	}
+        DEBUGINDENTLESS();
 
 	if (type_value != (u_char) (ASN_UNIVERSAL|ASN_PRIMITIVE|ASN_INTEGER))
 	{
-		  DEBUGINDENTLESS();
                   /* RETURN parse error */ return -1;
 	}
 
@@ -1302,7 +1755,8 @@ usm_parse_security_parameters (
 	 */
 	origNameLen = *secNameLen;
 
-        DEBUGDUMPHEADER("dump_recv", "Parsing msgUserName\n");
+
+        DEBUGDUMPHEADER("recv", "msgUserName");
 	if ( (next_ptr
 		= asn_parse_string (next_ptr, &remaining_bytes, &type_value,
 			(u_char *)secName, secNameLen)) == NULL )
@@ -1331,7 +1785,7 @@ usm_parse_security_parameters (
 	/* 
 	 * Retrieve the signature and blank it if there.
 	 */
-        DEBUGDUMPHEADER("dump_recv", "Parsing msgAuthenticationParameters\n");
+        DEBUGDUMPHEADER("recv", "msgAuthenticationParameters");
 	if ( (next_ptr
 		= asn_parse_string (next_ptr, &remaining_bytes, &type_value,
 			signature, signature_length)) == NULL )
@@ -1358,7 +1812,7 @@ usm_parse_security_parameters (
 	 *
 	 * Note that the next ptr is where the data section starts.
 	 */
-        DEBUGDUMPHEADER("dump_recv", "Parsing msgPrivacyParameters\n");
+        DEBUGDUMPHEADER("recv", "msgPrivacyParameters");
 	if ( (*data_ptr
 		= asn_parse_string (next_ptr, &remaining_bytes, &type_value,
 			salt, salt_length)) == NULL )
@@ -1406,13 +1860,13 @@ usm_check_and_update_timeliness(
 	int    *error)
 {
 	u_char	myID[USM_MAX_ID_LENGTH];
-	int	myIDLength = snmpv3_get_engineID(myID, USM_MAX_ID_LENGTH);
+	u_long	myIDLength = snmpv3_get_engineID(myID, USM_MAX_ID_LENGTH);
 	u_int	myBoots;
 	u_int	myTime;
 
 
 
-	if ( (myIDLength > USM_MAX_ID_LENGTH) || (myIDLength < 0) )
+	if ( (myIDLength > USM_MAX_ID_LENGTH) || (myIDLength == 0) )
 	{
 		/* We're probably already screwed...buffer overwrite.  XXX? */
 		DEBUGMSGTL(("usm","Buffer overflow.\n"));
@@ -1602,7 +2056,9 @@ usm_process_in_msg (
 	size_t  *scopedPduLen,	   /* IN/OUT - Len available, len returned.   */
 
 	size_t  *maxSizeResponse,  /* OUT    - Max size of Response PDU.      */
-	void   **secStateRf)	   /* OUT    - Ref to security state.         */
+	void   **secStateRf,       /* OUT    - Ref to security state.         */
+        struct snmp_session *sess, /* IN     - session which got the message  */
+        u_char msg_flags)	   /* IN     - v3 Message flags.              */
 {
 	size_t   remaining = wholeMsgLen
 				- (u_int)
@@ -1629,8 +2085,9 @@ usm_process_in_msg (
 	DEBUGMSGTL(("usm","USM processing begun...\n"));
 
 
-	if (secStateRef)		/* FIX -- huh?  destroy it? */
+	if (secStateRef)
 	{
+		usm_free_usmStateReference(*secStateRef);
 		*secStateRef = usm_malloc_usmStateReference();
 		if (*secStateRef == NULL)
 		{
@@ -1663,6 +2120,13 @@ usm_process_in_msg (
 	}
 
 
+	if (secLevel != SNMP_SEC_LEVEL_AUTHPRIV) {
+            /* pull these out now so reports can use them */
+            *scopedPdu = data_ptr;
+            *scopedPduLen = wholeMsgLen - (data_ptr - wholeMsg);
+            end_of_overhead = data_ptr;
+        }
+
 	if (secStateRef)
 	{
 		/* Cache the name, engine ID, and security level,
@@ -1695,7 +2159,11 @@ usm_process_in_msg (
 	 * Locate the engine ID record.
 	 * If it is unknown, then either create one or note this as an error.
 	 */
-	if (reportErrorOnUnknownID)
+	if ((sess && (sess->isAuthoritative == SNMP_SESS_AUTHORITATIVE ||
+                      (sess->isAuthoritative == SNMP_SESS_UNKNOWNAUTH &&
+                       (msg_flags & SNMP_MSG_FLAG_RPRT_BIT)))) ||
+            (!sess && (msg_flags & SNMP_MSG_FLAG_RPRT_BIT)))
+            
 	{
 		if (ISENGINEKNOWN(secEngineID, *secEngineIDLen)==FALSE)
 		{
@@ -1729,12 +2197,12 @@ usm_process_in_msg (
 		usm_get_user(secEngineID, *secEngineIDLen, secName))
 			== NULL )
 	{
-		DEBUGMSGTL(("usm","Unknown User.\n"));
-		if (snmp_increment_statistic (STAT_USMSTATSUNKNOWNUSERNAMES)==0)
-		{
-			DEBUGMSGTL(("usm","%s\n", "Failed to increment statistic."));
-		}
-		return SNMPERR_USM_UNKNOWNSECURITYNAME;
+	  DEBUGMSGTL(("usm","Unknown User(%s)\n",secName));		
+	  if (snmp_increment_statistic (STAT_USMSTATSUNKNOWNUSERNAMES)==0)
+	    {
+	      DEBUGMSGTL(("usm","%s\n", "Failed to increment statistic."));
+	    }
+	  return SNMPERR_USM_UNKNOWNSECURITYNAME;
 	}
 
 
@@ -1899,12 +2367,14 @@ usm_process_in_msg (
 
 #ifdef SNMP_TESTING_CODE
 		if ( debug_is_token_registered("usm/dump") == SNMPERR_SUCCESS) {
+			dump_chunk("usm/dump", "Cypher Text",
+						value_ptr, remaining);
+			dump_chunk("usm/dump", "salt + Encrypted form:",
+						salt, salt_length);
+			dump_chunk("usm/dump", "IV + Encrypted form:",
+						iv, iv_length);
 			dump_chunk("usm/dump", "Decrypted chunk:",
 						*scopedPdu, *scopedPduLen);
-			dump_chunk("usm/dump", "IV + Encrypted form:",
-						salt, salt_length);
-			dump_chunk("usm/dump", NULL,
-						value_ptr, remaining);
 		}
 #endif
 	}
@@ -1938,14 +2408,15 @@ usm_process_in_msg (
 
 void
 init_usm(void) {
-  snmp_register_callback(SNMP_CALLBACK_LIBRARY, SNMP_CALLBACK_POST_READ_CONFIG,
+  snmp_register_callback(SNMP_CALLBACK_LIBRARY,
+                         SNMP_CALLBACK_POST_PREMIB_READ_CONFIG,
                          init_usm_post_config, NULL);
 }
 
 /* 
  * initializations for the USM.
  *
- * Should be called after the configuration files have been read.
+ * Should be called after the (engineid) configuration files have been read.
  *
  * Set "arbitrary" portion of salt to a random number.
  */
@@ -1954,13 +2425,6 @@ init_usm_post_config(int majorid, int minorid, void *serverarg,
                      void *clientarg) {
   size_t	salt_integer_len = sizeof(salt_integer);
 
-  initialUser = usm_create_initial_user("initial", usmHMACMD5AuthProtocol,
-                                        USM_LENGTH_OID_TRANSFORM,
-                                        usmDESPrivProtocol,
-                                        USM_LENGTH_OID_TRANSFORM);
-  SNMP_FREE(initialUser->engineID);
-  initialUser->engineIDLen = 0;
-
   if ( sc_random((u_char *) &salt_integer, &salt_integer_len) != SNMPERR_SUCCESS )
   {
 	DEBUGMSGTL(("usm","sc_random() failed: using time() as salt.\n"));
@@ -1972,6 +2436,7 @@ init_usm_post_config(int majorid, int minorid, void *serverarg,
                                         USM_LENGTH_OID_TRANSFORM,
                                         usmDESPrivProtocol,
                                         USM_LENGTH_OID_TRANSFORM);
+
   SNMP_FREE(noNameUser->engineID);
   noNameUser->engineIDLen = 0;
 
@@ -2098,12 +2563,9 @@ usm_get_user_from_list(u_char *engineID, size_t engineIDLen,
           memcmp(ptr->engineID, engineID, engineIDLen) == 0)))
       return ptr;
   }
+
   /* return "" user used to facilitate engineID discovery */
   if (use_default && !strcmp(name, "")) return noNameUser;
-  /* this next line may be vestigial from when the draft used 'initial'
-     to discover engineID, also did not remove creation if 'inital' user
-     -gsm 2/6/99 */
-  if (use_default && !strcmp(name, "initial")) return initialUser;
   return NULL;
 }
 
@@ -2431,6 +2893,24 @@ usm_create_initial_user(const char *name, oid *authProtocol, size_t authProtocol
   return newUser;
 }
 
+/* this is a callback that can store all known users based on a
+   previously registered application ID */
+int
+usm_store_users(int majorID, int minorID, void *serverarg, void *clientarg)
+{
+  /* figure out our application name */
+  char *appname = (char *) clientarg;
+  if (appname == NULL)
+      appname = ds_get_string(DS_LIBRARY_ID, DS_LIB_APPTYPE);
+
+  /* save the user base */
+  usm_save_users("usmUser", appname);
+
+  /* never fails */
+  return SNMPERR_SUCCESS;
+}
+
+
 /* usm_save_users(): saves a list of users to the persistent cache */
 void
 usm_save_users(const char *token, const char *type)
@@ -2506,6 +2986,13 @@ usm_read_user(char *line)
   line = skip_token(line);
   line = read_config_read_octet_string(line, &user->engineID,
                                        &user->engineIDLen);
+
+  /* set the lcd entry for this engineID to the minimum boots/time
+     values so that its a known engineid and won't return a report pdu.
+     This is mostly important when receiving v3 traps so that the usm
+     will at least continue processing them. */
+  set_enginetime(user->engineID, user->engineIDLen, 1, 0, 0);
+
   line = read_config_read_octet_string(line, (u_char **)&user->name,
                                        &len);
   line = read_config_read_octet_string(line, (u_char **)&user->secName,
@@ -2619,6 +3106,7 @@ usm_set_user_password(struct usmUser *user, const char *token, char *line)
   size_t	 *keyLen;
   u_char	  userKey[SNMP_MAXBUF_SMALL];
   size_t	  userKeyLen = SNMP_MAXBUF_SMALL;
+  u_char	 *userKeyP = userKey;
   int		  type, ret;
 
   /*
@@ -2671,7 +3159,7 @@ usm_set_user_password(struct usmUser *user, const char *token, char *line)
       return;
     }
   } else if (type == 1) {
-    cp = read_config_read_octet_string(cp, (u_char **) &userKey, &userKeyLen);
+    cp = read_config_read_octet_string(cp, &userKeyP, &userKeyLen);
     
     if (cp == NULL) {
       config_perror("invalid user key");
