@@ -44,6 +44,14 @@
 #       include <stdlib.h>
 #endif
 
+/* Stuff needed for getHwAddress(...) */
+#ifdef HAVE_SYS_IOCTL_H
+#	include <sys/ioctl.h>
+#endif
+#ifdef HAVE_NET_IF_H
+#	include <net/if.h>
+#endif
+
 #if HAVE_DMALLOC_H
 #include <dmalloc.h>
 #endif
@@ -52,7 +60,6 @@
 #include "asn1.h"
 #include "snmpv3.h"
 #include "callback.h"
-#include "snmpusm.h"
 #include "snmp.h"
 #include "snmp_api.h"
 #include "snmp_impl.h"
@@ -65,12 +72,16 @@
 #include "snmp_debug.h"
 #include "snmp_logging.h"
 #include "default_store.h"
+#include "snmpusm.h"
 
 #include "transform_oids.h"
 
 static u_long		 engineBoots	   = 1;
+static unsigned int	 engineIDType	   = ENGINEID_TYPE_UCD_RND;
 static unsigned char	*engineID	   = NULL;
 static size_t		 engineIDLength	   = 0;
+static unsigned char	*engineIDNic	   = NULL;
+static unsigned int	 engineIDIsSet	   = 0; /* flag if ID set by config */
 static unsigned char	*oldEngineID	   = NULL;
 static size_t		 oldEngineIDLength = 0;
 static struct timeval	 snmpv3starttime;
@@ -83,6 +94,8 @@ static size_t	 defaultAuthTypeLen	= 0;
 static oid	*defaultPrivType	= NULL;
 static size_t	 defaultPrivTypeLen	= 0;
 
+static int getHwAddress(const char * networkDevice, char * addressOut);
+
 void
 snmpv3_authtype_conf(const char *word, char *cptr)
 {
@@ -193,6 +206,7 @@ setup_engineID(u_char **eidp, const char *text)
 {
   int		  enterpriseid	= htonl(ENTERPRISE_NUMBER),
 		  localsetup	= (eidp) ? 0 : 1;
+
 			/* Use local engineID if *eidp == NULL.  */
 #ifdef HAVE_GETHOSTNAME
   u_char	  buf[SNMP_MAXBUF_SMALL];
@@ -200,25 +214,81 @@ setup_engineID(u_char **eidp, const char *text)
 #endif
   u_char     *bufp = NULL;
   size_t	  len;
- 
-
-  /*
-   * Determine length of the engineID string.
-   */
-  if (text) {
-    len = 5+strlen(text);	/* 5 leading bytes+text. */
-
-  } else {
-    len = 5 + 4;		/* 5 leading bytes + four byte IPv4 address */
+  int	    localEngineIDType = engineIDType;
+  int tmpint;
+  time_t tmptime;
+  
+  engineIDIsSet = 1;
+  
+/* get the host name and save the information */
 #ifdef HAVE_GETHOSTNAME
-    gethostname((char *)buf, sizeof(buf));
-    hent = gethostbyname((char *)buf);
+  gethostname((char *)buf, sizeof(buf));
+  hent = gethostbyname((char *)buf);
+/* Determine if we are using IPV6 */
 #ifdef AF_INET6
+  /* see if they selected IPV4 or IPV6 support */
+  if ( (ENGINEID_TYPE_IPV6 == localEngineIDType ) || \
+       (ENGINEID_TYPE_IPV4 == localEngineIDType ) )
+  {
     if (hent && hent->h_addrtype == AF_INET6)
-      len += 12;		/* 16 bytes total for IPv6 address. */
+    {
+      localEngineIDType=ENGINEID_TYPE_IPV6;
+    }
+    else
+    {
+      /* Not IPV6 so we go with default */
+      localEngineIDType=ENGINEID_TYPE_IPV4;
+    }
+  }
+#else
+/* No IPV6 support.  Check if they selected IPV6 engineID type.  If so
+ * make it IPV4 for them */
+  if ( ENGINEID_TYPE_IPV6 == localEngineIDType )
+  {
+    localEngineIDType = ENGINEID_TYPE_IPV4;
+  }
 #endif
 #endif /* HAVE_GETHOSTNAME */
-  }  /* endif -- text (1) */
+
+  /* Determine if we have text and if so setup our localEngineIDType
+   * appropriately.  */
+  if ( NULL != text )
+  {
+      engineIDType = localEngineIDType = ENGINEID_TYPE_TEXT;
+  }
+  /* Determine length of the engineID string. */
+  len = 5;  /* always have 5 leading bytes */
+  switch(localEngineIDType)
+  {
+    case ENGINEID_TYPE_TEXT:
+      len += strlen(text);	/* 5 leading bytes+text. No NULL char */
+      break;
+#if defined(IFHWADDRLEN) && defined(SIOCGIFHWADDR)
+    case ENGINEID_TYPE_MACADDR: /* MAC address */
+      len += 6;   /* + 6 bytes for MAC address */
+      break;
+#endif
+    case ENGINEID_TYPE_IPV4: /* IPv4 */
+      len +=4; /* + 4 byte IPV4 address */
+      break;
+    case ENGINEID_TYPE_IPV6: /* IPv6 */
+      len += 16; /* + 16 byte IPV6 address */
+      break;
+    case ENGINEID_TYPE_UCD_RND: /* UCD specific encoding */
+        if (engineID) /* already setup, keep current value */
+            return engineIDLength;
+        if (oldEngineID) {
+            len = oldEngineIDLength;
+        } else {
+            len += sizeof(int) + sizeof(time_t);
+        }
+        break;
+    default:
+      snmp_log(LOG_ERR, "Unknown EngineID type requested for setup (%d).  Using IPv4.\n", localEngineIDType);
+      localEngineIDType=ENGINEID_TYPE_IPV4; /* make into IPV4 */
+      len += 4;		/* + 4 byte IPv4 address */
+      break;
+  } /* switch */
 
 
   /*
@@ -228,7 +298,6 @@ setup_engineID(u_char **eidp, const char *text)
     snmp_log_perror("setup_engineID malloc");
     return -1;
   }
-
   memcpy(bufp, &enterpriseid, sizeof(enterpriseid)); /* XXX Must be 4 bytes! */
   bufp[0] |= 0x80;
   
@@ -236,43 +305,88 @@ setup_engineID(u_char **eidp, const char *text)
   /*
    * Store the given text  -OR-   the first found IP address.
    */
-  if (text) {
-    bufp[4] = 4;
-    memcpy((char *)bufp+5, text, strlen(text));
+  switch (localEngineIDType)
+  {
+    case ENGINEID_TYPE_UCD_RND:
+      if (oldEngineID) {
+          /* keep our previous notion of the engineID */
+          memcpy(bufp, oldEngineID, oldEngineIDLength);
+      } else {
+          /*
+            Here we've desigend our own ENGINEID that is not based on
+            an address which may change and may even become conflicting
+            in the future like most of the default v3 engineID types
+            suffer from.
 
-  } else {
-    bufp[4] = 1;
+            Ours is built from 2 fairly random elements: a random number and
+            the current time in seconds.  This method suffers from boxes
+            that may not have a correct clock setting and random number
+            seed at startup, but few OSes should have that problem.
+          */
+          bufp[4] = ENGINEID_TYPE_UCD_RND;
+          tmpint = random();
+          memcpy(bufp+5, &tmpint, sizeof(tmpint));
+          tmptime = time(NULL);
+          memcpy(bufp+5+sizeof(tmpint), &tmptime, sizeof(tmptime));
+      }
+      break;
+    case ENGINEID_TYPE_TEXT:
+      bufp[4] = ENGINEID_TYPE_TEXT;
+      memcpy((char *)bufp+5, text, strlen(text));
+      break;
 #ifdef HAVE_GETHOSTNAME
-    gethostname((char *)buf, sizeof(buf));
-    hent = gethostbyname((char *)buf);
-
-    if (hent && hent->h_addrtype == AF_INET) {
-      memcpy(bufp+5, hent->h_addr_list[0], hent->h_length);
-
 #ifdef AF_INET6
-    } else if (hent && hent->h_addrtype == AF_INET6) {
-      bufp[4] = 2;
+    case ENGINEID_TYPE_IPV6:
+      bufp[4] = ENGINEID_TYPE_IPV6;
       memcpy(bufp+5, hent->h_addr_list[0], hent->h_length);
+      break;
 #endif
-
-    } else {		/* Unknown address type.  Default to 127.0.0.1. */
-
+#endif
+#if defined(IFHWADDRLEN) && defined(SIOCGIFHWADDR)
+    case ENGINEID_TYPE_MACADDR:
+	{ int x;
+		bufp[4] = ENGINEID_TYPE_MACADDR;
+		/* use default NIC if none provided */
+		if ( NULL == engineIDNic )
+		{
+			x = getHwAddress(DEFAULT_NIC,&bufp[5]);
+		}
+		else
+		{
+			x = getHwAddress(engineIDNic,&bufp[5]);
+		}
+		if ( 0 != x)
+		/* function failed fill MAC address with zeros */
+		{
+			memset(&bufp[5], 0, 6);
+		}
+	}
+      break;
+#endif
+    case ENGINEID_TYPE_IPV4:
+    default:
+      bufp[4] = ENGINEID_TYPE_IPV4;
+#ifdef HAVE_GETHOSTNAME
+      if (hent && hent->h_addrtype == AF_INET)
+      {
+	memcpy(bufp+5, hent->h_addr_list[0], hent->h_length);
+      }
+      else /* Unknown address type.  Default to 127.0.0.1. */
+      {
+	bufp[5] = 127;
+	bufp[6] = 0;
+	bufp[7] = 0;
+	bufp[8] = 1;
+      }
+#else /* HAVE_GETHOSTNAME */
+      /* Unknown address type.  Default to 127.0.0.1. */
       bufp[5] = 127;
       bufp[6] = 0;
       bufp[7] = 0;
       bufp[8] = 1;
-    }
-#else /* HAVE_GETHOSTNAME */
-    /* Unknown address type.  Default to 127.0.0.1. */
-    
-    bufp[5] = 127;
-    bufp[6] = 0;
-    bufp[7] = 0;
-    bufp[8] = 1;
 #endif /* HAVE_GETHOSTNAME */
-    
-  }  /* endif -- text (2) */
-
+      break;
+  }
 
   /*
    * Pass the string back to the calling environment, or use it for
@@ -295,26 +409,41 @@ setup_engineID(u_char **eidp, const char *text)
 void
 usm_parse_create_usmUser(const char *token, char *line) {
   char *cp;
-  char buf[SNMP_MAXBUF_MEDIUM];
+  char buf[SNMP_MAXBUF_MEDIUM], buf2[SNMP_MAXBUF_SMALL];
   struct usmUser *newuser;
   u_char	  userKey[SNMP_MAXBUF_SMALL];
   size_t	  userKeyLen = SNMP_MAXBUF_SMALL;
-  int ret;
+  size_t ret;
 
   newuser = usm_create_user();
 
   /* READ: Security Name */
   cp = copy_word(line, buf);
+
+  /* might be a -e ENGINEID argument */
+  if (strcmp(buf,"-e") == 0) {
+      /* get the specified engineid from the line */
+      cp = copy_word(cp, buf);
+      newuser->engineIDLen = hex_to_binary(buf, (u_char *)buf2);
+      if (newuser->engineIDLen <= 0) {
+          usm_free_user(newuser);
+          config_perror("invalid EngineID argument to -e");
+          return;
+      }
+      memdup(&newuser->engineID, (u_char *)buf2, newuser->engineIDLen);
+      cp = copy_word(cp, buf);
+  } else {
+      newuser->engineID = snmpv3_generate_engineID(&ret);
+      if ( ret == 0 ) {
+          usm_free_user(newuser);
+          return;
+      }
+      newuser->engineIDLen = ret;
+  }
+  
   newuser->secName = strdup(buf);
   newuser->name = strdup(buf);
 
-  newuser->engineID = snmpv3_generate_engineID(&ret);
-  if ( ret < 0 ) {
-    usm_free_user(newuser);
-    return;
-  }
-  newuser->engineIDLen = ret;
-
   if (!cp)
     goto add; /* no authentication or privacy type */
 
@@ -380,6 +509,7 @@ usm_parse_create_usmUser(const char *token, char *line) {
   if (!cp) {
     /* assume the same as the authentication key */
     memdup(&newuser->privKey, newuser->authKey, newuser->authKeyLen);
+    newuser->privKeyLen = newuser->authKeyLen;
   } else {
     cp = copy_word(cp, buf);
     /* And turn it into a localized key */
@@ -413,7 +543,9 @@ usm_parse_create_usmUser(const char *token, char *line) {
   }
 add:
   usm_add_user(newuser);
-  DEBUGMSGTL(("usmUser","created a new user %s\n", newuser->secName));
+  DEBUGMSGTL(("usmUser","created a new user %s at ", newuser->secName));
+  DEBUGMSGHEX(("usmUser", newuser->engineID, newuser->engineIDLen));
+  DEBUGMSG(("usmUser", "\n"));
 }
 
 /*******************************************************************-o-******
@@ -433,7 +565,82 @@ engineBoots_conf(const char *word, char *cptr)
   DEBUGMSGTL(("snmpv3","engineBoots: %d\n",engineBoots));
 }
 
+/*******************************************************************-o-******
+ * engineIDType_conf
+ *
+ * Parameters:
+ *	*word
+ *	*cptr
+ *
+ * Line syntax:
+ *	engineIDType <1 or 3>
+ *		1 is default for IPv4 engine ID type.  Will automatically
+ *		    chose between IPv4 & IPv6 if either 1 or 2 is specified.
+ *		2 is for IPv6.
+ *		3 is hardware (MAC) address, currently supported under Linux
+ */
+void
+engineIDType_conf(const char *word, char *cptr)
+{
+    engineIDType = atoi(cptr);
+    /* verify valid type selected */
+    switch (engineIDType) {
+        case ENGINEID_TYPE_IPV4: /* IPv4 */
+        case ENGINEID_TYPE_IPV6: /* IPv6 */
+            /* IPV? is always good */
+            break;
+#if defined(IFHWADDRLEN) && defined(SIOCGIFHWADDR)
+        case ENGINEID_TYPE_MACADDR: /* MAC address */
+            break;
+#endif
+        default:
+            /* unsupported one chosen */
+            config_perror("Unsupported enginedIDType, forcing IPv4");
+            engineIDType=ENGINEID_TYPE_IPV4;
+    }
+    DEBUGMSGTL(("snmpv3","engineIDType: %d\n",engineIDType));
+}
 
+/*******************************************************************-o-******
+ * engineIDNic_conf
+ *
+ * Parameters:
+ *	*word
+ *	*cptr
+ *
+ * Line syntax:
+ *	engineIDNic <string>
+ *		eth0 is default
+ */
+void
+engineIDNic_conf(const char *word, char *cptr)
+{
+  /* Make sure they haven't already specified the engineID via the
+   * configuration file */
+  if ( 0 == engineIDIsSet )
+  /* engineID has NOT been set via configuration file */
+  {
+    /* See if already set if so erase & release it */
+    if ( NULL != engineIDNic )
+    {
+      free(engineIDNic);
+    }
+    engineIDNic=(u_char *)malloc(strlen(cptr)+1);
+    if ( NULL != engineIDNic )
+    {
+      strcpy((char *)engineIDNic,cptr);
+      DEBUGMSGTL(("snmpv3","Initializing engineIDNic: %s\n", engineIDNic));
+    }
+    else
+    {
+      DEBUGMSGTL(("snmpv3","Error allocating memory for engineIDNic!\n"));
+    }
+  }
+  else
+  {
+    DEBUGMSGTL(("snmpv3","NOT setting engineIDNic, engineID already set\n"));
+  }
+}
 
 /*******************************************************************-o-******
  * engineID_conf
@@ -493,27 +700,23 @@ oldengineID_conf(const char *word, char *cptr)
  */
 void
 init_snmpv3(const char *type) {
+
   gettimeofday(&snmpv3starttime, NULL);
 
-  if (type == NULL)
-     type = "snmpapp";
-
-  if (type && !strcmp(type,"snmpapp")) {
-     setup_engineID(NULL,"__snmpapp__");
-  } else {
-     setup_engineID(NULL, NULL);
-  }
-
-  /* initialize submodules */
-  init_usm();
-
   /* we need to be called back later */
   snmp_register_callback(SNMP_CALLBACK_LIBRARY, SNMP_CALLBACK_POST_READ_CONFIG,
                          init_snmpv3_post_config, NULL);
+  snmp_register_callback(SNMP_CALLBACK_LIBRARY,
+                         SNMP_CALLBACK_POST_PREMIB_READ_CONFIG,
+                         init_snmpv3_post_premib_config, NULL);
   /* we need to be called back later */
   snmp_register_callback(SNMP_CALLBACK_LIBRARY, SNMP_CALLBACK_STORE_DATA,
-                         snmpv3_store, (void *) type);
+                         snmpv3_store, (void *) strdup(type));
 
+  /* initialize submodules */
+  /*   NOTE: this must be after the callbacks are registered above,
+             since they need to be called before the USM callbacks. */
+  init_usm();
 
 #if		!defined(USE_INTERNAL_MD5)
   /* doesn't belong here at all */
@@ -525,6 +728,8 @@ init_snmpv3(const char *type) {
   /* handle engineID setup before everything else which may depend on it */
   register_premib_handler(type,"engineID", engineID_conf, NULL, "string");
   register_premib_handler(type,"oldEngineID", oldengineID_conf, NULL, NULL);
+  register_premib_handler(type,"engineIDType", engineIDType_conf, NULL,"num");
+  register_premib_handler(type,"engineIDNic", engineIDNic_conf, NULL,"string");
   register_config_handler(type,"engineBoots", engineBoots_conf, NULL, NULL);
 
   /* default store config entries */
@@ -547,17 +752,17 @@ init_snmpv3(const char *type) {
   register_config_handler("snmp","defSecurityLevel", snmpv3_secLevel_conf,
                           NULL, "noAuthNoPriv|authNoPriv|authPriv");
   register_config_handler(type,"userSetAuthPass", usm_set_password, NULL,
-                          "secname engineIDLen engineID pass");
+                          NULL);
   register_config_handler(type,"userSetPrivPass", usm_set_password, NULL,
-                          "secname engineIDLen engineID pass");
+                          NULL);
   register_config_handler(type,"userSetAuthKey", usm_set_password, NULL,
-                          "secname engineIDLen engineID KuLen Ku");
+                          NULL);
   register_config_handler(type,"userSetPrivKey", usm_set_password, NULL,
-                          "secname engineIDLen engineID KuLen Ku");
+                          NULL);
   register_config_handler(type,"userSetAuthLocalKey", usm_set_password, NULL,
-                          "secname engineIDLen engineID KulLen Kul");
+                          NULL);
   register_config_handler(type,"userSetPrivLocalKey", usm_set_password, NULL,
-                          "secname engineIDLen engineID KulLen Kul");
+                          NULL);
 }
 
 /*
@@ -569,12 +774,12 @@ int
 init_snmpv3_post_config(int majorid, int minorid, void *serverarg,
                         void *clientarg) {
 
-  int engineIDLen;
+  size_t engineIDLen;
   u_char *c_engineID;
 
   c_engineID = snmpv3_generate_engineID(&engineIDLen);
 
-  if ( engineIDLen < 0 ) {
+  if ( engineIDLen == 0 ) {
     /* Somethine went wrong - help! */
     return SNMPERR_GENERR;
   }
@@ -596,6 +801,15 @@ init_snmpv3_post_config(int majorid, int minorid, void *serverarg,
   return SNMPERR_SUCCESS;
 }
 
+int
+init_snmpv3_post_premib_config(int majorid, int minorid, void *serverarg,
+                               void *clientarg) {
+    if (!engineIDIsSet)
+        setup_engineID(NULL, NULL);
+    
+    return SNMPERR_SUCCESS;
+}
+
 /*******************************************************************-o-******
  * store_snmpv3
  *
@@ -649,14 +863,14 @@ snmpv3_local_snmpEngineBoots(void)
  * Store engineID in buf; return the length.
  *
  */
-int
+size_t
 snmpv3_get_engineID(u_char *buf, size_t buflen)
 {
   /*
    * Sanity check.
    */
   if ( !buf || (buflen < engineIDLength) ) {
-    return SNMPERR_GENERR;
+    return 0;
   }
 
   memcpy(buf,engineID,engineIDLength);
@@ -689,9 +903,9 @@ snmpv3_clone_engineID(u_char **dest, size_t* destlen, u_char*src, size_t srclen)
   *dest = NULL; *destlen = 0;
 
   if (srclen && src) {
-    *dest = (u_char*)malloc((unsigned)srclen * sizeof(u_char));
+    *dest = (u_char*)malloc(srclen);
     if (*dest == NULL) return 0;
-    memmove(*dest, src, srclen * sizeof(u_char));
+    memmove(*dest, src, srclen);
     *destlen = srclen;
   }
   return *destlen;
@@ -714,7 +928,7 @@ snmpv3_clone_engineID(u_char **dest, size_t* destlen, u_char*src, size_t srclen)
  * 'length' is set to the length of engineID  -OR-  < 0 on failure.
  */
 u_char *
-snmpv3_generate_engineID(int *length)
+snmpv3_generate_engineID(size_t *length)
 {
   u_char *newID;
   newID = (u_char *) malloc(engineIDLength);
@@ -723,7 +937,7 @@ snmpv3_generate_engineID(int *length)
     *length = snmpv3_get_engineID(newID, engineIDLength);
   }
 
-  if (*length < 0) {
+  if (*length == 0) {
     SNMP_FREE(newID);
     newID = NULL;
   }
@@ -743,6 +957,66 @@ snmpv3_local_snmpEngineTime(void)
   return calculate_time_diff(&now, &snmpv3starttime)/100;
 }
 
+
+/* Code only for Linux systems */
+#if defined(IFHWADDRLEN) && defined(SIOCGIFHWADDR)
+static int getHwAddress(
+	const char * networkDevice, /* e.g. "eth0", "eth1" */
+	char * addressOut)	    /* return address. Len=IFHWADDRLEN */
+/*  getHwAddress(...)
+ *
+ *  This function will return a Network Interfaces Card's Hardware
+ *  address (aka MAC address).
+ *
+ *  Input Parameter(s):
+ *	networkDevice - a null terminated string with the name of a network
+ *			device.  Examples: eth0, eth1, etc...
+ *
+ *  Output Parameter(s):
+ *	addressOut -	This is the binary value of the hardware address.
+ *			This value is NOT converted into a hexadecimal string.
+ *			The caller must pre-allocate for a return value of
+ *			length IFHWADDRLEN
+ *
+ *  Return value:   This function will return zero (0) for success.  If
+ *		    an error occurred the function will return -1.
+ *
+ *  Caveats:	This has only been tested on Ethernet networking cards.
+ */
+{
+    int sock;		    /* our socket */
+    struct ifreq request;   /* struct which will have HW address */
+
+    if ( (NULL == networkDevice) || (NULL == addressOut) )
+    {
+	    return -1;
+    }
+    /* In order to find out the hardware (MAC) address of our system under
+     * Linux we must do the following:
+     * 1.  Create a socket
+     * 2.  Do an ioctl(...) call with the SIOCGIFHWADDRLEN operation.
+     */
+    sock = socket(AF_INET, SOCK_DGRAM, 0);
+    if (sock < 0)
+    {
+	return -1;
+    }
+    /* erase the request block */
+    memset(&request,0,sizeof(request));
+    /* copy the name of the net device we want to find the HW address for */
+    strncpy(request.ifr_name, networkDevice,IFNAMSIZ-1);
+    /* Get the HW address */
+    if (ioctl(sock, SIOCGIFHWADDR, &request))
+    {
+	close(sock);
+	return -1;
+    }
+    close(sock);
+    memcpy(addressOut, request.ifr_hwaddr.sa_data, IFHWADDRLEN);
+    return 0;
+}
+#endif
+
 #ifdef SNMP_TESTING_CODE
 /* snmpv3_set_engineBootsAndTime(): this function does not exist.  Go away. */
 /*   It certainly should never be used, unless in a testing scenero,
