@@ -54,7 +54,9 @@ DIR *opendir (const char *filename);
 struct direct *readdir (DIR *dirp);
 int closedir (DIR *dirp);
 
+#ifndef HAVE_GETTIMEOFDAY
 int gettimeofday (struct timeval *, struct timezone *tz);
+#endif
 #ifndef HAVE_STRCASECMP
 int strcasecmp(const char *s1, const char *s2);
 #endif
@@ -116,6 +118,15 @@ int calculate_time_diff(struct timeval *, struct timeval *);
 #ifndef HAVE_STRCASESTR
 char *strcasestr(const char *, const char *);
 #endif
+#ifndef HAVE_STRTOL
+long strtol(const char *, char **, int);
+#endif
+#ifndef HAVE_STRTOUL
+unsigned long strtoul(const char *, char **, int);
+#endif
+#ifndef HAVE_SNPRINTF
+int snprintf(char *, size_t, const char *, ...);
+#endif
 
 int mkdirhier(const char *pathname, mode_t mode, int skiplast);
 
