@@ -24,17 +24,17 @@
    counter64, i64, ui64, union? */
 #define OPAQUE_SPECIAL_TYPES 1
 
-/* comment the next line if you are compiling with libsnmp.h 
+/* comment the next line if you are compiling with libsnmp.h
    and are not using the UC-Davis SNMP library. */
 #define UCD_SNMP_LIBRARY 1
 
 /* define if you want to compile support for both authentication and
    privacy support. */
-#define SCAPI_AUTHPRIV 1 
- 
+#define SCAPI_AUTHPRIV 1
+
 /* define if you are using the internal MD5 code */
-#define USE_INTERNAL_MD5 1 
- 
+#define USE_INTERNAL_MD5 1
+
 /* add in recent CMU library extensions (not complete) */
 #define CMU_COMPATIBLE 1
 
@@ -155,7 +155,7 @@
 #define UNAMEPROG "/bin/uname"
 
 /* testing code sections. */
-/* #undef SNMP_TESTING_CODE */ 
+/* #undef SNMP_TESTING_CODE */
 
 /* If you don't have root access don't exit upon kmem errors */
 #undef NO_ROOT_ACCESS
@@ -178,6 +178,12 @@
 /* Define if you have the setenv function.  */
 #undef HAVE_SETENV
 
+/* Define if you have the strtoul function.  */
+#define HAVE_STRTOUL 1
+
+/* Define if you have the vsnprintf function.  */
+#define HAVE_VSNPRINTF 1
+
 /* Define if you have the index function.  */
 #undef HAVE_INDEX
 
@@ -526,10 +532,10 @@
 
 /* type check for in_addr_t */
 #define in_addr_t u_long
- 
+
 /* define if your compiler (processor) defines __FUNCTION__ for you */
 #undef HAVE_CPP_UNDERBAR_FUNCTION_DEFINED
- 
+
 /* mib pointer to the top of the extensible tree.  This has been
  assigned to UCDavis by the iana group.  Optionally, point this to the
  location in the tree your company/organization has been allocated. */
@@ -585,12 +591,12 @@
 
 /* which version are you using? This mibloc will tell you */
 #define VERSIONMIBNUM 100
-#define USEVERSIONMIB    
+#define USEVERSIONMIB
 
 /* Reports errors the agent runs into */
 /* (typically its "can't fork, no mem" problems) */
 #define ERRORMIBNUM 101
-#define USEERRORMIB    
+#define USEERRORMIB
 
 /* The sub id of EXENSIBLEMIB returned to queries of
    .iso.org.dod.internet.mgmt.mib-2.system.sysObjectID.0 */
@@ -601,7 +607,7 @@
    EXTENSIBLEMIB.AGENTID.???, where ??? is defined below by OSTYPE */
 
 #define HPUX9ID 1
-#define SUNOS4ID 2 
+#define SUNOS4ID 2
 #define SOLARISID 3
 #define OSFID 4
 #define ULTRIXID 5
@@ -656,7 +662,7 @@
 #endif
 
 /* how long to wait (seconds) for error querys before reseting the error trap.*/
-#define ERRORTIMELENGTH 600 
+#define ERRORTIMELENGTH 600
 
 /* Exec command to fix PROC problems */
 /* %s will be replaced by the process name in error */
@@ -673,7 +679,7 @@
    Don't define CASHETIME to disable cashing completely */
 
 #define EXCACHETIME 30
-#define CACHEFILE "/tmp/.snmp-exec-cache"
+#define CACHEFILE ".snmp-exec-cache"
 #define MAXCACHESIZE (200*80)   /* roughly 200 lines max */
 
 #define MAXDISKS 10                      /* can't scan more than this number */
@@ -797,6 +803,7 @@
 #define IN_UCD_SNMP_SOURCE 1
 
 int strcasecmp(const char *s1, const char *s2);
+#define vsnprintf _vsnprintf
 
 #define ENV_SEPARATOR ";"
 #define ENV_SEPARATOR_CHAR ';'
@@ -808,6 +815,11 @@ int strcasecmp(const char *s1, const char *s2);
 
 #endif
 
-typedef int mode_t;
+typedef unsigned short mode_t;
 
 #define AGENT_DIRECTORY_MODE 0700
+
+/* reverse encoding BER packets is both faster and more efficient in space. */
+#define USE_REVERSE_ASNENCODING       1
+#define DEFAULT_ASNENCODING_DIRECTION 1 /* 1 = reverse, 0 = forwards */
+
